package com.openexchange.datagenerator

import Values._

import scala.io.Source
import scala.util.matching.Regex
import scala.util.parsing.combinator.RegexParsers

/**
  * Parses a list of configuration files
  * @param files A list of configuration files as (filename, content) tuples
  * @param functions Function definitions used in the configuration files
  */
class Configuration(files: TraversableOnce[Configuration.Definition],
                    val functions: Functions)
{
  object parseConfig extends RegexParsers {
    override protected val whiteSpace: Regex = """(\s|#.*|//.*)+""".r
    def conf: Parser[List[(String, Val[_])]] = keyvalue.*
    def keyvalue: Parser[(String, Val[_])] = """\w+""".r ~ "=" ~ expr ^^ {
      case key ~ _ ~ value => (key, value)
    }
    def expr:     Parser[Val[_]]  = function | constant
    def function: Parser[Val[_]]  = """[A-Za-z]\w+""".r ~ "(" ~ params <~ ")" ^^ {
      case name ~ _ ~ params => functions(name)(params)
    }
    def params:   Parser[List[Val[_]]] = repsep(expr, ",")
    def constant: Parser[Val[_]]  = hex | number | string | boolean
    def hex:      Parser[NumVal]  = """0[Xx][0-9A-Fa-f]+""".r ^^
      (s => NumVal(s.substring(2).toLong))
    def number:   Parser[NumVal]  = """[+-]?(?:\d+(?:\.\d*)?|\.\d+)(?:[Ee][+-]?\d+)?""".r ^^
                                    ((s) => NumVal(s.toDouble))
    def string:   Parser[StrVal]  = """\"(?:\\.|[^\\"])*\"""".r ^^
                                    ((s) => StrVal(s.slice(1, s.length - 1)))
    def boolean:  Parser[BoolVal] = "true" ^^^ trueVal | "false" ^^^ falseVal

    def apply(definition: Configuration.Definition): List[(String, Val[_])] = {
      parseAll(conf, definition.content) match {
        case Success(result, _) => result
        case NoSuccess(msg, next) => scala.sys.error(s"$msg in ${definition.name} at ${next.pos}")
      }
    }
  }

  private val conf = files.flatMap(parseConfig(_)).toMap

  private def get[T](key: String, typeName: String): Val[T] = {
    val v = conf.getOrElse(key, throw new Exception(s"Missing configuration value: ${key}"))
    try {
      v.asInstanceOf[Val[T]]
    } catch {
      case e: ClassCastException => {
        throw new Exception(s"Expected ${key} to be ${typeName}")
      }
    }
  }

  def apply(key: String): Generator = new Generator(get(key, "num"))

  def has(key: String): Boolean = conf.contains(key)

  def num(key: String): Double      = get(key, "num").value()
  def int(key: String): Int         = get[Double](key, "num").value().toInt
  def str(key: String): String      = get(key, "str").value()
  def bool(key: String): Boolean    = get(key, "bool").value()
  def list[T](key: String): List[T] = get(key, "list").value()
  
  class Generator(count: NumVal) {
    def generate[T](body: => T): TraversableOnce[T] = {
      for (_ <- 0 until count.value().toInt) yield body
    }
    def generate[T](body: Int => T): TraversableOnce[T] = {
      for (i <- 0 until count.value().toInt) yield body(i)
    }
  }
}

object Configuration {

  trait Definition {
    def name: String
    def content: String
  }

  case class File(filename: String) extends Definition {
    def name = filename
    def content = Source.fromFile(filename).getLines().mkString("\n")
  }

  case class Line(line: String) extends Definition {
    def name = line
    def content = line
  }

}
