package com.openexchange.datagenerator.mail

import java.net.URL
import java.net.URLEncoder
import java.net.HttpURLConnection
import java.io.InputStream
import java.io.BufferedReader
import java.nio.charset.StandardCharsets.UTF_8
import java.util.UUID
import java.net.CookieHandler
import java.net.CookieManager
import java.io.OutputStream
import java.io.BufferedOutputStream

import Multipart._

class API(opt: Main.Options) {
  val server = new URL(opt.server)
  
  if (CookieHandler.getDefault == null) {
    CookieHandler.setDefault(new CookieManager())
  }
  
  case class HTTPError(code: Int, message: String) extends Exception {
    override def toString(): String = s"HTTP Error ${code} ${message}"
  }
  
  def encode(params: Map[String, String]): String = {
    params.map(x => URLEncoder.encode(x._1, "UTF-8") + '=' +
                    URLEncoder.encode(x._2, "UTF-8")
              ).mkString("&")
  }
  
  // Generic HTTP request
  
  def request(method: String, path: String,
              params: Map[String, String] = Map.empty,
              body: Writable = Iterator.empty,
              headers: Map[String, String] = Map("Accept-Charset" -> "UTF-8")):
  String = {
    val conn = new URL(server, path + '?' + encode(params)).openConnection()
    val http = conn.asInstanceOf[HttpURLConnection]
    http.setRequestMethod(method)
    headers.foreach(header => http.setRequestProperty(header._1, header._2))
    if (body.nonEmpty) {
      http.setDoOutput(true)
      val out = new BufferedOutputStream(http.getOutputStream())
      body.foreach(out.write(_))
      out.flush()
    }
    if (http.getResponseCode() != HttpURLConnection.HTTP_OK) {
      throw new HTTPError(http.getResponseCode(), http.getResponseMessage())
    }
    val is = http.getInputStream()
    val buf = new Array[Byte](4096)
    val sb = new StringBuilder()
    Iterator.continually(is.read(buf)).takeWhile(_ != -1).foreach(n => {
      sb.append(new String(buf, 0, n, UTF_8))
    })
    sb.result()
  }
  
  // Various HTTP methods
  
  def post(path: String, params: Map[String, String]): String =
    request("POST", path, Map.empty, Iterator(encode(params).getBytes(UTF_8)),
      Map("Content-Type" -> "application/x-www-form-urlencoded"))
  
  def post(path: String, params: Map[String, String],
           files: Iterator[Writable]): String = {
    val boundary = UUID.randomUUID.toString

    request("POST", path, params, multipart(files.zipWithIndex map {
        case (file, i) => part(Map(
            "Content-Disposition" ->
              s"form-data; name=file${i}; filename=file${i}.eml",
            "Content-Type" -> "message/rfc822"
          ), file)
      }, boundary),
      Map("Accept-Charset" -> "UTF-8",
          "Content-Type" -> s"multipart/form-data; boundary=${boundary}"))
  }
  
  // HTTP API calls
  
  def login(name: String, password: String): String = {
    post("api/login", Map("action" -> "login",
                          "name" -> name,
                          "password" -> password))
  }
  
  def logout(session: String): String =
    request("GET", "api/login", Map("action" -> "logout", "session" -> session))
  
  def getMyUser(session: String): String =
    request("GET", "api/user/me", Map("session" -> session))
  
  def importMails(mails: TraversableOnce[Writable],
                  folder: String, session: String): String = {
    post("api/mail",
      Map("action"  -> "import",
          "folder"  -> s"default0/${folder}",
          "force"   -> "true",
          "session" -> session),
      mails.toIterator)
  }
}
