package com.openexchange.datagenerator.mail

import java.time.Instant
import java.util.Date

import com.openexchange.datagenerator.Configuration
import com.openexchange.datagenerator.Functions

import scala.collection.mutable

object Main {
  def error(message: String) {
    println("Error: " + message)
    println
    usage(1)
  }
  
  def usage(exitCode: Int) {
    println("""Usage: scala mailgenerator.jar [OPTIONS] [INPUTFILE]

Where INPUTFILE is a text file with one full login name per line.
If not specified, the standard input is used.

The OPTIONS are:

  -f  --config=FILE        Adds a configuration file for generated data.
                           This option must be specified at least once.
                           All files are effectively concatenated.
                           The last definition of each key wins.

  -D KEY=VALUE             Defines a configuration value, effectively
  --define=KEY=VALUE       inserting a line between previous and subsequent
                           configuration files and values.

  -h  --help               Print this usage information.

  -n  --now=DATETIME       The starting date and time in ISO-8601 format
                           (yyyy-mm-ddThh:mm:ssZ).
                           Defaults to the current date and time.

  -p  --password=PASSWORD  Password to use for all users
                           Default: secret

  -r  --seed=NUMBER        Initializes the random number generator with
                           the specified seed number.
                           Defaults to the current timestamp in ms.

  -s  --server=URL         URL of an OX App Suite server.
                           Default: http://localhost:8009/appsuite/
""")
    System.exit(exitCode)
  }
  
  class Options {
    val config: mutable.MutableList[Configuration.Definition] = mutable.MutableList.empty
    var input: String = "-"
    var now: Instant = Instant.now
    var password: String = "secret"
    var seed: Long = new Date().getTime
    var server: String = "http://localhost:8009/appsuite/"
  }
  
  def parseArgs(args: Iterator[String]): Options = {
    val opt = new Options
    
    def nextArg(value: String): String = {
      if (value != "") value
      else if (args.hasNext) args.next()
      else { usage(1); "" }
    }
    
    val shortArgs = """-([h]*)(?:([Dfnprst])(.*))?""".r
    val longArg = """--(define|config|now|password|seed|server)=(.*)""".r
    val invalidArgs = """--?(.+)""".r
    
    while (args.hasNext) {
      args.next() match {
        case shortArgs(flags, param, value) => {
          for (flag <- flags) flag match {
            case 'h' => usage(0)
          }
          param match {
            case "D" => opt.config  += Configuration.Line(nextArg(value))
            case "f" => opt.config  += Configuration.File(nextArg(value))
            case "n" => opt.now      = Instant.parse(nextArg(value))
            case "p" => opt.password = nextArg(value)
            case "r" => opt.seed     = nextArg(value).toLong
            case "s" => opt.server   = nextArg(value)
            case ""  => if (flags == "") opt.input = "-"
          }
        }
        case "--help"   => usage(0)
        case longArg(name, value) => name match {
          case "define"   => opt.config  += Configuration.Line(value)
          case "config"   => opt.config  += Configuration.File(value)
          case "now"      => opt.now      = Instant.parse(value)
          case "password" => opt.password = value
          case "seed"     => opt.seed     = value.toLong
          case "server"   => opt.server   = value
        }
        case "--" => while (args.hasNext) opt.input = args.next()
        case invalidArgs(arg) => error(s"Invalid argument '$arg'.")
        case value => opt.input = value
      }
    }
    
    opt
  }
  
  def main(args: Array[String]) {
    val opt = parseArgs(args.iterator)
    if (opt.config.isEmpty) error("No configuration files (-f) specified.")
    if (opt.input == "-") opt.input = "/dev/stdin"
    val config = new Configuration(opt.config, new Functions(opt.now, opt.seed))
    val generator = new Mails(config, opt)
    generator.generate()
  }
}
