package com.openexchange.datagenerator.sql

import java.util.Date

import scala.io.Source
import java.time.Instant

import com.openexchange.datagenerator.{Configuration, Functions}

import scala.collection.mutable

object Main {
  def error(message: String) {
    println("Error: " + message)
    println
    usage(1)
  }
  
  def usage(exitCode: Int) {
    println("""Usage: ox-data-generator [OPTIONS] INPUT

Where INPUT is an SQL dump of an empty database with already created contexts.

The OPTIONS are:

  -b NUMBER  --batch=NUMBER  Number of rows per generated INSERT statement.
                             Default: 1000.

  -c CID     --cid=CID       The context to use as template.
                             Defaults to the first CID found in the input dump.

  -d PATH    --drive=PATH    Root path of the Drive filestore.
                             Default: /var/opt/open-xchange/filestore/

  -f FILE    --config=FILE   Adds a configuration file for generated data.
                             This option must be specified at least once.
                             All files are effectively concatenated.
                             The last definition of each key wins.

  -D KEY=VALUE               Defines a configuration value, effectively
  --define=KEY=VALUE         inserting a line between previous and subsequent
                             configuration files and values.

  -h         --help          Print this usage information.

  -l FILE    --logins=FILE   Dumps all generated logins to a file for use by
                             the mail generator.

  -n         --now=DATETIME  Specifies the current date and time in ISO-8601
                             format (yyyy-mm-ddThh:mm:ssZ).
                             Defaults to the actual current date and time.

  -o FILE    --output=FILE   The output file. Defaults to stdout.

  -r NUMBER  --seed=NUMBER   Initializes the random number generator with
                             the specified seed number.
                             Defaults to the current timestamp in ms.
  -s NUMBER  --swap=NUMBER   Threshold file size at which an in-memory table is
                             swapped to a temporary file.
                             Defaults to 1 MB.
  -t PATH    --tmpdir=PATH   Directory for temporary files.
                             Defaults to the value of java.io.tmpdir.
""")
    System.exit(exitCode)
  }

  class Options {
    var batch: Int = 1000
    var cid: Int = -1
    var drive: String = "/var/opt/open-xchange/filestore/"
    val config: mutable.MutableList[Configuration.Definition] = mutable.MutableList.empty
    var input: String = "-"
    var logins: Option[String] = None
    var now: Instant = Instant.now
    var output: String = "-"
    var seed: Long = new Date().getTime
    var tmpdir: String = sys.props("java.io.tmpdir")
    var swap: Int = 1 << 20
  }
  
  def parseArgs(args: Iterator[String]): Options = {
    val opt = new Options
    
    def nextArg(value: String): String = {
      if (value != "") value
      else if (args.hasNext) args.next()
      else { usage(1); "" }
    }
    
    val shortArgs = """-([h]*)(?:([bcdDflmnorst])(.*))?""".r
    val longArg = """--(batch|cid|drive|define|config|logins|maildir|now|output|seed|swap|tmpdir)=(.*)""".r
    val invalidArgs = """--?(.+)""".r
    
    while (args.hasNext) {
      args.next() match {
        case shortArgs(flags, param, value) => {
          for (flag <- flags) flag match {
            case 'h' => usage(0)
          }
          param match {
            case "b" => opt.batch   = nextArg(value).toInt
            case "c" => opt.cid     = nextArg(value).toInt
            case "d" => opt.drive   = nextArg(value)
            case "D" => opt.config += Configuration.Line(nextArg(value))
            case "f" => opt.config += Configuration.File(nextArg(value))
            case "l" => opt.logins  = Some(nextArg(value))
            case "n" => opt.now     = Instant.parse(nextArg(value))
            case "o" => opt.output  = nextArg(value)
            case "r" => opt.seed    = nextArg(value).toLong
            case "s" => opt.swap    = nextArg(value).toInt
            case "t" => opt.tmpdir  = nextArg(value)
            case ""  => if (flags == "") opt.input = "-"
          }
        }
        case "--help"   => usage(0)
        case longArg(name, value) => name match {
          case "batch"   => opt.batch   = value.toInt
          case "cid"     => opt.cid     = value.toInt
          case "drive"   => opt.drive   = value
          case "define"  => opt.config += Configuration.Line(value)
          case "config"  => opt.config += Configuration.File(value)
          case "logins"  => opt.logins  = Some(value)
          case "now"     => opt.now     = Instant.parse(value)
          case "output"  => opt.output  = value
          case "seed"    => opt.seed    = value.toLong
          case "swap"    => opt.swap    = value.toInt
          case "tmpdir"  => opt.tmpdir  = value
        }
        case "--" => while (args.hasNext) opt.input = args.next()
        case invalidArgs(arg) => error(s"Invalid argument '$arg'.")
        case value => opt.input = value
      }
    }
    
    opt
  }
  
  def main(args: Array[String]) {
    val opt = parseArgs(args.iterator)
    if (opt.config.isEmpty) error("No configuration specified (use -f, -D).")
    if (opt.input == "-") opt.input = "/dev/stdin"
    if (opt.output == "-") opt.output = "/dev/stdout"
    if (!opt.drive.endsWith("/")) opt.drive += "/"
    
    val config = new Configuration(opt.config, new Functions(opt.now, opt.seed))
    val input = Source.fromFile(opt.input).getLines()
    val sql = new SQL(input, opt)
    val tables = new Tables(config, sql)
    tables.generate()
    sql.writeTo(opt.output)
  }
}
