package com.openexchange.datagenerator.sql

import java.nio.ByteBuffer
import java.nio.channels.WritableByteChannel
import java.nio.channels.FileChannel
import java.nio.charset.StandardCharsets.UTF_8
import java.nio.file.Files
import java.nio.file.Paths
import java.nio.file.StandardOpenOption._
import java.nio.CharBuffer
import java.nio.charset.CoderResult
import scala.annotation.tailrec

// 100 contexts without buffering
// real	0m4.490s
// user	0m9.168s
// sys	0m0.392s

class Output(table: String, opt: Main.Options,
             prefix: StringBuilder = new StringBuilder(0x4000))
{

  trait ConcreteOutput {
    def append(s: String): Boolean
    def writeTo(output: WritableByteChannel)
  }
  
  class StringBufferOutput(sb: StringBuilder) extends ConcreteOutput {
    def append(s: String): Boolean = {
      sb ++= s
      sb.length > opt.swap
    }
    
    def writeTo(output: WritableByteChannel) {
      output.write(ByteBuffer.wrap(sb.result().getBytes(UTF_8)))
    }
  }
  
  class FileOutput(prefix: StringBuilder) extends ConcreteOutput {
    val path = Files.createTempFile(Paths.get(opt.tmpdir), null, null)
    val fc = FileChannel.open(path, READ, WRITE, CREATE, DELETE_ON_CLOSE)
    val enc = UTF_8.newEncoder()
    val buf = ByteBuffer.allocate(opt.swap)

    append(prefix.result())
    
    def append(s: String): Boolean = {
      val inbuf = CharBuffer.wrap(s)
      encode(() => enc.encode(inbuf, buf, false))
      false
    }
    
    @tailrec
    private def encode(f: () => CoderResult) {
        f() match {
          case CoderResult.OVERFLOW => {
            buf.flip()
            fc.write(buf)
            buf.compact()
            encode(f)
          }
          case CoderResult.UNDERFLOW =>
          case error => error.throwException()
        }
    }
    
    def writeTo(output: WritableByteChannel) {
      enc.encode(CharBuffer.allocate(0), buf, true)
      encode(() => enc.flush(buf))
      buf.flip()
      fc.write(buf)
      fc.transferTo(0, fc.size, output)
      fc.close()
    }
  }
  
  private var out: ConcreteOutput = new StringBufferOutput(prefix)
  private var batch = 0

  def append(s: String) = if (out.append(s)) out = new FileOutput(prefix)

  def insert(values: String) = {
    if (batch == 0) {
      append(s"INSERT INTO $table VALUES (")
      batch = opt.batch - 1
    } else {
      append("),(")
      batch -= 1
    }
    append(values)
    if (batch == 0) append(");\n")
  }

  def writeTo(output: WritableByteChannel) = {
    if (batch > 0) append(");\n")
    out.writeTo(output)
  }
  
}
