package com.openexchange.datagenerator.sql

import java.nio.ByteBuffer
import java.nio.channels.FileChannel
import java.nio.channels.WritableByteChannel
import java.nio.charset.StandardCharsets.UTF_8
import java.nio.file.Files
import java.nio.file.Paths
import java.nio.file.StandardOpenOption._
import java.time.Instant
import java.time.LocalDate
import java.time.LocalTime

import scala.util.parsing.combinator.RegexParsers
import scala.collection.mutable

class SQL(input: Iterator[String], val opt: Main.Options) {
  val create = """CREATE TABLE `(\w+)` \(""".r
  val column = """  `(\w+)` (\w+).*?(?:DEFAULT (NULL|CURRENT_TIMESTAMP|'.+?'|.*?),)?.*""".r
  val start  = """/\*!40000 ALTER TABLE `(\w+)` DISABLE KEYS \*/;""".r
  val end    = """/\*!40000 ALTER TABLE `(\w+)` ENABLE KEYS \*/;""".r
  val insert = """INSERT INTO `(\w+)` VALUES (.*);""".r

  val value  = """('(?:''|\\.|[^'\\])*'|[^,\)]*)(,|\))""".r

  object parseValues extends RegexParsers {
    type SQLParser = Parser[SQLValue]
    type ListParser = Parser[List[SQLValue]]
    
    def int:    Parser[SQLInt]    = """[+-]?\d+""".r               ^^ { _.toLong }
    def float:  Parser[SQLFloat]  = """[+-]?(?:\d+(?:\.\d*)?|\.\d+)(?:[Ee][+-]?\d+)?""".r ^^ { _.toDouble }
    def string: Parser[SQLString] = "'" ~> """(?:''|\\.|[^'\\])*""".r <~ "'" ^^ Literals.unescape
    def bin1:   Parser[SQLBinary] = """[Xx]'""".r ~> byte.* <~ "'" ^^ { _.toSeq }
    def bin2:   Parser[SQLBinary] = """0[Xx]""".r ~> byte.*        ^^ { _.toSeq }
    def byte:   Parser[Byte]      = """[0-9A-Fa-f]{2}""".r         ^^ { Integer.parseInt(_, 16).toByte }
    def binary: Parser[SQLBinary] = bin1 | bin2
    def date:   Parser[SQLDate]   = """\d{4}-\d\d-\d\d""".r        ^^ { LocalDate.parse(_) }
    def time:   Parser[SQLTime]   = """\d\d:\d\d:\d\d""".r         ^^ { LocalTime.parse(_) }
    def Null:   Parser[SQLValue]  = """(?i)NULL""".r               ^^ ((_: String) => NULL)
    
    def datetime: Parser[SQLDateTime] = date ~ """[ T]""".r ~ time ^^ {
      case d ~ _ ~ t => d.value.atTime(t.value)
    }
    def timestamp: Parser[SQLTimestamp] = """[+-]?\d+""".r         ^^ { n => Instant.ofEpochMilli(n.toLong) }
    
    def insert(values: ListParser) = repsep("(" ~> values <~ ")", ",")
    
    val typeParsers: Map[Symbol, SQLParser] = Map(
        'bigint -> int,
        'binary -> binary,
        'blob -> binary,
        'date -> date,
        'datetime -> datetime,
        'decimal -> float,
        'int -> int,
        'longblob -> binary,
        'mediumblob -> binary,
        'smallint -> int,
        'text -> string,
        'timestamp -> timestamp,
        'tinyint -> int,
        'varbinary -> binary,
        'varchar -> string)
    
    def parseList(list: List[SQLParser]): ListParser = {
      val rev = list.reverse
      val initial: ListParser = (rev.head | Null) ^^ { List(_) }
      rev.tail.foldLeft(initial) { (lp: ListParser, p: SQLParser) =>
        (p | Null) ~ ("," ~> lp) ^^ { case v ~ vs => v :: vs }
      }
    }
        
    def apply(input: String, types: List[Symbol]): List[List[SQLValue]] = {
      val parser = insert(parseList(types map typeParsers))
      parseAll(parser, input) match {
        case Success(result, _) => result
        case failure: NoSuccess => scala.sys.error(failure.msg + " at:\n" + failure.next.pos.longString)
      }
    }

    val quote = "'".r
    val singleParsers = typeParsers mapValues (value => {
      val quoted = quote ~ value ~ quote ^^ { case _ ~ v ~ _ => v }
      value | quoted | Null
    })
    def single(input: String, typ: Symbol): SQLValue = {
      parseAll(singleParsers(typ), input) match {
        case Success(result, _) => result
        case failure: NoSuccess => scala.sys.error(failure.msg + " at:\n" + failure.next.pos.longString)
      }
    }
  }
  
  var contexts: mutable.SortedMap[Int, String] = mutable.SortedMap.empty
  var tables: List[Table] = Nil
  var tableMap: Map[Symbol, Table] = Map.empty
  def apply(key: Symbol): Option[Table] = tableMap.get(key)
  
  class Table(val name: String, val prefix: StringBuilder, val ignored: Boolean) {
    val output: Output = new Output(s"`$name`", opt, prefix)
    var columns: List[Symbol] = Nil
    var types: List[Symbol] = Nil
    var hasCID = false
    type ContextData = List[Map[Symbol, SQLValue]]
    val data: mutable.Map[Int, ContextData] = mutable.Map.empty.withDefaultValue(Nil)
    val defaults: mutable.Map[Symbol, SQLValue] = mutable.Map.empty
    def addColumn(col: String, typ: String) {
      if (!ignored && Symbol(col) == 'cid) hasCID = true
      columns = Symbol(col) :: columns
      types = Symbol(typ) :: types
    }
    def addColumn(col: String, typ: String, default: String) {
      addColumn(col, typ)
      val value: SQLValue = if (default == "CURRENT_TIMESTAMP") opt.now else
        parseValues.single(default, Symbol(typ))
      defaults.put(Symbol(col), value)
    }
    def initRows(values: String, filter: Option[Map[Symbol, SQLValue] => Boolean]) {
      val rows = parseValues(values, types)
      for (row <- rows) {
        val asMap = (columns zip row).toMap
        if (filter.isEmpty || filter.get(asMap)) {
          val cid = if (hasCID) {
            val cid = asMap('cid).asInstanceOf[SQLInt].value.toInt
            if (opt.cid == -1) opt.cid = cid
            cid
          } else {
            0
          }
          data.update(cid, asMap :: data(cid))
        }
      }
    }
    def initColumns() {
      columns = columns.reverse
      types = types.reverse
    }
    def init() {
      types = Nil
      data foreach { case (cid, d) => data(cid) = d.reverse }
      if (data.nonEmpty) defaults ++= data(if (hasCID) opt.cid else 0).head
    }
    tables = this :: tables
    if (!ignored) tableMap += (Symbol(name) -> this)
    
    var tmpFile: Option[FileChannel] = None
    def addRow(row: Map[Symbol, SQLValue]) {
      def getValue(col: Symbol): SQLValue = {
        row.get(col) orElse defaults.get(col) getOrElse NULL
      }
      output insert (columns map getValue mkString ",")
    }
    
    def addOriginalData(cid: Int) {
      for (row <- data(cid)) addRow(row)
    }
    
    var explicit = false
    
    private class SeqCol(col: Symbol, start: Long) {
      val column = col
      var value = start
    }
    private var seq: Option[SeqCol] = None

    def sequence(col: Symbol) {
      if (!defaults.contains(col)) {
        throw new Exception(s"Sequence table `${name}` does not contains column ${col}")
      }
      seq = Some(new SeqCol(col, defaults(col).asInstanceOf[SQLInt].value))
    }
    def next(): Long = {
      seq match {
        case None => throw new Exception(s"No sequence on table `$name`")
        case Some(s) => s.value = s.value + 1; s.value
      }
    }
    
    def finish(cid: Int) {
      seq match {
        case Some(s) => addRow(Map('cid -> cid, s.column -> s.value))
        case None => if (!explicit && hasCID) addOriginalData(cid)
      }
    }
    def writeTo(dest: WritableByteChannel) {
      if (seq.isEmpty && !explicit && !hasCID) {
        for (row <- data(0)) addRow(row)
      }
      output writeTo dest
    }
  }

  var lastSegment = ""
  
  def addContext(row: Map[Symbol, SQLValue]): Boolean = {
    val cid = row('cid).asInstanceOf[SQLInt].value
    val mail = row('mail).asInstanceOf[SQLString].value
    contexts(cid.toInt) = mail.replaceFirst(".*@", "")
    true
  }

  def parse(ignoredTables: Set[String]) {
    var output = new StringBuilder
    var table: Table = null
  
    var parsingColumns = false
    for (line <- input) {
      var add = true
      if (parsingColumns) {
        line match {
          case column(name, typ, null) => table.addColumn(name, typ)
          case column(name, typ, default) => table.addColumn(name, typ, default)
          case _ => {
            table.initColumns()
            parsingColumns = false
          }
        }
      } else {
        line match {
          case create(name) => {
            table = new Table(name, output, ignoredTables(name))
            parsingColumns = true
          }
          case insert(_, values) => {
            add = false
            table.initRows(values, if (table.name == "user") Some(addContext) else None)
          }
          case end(_) => {
            table.init()
            output = new StringBuilder
          }
          case _ => {}
        }
      }
      if (add) output ++= line + '\n'
    }
    tables = tables.reverse
    lastSegment = output.result
  }
  
  def finishTables(cid: Int) = for (t <- tables) t.finish(cid)
  
  def writeTo(output: String) {
    val dest = Files.newByteChannel(Paths.get(output),
                                    WRITE, CREATE, TRUNCATE_EXISTING)
    val header = s"-- Seed: ${opt.seed}\n-- Time: ${opt.now}\n"
    dest.write(ByteBuffer.wrap(header.getBytes(UTF_8)))
    for (t <- tables) t.writeTo(dest)
    dest.write(ByteBuffer.wrap(lastSegment.getBytes(UTF_8)))
    dest.close()
  }
}
