package com.openexchange.datagenerator.sql

import java.nio.file.Files
import java.nio.file.Path
import java.nio.file.Paths
import java.nio.file.StandardOpenOption._

import com.openexchange.datagenerator.Configuration
import com.openexchange.datagenerator.Generator
import com.openexchange.datagenerator.FileWriter

import scala.collection.mutable

/**
 * @author Viktor Pracht <viktor.pracht@open-xchange.com>
 */
abstract class SQLGenerator(config: Configuration, sql: SQL) extends Generator(config) {
  sql.parse(ignoredTables)
  
  var cid: Int = 0
  
  implicit class TableAPI(name: Symbol) {
    private val table = sql(name) getOrElse {
      throw new Exception(s"Missing table `${name}`")
    }
    table.explicit = true

    def values(row: (Symbol, SQLValue)*): TableAPI = addRow(row.toMap)
    def addRow(row: Map[Symbol, SQLValue]): TableAPI = {
      table.addRow(row + ('cid -> SQLInt(cid)))
      this
    }
    
    def originalData(): TableAPI = { table.addOriginalData(cid); this }
    def data: List[Map[Symbol, SQLValue]] = table.data(cid)
    var seqColumn: Option[Symbol] = None
    def sequence(col: Symbol): TableAPI = {
      table.sequence(col)
      this
    }
    def next() = table.next()
    def defaults(values: (Symbol, SQLValue)*): TableAPI = {
      table.defaults ++= values
      this
    }
  }
  
  def exists(table: Symbol) = sql(table).isDefined
  
  object insert { def into = TableAPI _ }
  
  def binUUID(): SQLBinary = {
    val data = new Array[Byte](16)
    rnd.nextBytes(data)
    data(6) = (data(6) & 0x0f | 0x40).toByte
    data(8) = (data(8) & 0x3f | 0x80).toByte
    data.toSeq
  }

  def hexBlob(hex: String): SQLBinary = {
    hex.sliding(2, 2).map(Integer.parseInt(_, 16).toByte).toSeq
  }

  def ignoredTables: Set[String]

  case class LoginData(cid: Int, id: Int, oxLogin: String, email: String, imapLogin: String)

  def context(cid: Int, domain: String): TraversableOnce[LoginData]
  
  override def generate() {
    if (sql.opt.logins.isEmpty) {
      for ((cid, domain) <- sql.contexts) {
        this.cid = cid
        context(cid, domain)
        sql.finishTables(cid)
      }
    } else {
      val logins = new Output("", sql.opt)
      val domainSQL = new Output("virtual_domains (id, name)", sql.opt)
      val userSQL = new Output("virtual_users (domain_id, email, password)", sql.opt)
      val domains = mutable.Map.empty[String, Int]
      var domID = 0
      for ((cid, domain) <- sql.contexts) {
        this.cid = cid
        for (login <- context(cid, domain)) {
          logins append s"${login.cid},${login.id},${login.oxLogin},"
          logins append s"${login.email},${login.imapLogin}\n"
          val domainID = domains.getOrElse(domain, {
            domID += 1
            domains.put(domain, domID)
            domainSQL insert s"$domID, '$domain'"
            domID
          })
          userSQL insert s"${domainID}, '${login.imapLogin}', '{PLAIN}secret'"
        }
        sql.finishTables(cid)
      }
      val dest = Files.newByteChannel(Paths.get(sql.opt.logins.get),
                                      WRITE, CREATE, TRUNCATE_EXISTING)
      try logins writeTo dest  finally dest.close()
      val sqlDest = Files.newByteChannel(Paths.get(sql.opt.logins.get + "-sql"),
                                         WRITE, CREATE, TRUNCATE_EXISTING)
      try {
        domainSQL writeTo sqlDest
        userSQL writeTo sqlDest
      } finally sqlDest.close()
    }
  }

  private val permissionBits: Map[String, Int] = Map(
    "webmail" -> (1 << 0),
    "calendar" -> (1 << 1),
    "contacts" -> (1 << 2),
    "infostore" -> (1 << 4),
    "tasks" -> (1 << 3),
    "webdav" -> (1 << 9),
    "webdavxml" -> (1 << 8),
    "ical" -> (1 << 10),
    "vcard" -> (1 << 11),
    "syncml" -> (1 << 14),
    "usm" -> (1 << 26),
    "olox20" -> (1 << 27),
    "activesync" -> (1 << 25),
    "readcreatesharedfolders" -> (1 << 16),
    "delegatetask" -> (1 << 17),
    "editpublicfolders" -> (1 << 15),
    "editgroup" -> (1 << 18),
    "editresource" -> (1 << 19),
    "editpassword" -> (1 << 20),
    "collectemailaddresses" -> (1 << 21),
    "multiplemailaccounts" -> (1 << 22),
    "subscription" -> (1 << 23),
    "publication" -> (1 << 24),
    "deniedportal" -> (1 << 28)
  )

  class Permissions(accessCombination:TraversableOnce[String]) {
    val (bits, disableGAB) = {
      var disableGAB = false
      (accessCombination.foldLeft(0)((bits, p) => {
        disableGAB |= p == "globaladdressbookdisabled"
        bits | permissionBits.getOrElse(p, 0)
      }), disableGAB)
    }
  }

  val fileWriter = new FileWriter(sql.opt.swap)
  def drivePath(names: String*): Path = Paths.get(sql.opt.drive, names:_*)
/*
  def mailPath(user: String, names: String*): Path =
    Paths.get(sql.opt.maildir.replace("%u", user), names:_*)

  def writeDriveFile(name: String, size: Long, data: Array[Byte] => Unit) {
    val path = Paths.get(sql.opt.drive, name)
    Files.createDirectories(path.getParent)
    val fc = FileChannel.open(path, CREATE_NEW, WRITE)
    try {
      val blockSize = sql.opt.swap
      val buf = ByteBuffer.allocate(blockSize)
      val arr = new Array[Byte](blockSize)
      for (block <- 0 until (size / blockSize).toInt) {
        buf.clear()
        data(arr)
        buf.put(arr).flip()
        fc.write(buf)
      }
      val lastArr = new Array[Byte]((size % blockSize).toInt)
      buf.clear()
      data(lastArr)
      buf.put(lastArr).flip()
      fc.write(buf)
    } finally {
      fc.close()
    }
  }
*/
}
