package com.openexchange.datagenerator.sql

import java.time._
import java.time.format.DateTimeFormatter
import java.time.ZoneOffset.UTC
import java.time.temporal.ChronoField._
import java.time.temporal.ChronoUnit._

import scala.annotation.tailrec
import scala.collection.mutable

import com.openexchange.datagenerator.Configuration

/**
 * @author Viktor Pracht <viktor.pracht@open-xchange.com>
 */
class Tables(config: Configuration, sql: SQL) extends SQLGenerator(config, sql) {
  override def ignoredTables = Set("presenceSubscriptions", "updateTask")

  override def str(len: Int): String = text().take(len)

  val permissions = new Permissions(config list "permissions")

  val units = Array(DAYS, WEEKS, MONTHS, YEARS, MONTHS, YEARS)
  val freqs = Array("DAILY", "WEEKLY", "MONTHLY", "YEARLY", "MONTHLY", "YEARLY")

  def powerset[A](list: List[A]): List[List[A]] = list match {
    case Nil => List(Nil)
    case x :: xs => {
      val subset = powerset(xs)
      subset ::: (subset map (x :: _))
    }
  }
  val iCalDays = powerset(List("SO", "MO", "TU", "WE", "TH", "FR", "SA").reverse)
    .map(_.reverse.mkString(",")).toIndexedSeq

  def iCalDuration(minutes: Int): String = {
    if (minutes == 0) return "PT0S"
    val sign = if (minutes < 0) "-" else ""
    val mins = Math.abs(minutes)
    if (mins >= 7*24*60) s"${sign}P${mins / (7*24*60)}W"  // weeks
    else if (mins >= 24*60) s"${sign}P${mins / (24*60)}D" // days
    else if (mins >= 60) s"${sign}PT${mins / 60}H"        // hours
    else s"${sign}PT${mins}M"                             // minutes
  }

  val iCalDate = DateTimeFormatter ofPattern "yyyyMMdd"
  val iCalDateTime = DateTimeFormatter ofPattern "yyyyMMdd'T'HHmmss'Z'"
  val iCalDate2 = DateTimeFormatter ofPattern "yyyy-MM-dd"
  val iCalDateTime2 = DateTimeFormatter ofPattern "yyyy-MM-dd HH:mm:ss"

  override def context(cid: Int, domain: String): TraversableOnce[LoginData] = {
    'sequence_calendar  sequence 'id
    'sequence_contact   sequence 'id
    'sequence_folder    sequence 'id
    'sequence_infostore sequence 'id
    'sequence_principal sequence 'id
    'sequence_reminder  sequence 'id
    if (exists('calendar_event_sequence)) {
      'calendar_alarm_sequence defaults ('id -> 0, 'account -> 0) sequence 'id
      'calendar_event_sequence defaults ('id -> 0, 'account -> 0) sequence 'id
    }
    
    insert into 'groups_member          originalData()
    insert into 'login2user             originalData()
    insert into 'oxfolder_permissions   originalData()
    insert into 'oxfolder_tree          originalData()
    insert into 'prg_contacts           originalData()
    insert into 'user                   originalData()
    insert into 'user_attribute         originalData()
    insert into 'user_configuration     originalData()
    insert into 'user_mail_account      originalData()
    insert into 'user_setting_mail      originalData()
    insert into 'user_transport_account originalData()
    insert into 'user_transport_account_properties originalData()

    var filestore_usage = 0L

    // Extract the list of user-specific default folders

    class DefaultFolder(row: Map[Symbol, SQLValue]) {
      val name = SQLString primitiveFrom row('fname)
      val module = SQLInt primitiveFrom row('module)
      val fType = SQLInt primitiveFrom row('type)
      val flag = SQLInt primitiveFrom row('permission_flag)

      var id: Long = SQLInt primitiveFrom row('fuid)
      def next() { id = 'sequence_folder.next() }

      val originalParentID = SQLInt primitiveFrom row('parent)
      var parent: Option[DefaultFolder] = None
      def parentID = parent.map(_.id).getOrElse(originalParentID)
    }

    val defaultFolders = 'oxfolder_tree.data
      .filter(row => (SQLInt primitiveFrom row('type)) != 5)
      .map(row => new DefaultFolder(row))

    // Resolve the parent methods to return each other's current ID
    val folderMap = defaultFolders.map(folder => (folder.id, folder)).toMap
    for (folder <- defaultFolders) {
      folderMap.get(folder.originalParentID)
        .foreach(parent => folder.parent = Some(parent))
    }

    // Some default folders are referenced directly
    def findFolder(module: Int, name: String) = defaultFolders
      .find(_.module == module)
      .getOrElse(throw new Exception(s"Default ${name} folder not found"))
    val calendarFolder = findFolder(2, "calendar")
    val contactsFolder = findFolder(3, "contacts")
    val infostoreFolder = findFolder(8, "drive")

    // Users
    val loginPattern = config str "login"
    val emailPattern = config str "email"
    val oxLoginPattern = config str "oxLogin"
    val imapLoginPattern = config str "imapLogin"
    val logins: TraversableOnce[LoginData] = config("usersPerContext") generate ((user_index: Int) => {
      val user_id = 'sequence_principal.next()
      val contact_id = 'sequence_contact.next()
      
      val login = loginPattern.replace("#id", user_id.toString)
        .replace("#cid", cid.toString)
      val firstName = name().mkString
      val lastName = name().mkString
      val displayName = s"$firstName $lastName"
      def evalPattern(s: String): String = s
        .replace("#id", user_id.toString)
        .replace("#cid", cid.toString)
        .replace("#login", login)
        .replace("#domain", domain)
      val email = evalPattern(emailPattern)
      val imapLogin = evalPattern(imapLoginPattern)
      val oxLogin = evalPattern(oxLoginPattern)
      val timezone = "Europe/Berlin"
      val zoneId = ZoneId.of(timezone)
      
      val user_uuid = uuid()
      val organizerURI = s"urn:uuid:${f"$cid%08x"}-${f"${user_id >> 16}%04x"}-${f"${user_id & 0xffff}%04x"}-00fc-c0e11e000001"
      
      insert into 'groups_member values (
          'id -> 1,
          'member -> user_id
      )
      insert into 'login2user values ( 
          'id -> user_id,
          'uid -> login
      )
      insert into 'user values (
          'id -> user_id,
          'imapLogin -> imapLogin,
          'mail -> email,
          'timeZone -> timezone,
          'contactId -> contact_id,
          'homeDirectory -> s"/home/${login}.${cid}"
      )
      if (exists('user_alias)) {
        insert into 'user_alias values (
            'user -> user_id,
            'alias -> email,
            'uuid -> binUUID()
        )
      } else {
        insert into 'user_attribute values (
            'id -> user_id,
            'name -> "alias",
            'value -> email,
            'uuid -> binUUID()
        )
      }
      insert into 'user_configuration values (
          'user -> user_id,
          'permissions -> permissions.bits // 268422943
      )
      insert into 'user_mail_account values (
          'user -> user_id,
          'login -> imapLogin,
          'primary_addr -> email,
          'trash -> "Trash",
          'sent -> "Sent",
          'drafts -> "Drafts",
          'spam -> "Spam",
          'trash_fullname -> "Trash",
          'sent_fullname -> "Sent",
          'drafts_fullname -> "Drafts",
          'spam_fullname -> "Spam",
      )
      insert into 'user_setting_mail values (
          'user -> user_id,
          'send_addr -> email
      )

      insert into 'user_setting_server values (
        'user -> user_id,
        'contact_collect_folder -> 0,
        'uuid -> binUUID()
      )

      insert into 'user_transport_account values (
          'user -> user_id,
          'send_addr -> email
      )
      insert into 'user_transport_account_properties values (
        'user -> user_id
      )
      
      // Folders

      insert into 'oxfolder_permissions values(
        'fuid -> 6, 'permission_id -> user_id,
        'fp -> (if (permissions.disableGAB) 0 else 2),
        'orp -> (if (permissions.disableGAB) 0 else 4),
        'owp -> (if (permissions.disableGAB) 0 else 2),
        'odp -> 0, 'admin_flag -> 0, 'group_flag -> 0, 'system -> 0
      )
      for (folder <- defaultFolders) {
        folder.next()
        insert into 'oxfolder_permissions values (
            'fuid -> folder.id,
            'permission_id -> user_id,
            'fp -> 128, 'orp -> 128, 'owp -> 128, 'odp -> 128,
            'admin_flag -> 1, 'group_flag -> 0, 'system -> 0
        )
        insert into 'oxfolder_tree values(
            'fuid -> folder.id,
            'parent -> folder.parentID,
            'fname -> (if (folder.module == 8 && folder.fType == 2) displayName
                       else folder.name),
            'module -> folder.module,
            'type -> folder.fType,
            'creating_date -> now,
            'created_from -> user_id,
            'changing_date -> now,
            'changed_from -> user_id,
            'permission_flag -> folder.flag,
            'subfolder_flag -> 0,
            'default_flag -> 1,
            'meta -> NULL
        )
      }
      
      // Contacts
      
      insert into 'prg_contacts values (
          'creating_date -> now,
          'created_from -> user_id,
          'changing_date -> now,
          'changed_from -> user_id,
          'fid -> 6,
          'userid -> user_id,
          'intfield01 -> contact_id,
          'field01 -> displayName,
          'field02 -> lastName,
          'field03 -> firstName,
          'field65 -> email,
          'field90 -> displayName,
          'useCount -> 0,
          'uid -> user_uuid
      )
      config("contactsPerUser") generate {
        val firstName = str(128)
        val lastName = str(128)
        val birthday = today minusDays gauss(13*365, 100*365)
        val anniversary = today minusDays uniform(0, 365)
        insert into 'prg_contacts values (
            'creating_date -> now, 'created_from -> user_id,
            'changing_date -> now, 'changed_from -> user_id,
            'fid -> contactsFolder.id,
            'userid -> NULL,
            'pflag -> (if (config bool "privateContacts") 1 else NULL),
            'timestampfield01 -> birthday,
            'timestampfield02 -> anniversary,
            'intfield01 -> 'sequence_contact.next(),
            'intfield08 -> 1,
            'field01 -> (firstName + " " + lastName),
            'field02 -> lastName, 'field03 -> firstName,
            'field04 -> str(128),  'field05 -> str(64),  'field06 -> str(64),
            'field07 -> str(256),  'field08 -> str(64),  'field09 -> str(64),
            'field10 -> str(64),   'field11 -> str(64),  'field14 -> str(64),
            'field17 -> str(5680), 'field18 -> str(512), 'field19 -> str(128),
            'field20 -> str(128),  'field22 -> str(64),  'field23 -> str(256),
            'field24 -> str(64),   'field25 -> str(128), 'field26 -> str(64),
            'field27 -> str(64),   'field37 -> str(256), 'field38 -> str(64),
            'field39 -> str(64),   'field40 -> str(64),  'field41 -> str(64),
            'field43 -> str(64),   'field44 -> str(128), 'field45 -> str(64),
            'field49 -> str(64),   'field50 -> str(64),  'field51 -> str(64),
            'field53 -> str(64),   'field54 -> str(64),  'field55 -> str(64),
            'field62 -> str(64),   'field63 -> str(64),  'field65 -> str(256),
            'field66 -> str(256),  'field67 -> str(256), 'field68 -> str(128),
            'field70 -> str(64),   'field71 -> str(64),  'field72 -> str(64),
            'field73 -> str(64),   'field74 -> str(64),  'field75 -> str(64),
            'field76 -> str(64),   'field77 -> str(64),  'field78 -> str(64),
            'field79 -> str(64),   'field80 -> str(64),  'field81 -> str(64),
            'field82 -> str(64),   'field83 -> str(64),  'field84 -> str(64),
            'field85 -> str(64),   'field86 -> str(64),  'field87 -> str(64),
            'field88 -> str(64),   'field89 -> str(64),
            'field90 -> displayName,
            'uid -> uuid
        )
      }
      
      // Appointments
      
      config("appointmentsPerUser") generate {
        
        // Participant types
        val USER = 1
        val GROUP = 2
        val RESOURCE = 3
        val RESOURCEGROUP = 4
        val EXTERNAL_USER = 5
        val EXTERNAL_GROUP = 6
        val NO_ID = -1
        
        // Shown as
        val RESERVED = 1
        val TEMPORARY = 2
        val ABSENT = 3
        val FREE = 4
        
        // Confirmation status
        val NONE = 0
        val ACCEPT = 1
        val DECLINE = 2
        val TENTATIVE = 3
        
        // Module numbers
        val APPOINTMENT = 1
        val TASK = 4
        val CONTACT = 7
        val EMAIL = 19
        val INFOSTORE = 137

        val hasOld = exists('prg_dates)
        val hasNew = exists('calendar_event)
        val id = 'sequence_calendar.next()
        val eventID = if (hasNew) Some('calendar_event_sequence.next()) else None
        val uid = uuid()
        val fulltime = config bool "fulltimeAppointments"
        val start = today.plusDays(gauss(0, 100))
          .atTime(if (fulltime) 0 else uniform(8, 12), 0)
        val end = if (fulltime) {
          start plusDays truncate(1, 1000, gauss(1, 4))
        } else {
          start plusHours truncate(1, 100, gauss(1, 4))
        }
        
        val color = if (config bool "colorLabels") uniform(1, 10) else 0
        
        val recurring = config bool "recurringAppointments"

        def asICal(t: LocalDateTime): String = t.format(if (fulltime) iCalDate else iCalDateTime)
        def asICal2(t: LocalDateTime): String = t.format(if (fulltime) iCalDate2 else iCalDateTime2)

        val summary = str(255)
        val location = str(255)
        val description = str(1024)

        val appointment = mutable.Map[Symbol, SQLValue](
          'creating_date -> now,
          'created_from -> user_id,
          'changing_date -> now,
          'changed_from -> user_id,
          'fid -> 0, // private folder of every participant
          'pflag -> (if (config bool "privateAppointments") 1 else 0),
          'timestampfield01 -> start,
          'timezone -> timezone,
          'intfield01 -> id,
          'intfield03 -> color,
          'intfield05 -> NULL, // TODO: recurrence position
          'intfield06 -> (config int "shownAs"),
          'intfield07 -> (if (fulltime) 1 else 0),
          'intfield08 -> 0, // TODO: number of attachments
          'field01 -> summary,
          'field02 -> location,
          'field04 -> description,
          'field09 -> NULL, // TODO: categories (varchar,NULL,)
          'uid -> uid,
          'organizer -> email,
          'sequence -> NULL, // TODO: iCal (int,NULL,)
          'organizerId -> user_id,
          'principal -> NULL, // TODO: iCal (varchar,NULL,)
          'principalId -> NULL, // TODO: ical (int,NULL,)
          'filename -> NULL // TODO: (varchar,NULL)
        )

        val event = mutable.Map.empty[Symbol, SQLValue]
        if (hasNew) {
          config("alarmsPerAppointment") generate {
            val action = config str "alarmActions"
            val extProps = if (action == "DISPLAY") {
              // 0x05 followed by DEFLATE-compressed JSON: [{"name":"DESCRIPTION","value":""}]
              hexBlob("05789C8BAE56CA4BCC4D55B25272710D760EF20C08F1F4F753D2512A4BCC2905892AD5C60200BE360A63")
            } else {
              NULL
            }
            val alarmDuration = config int "alarmDurations"
            val alarmID = 'calendar_alarm_sequence.next()
            insert into 'calendar_alarm values(
              'account -> 0,
              'id -> alarmID,
              'event -> eventID.get,
              'user -> user_id,
              'uid -> uuid,
              'action -> action,
              'triggerRelated -> "START",
              'triggerDuration -> iCalDuration(alarmDuration),
              'extendedProperties -> extProps
            )
            // TODO: alarms for birthdays
            val trigger: Instant = start.atZone(zoneId).toInstant
            insert into 'calendar_alarm_trigger values(
              'account -> 0,
              'alarm -> alarmID,
              'user -> user_id,
              'eventId -> eventID.get,
              'folder -> calendarFolder.id,
              'triggerDate -> trigger.plus(alarmDuration, MINUTES),
              'action -> action,
              'recurrence -> asICal(start),
              'floatingTimezone -> (if (fulltime) timezone else NULL),
              'relatedTime -> trigger
            )
          }
          insert into 'calendar_attendee values (
            'account -> 0,
            'event -> eventID.get,
            'entity -> user_id,
            'uri -> s"mailto:${email}",
            'cn -> displayName,
            'folder -> calendarFolder.id,
            'cuType -> "INDIVIDUAL",
            'partStat -> "ACCEPTED"
          )
          event ++= List[(Symbol, SQLValue)](
            'account-> 0,
            'id -> eventID.get,
            'user -> user_id,
            'uid -> uid,
            'timestamp -> now.atZone(zoneId).toInstant,
            'created -> now.atZone(zoneId).toInstant,
            'createdBy -> user_id,
            'modified -> now.atZone(zoneId).toInstant,
            'modifiedBy -> user_id,
            'start -> asICal2(start),
            'startTimezone -> (if (fulltime) NULL else timezone),
            'end -> asICal2(end),
            'endTimezone -> (if (fulltime) NULL else timezone),
            'allDay -> (if (fulltime) 1 else 0),
            'sequence -> 0,
            'transp -> 1,
            'class -> "PUBLIC",
            'organizer -> organizerURI,
            'summary -> summary,
            'location -> location,
            'description -> description,
            'rangeFrom -> start.atZone(zoneId).toInstant
          )
        }

        if (hasOld) insert into 'prg_date_rights values (
            'object_id -> id,
            'id -> user_id,
            'type -> USER
        )
        
        if (recurring) {
          val rt = config int "recurrenceTypes"
          val unit = units(rt - 1)
          val interval = config int ("recurrenceInterval" + rt)
          val infinite = config bool ("infiniteRecurrence" + rt)
          
          val recurrence = new StringBuilder
          recurrence.append(s"t|${rt}|i|${interval}|")

          def opt(types: Set[Int], param: => Int): Option[Int] = {
            if (types contains rt) Some(param) else None
          }
          val weekdays = opt(Set(2, 5, 6), config int ("weekdays" + rt))
          val month = opt(Set(4, 6), config int ("month" + rt))
          val dayMonth = if (rt == 4) start.withMonth(month.get) else start
          val dayLimit = YearMonth.from(dayMonth).lengthOfMonth
          val day = opt(Set(3, 4, 5, 6),
                        truncate(1, dayLimit, config int ("day" + rt)))
          
          if (weekdays.isDefined) recurrence.append(s"a|${weekdays.get}|")
          if (day.isDefined) recurrence.append(s"b|${day.get}|")
          if (month.isDefined) recurrence.append(s"c|${month.get}|")
          recurrence.append(s"s|${start.toEpochSecond(UTC)}000|")
          
          val (recurrenceEnd, iCalEnd) = if (infinite) {
            (end plusYears 20, None)
          } else {
            val counted = config bool ("countedRecurrence" + rt)
            val (recEnd, icalEnd) = if (counted) {
              val count = config int ("recurrenceCount" + rt)
              @tailrec
              def iterate[T](f: T => T, x: T, n: Int): T = {
                if (n == 0) x else iterate(f, f(x), n - 1)
              }
              def getEnd(f: LocalDateTime => LocalDateTime): LocalDateTime = {
                val skipStart = if (f(start).isBefore(start)) 1 else 0
                f(start.plus(interval * (count + skipStart), unit))
              }
              def findDayOfMonth(t: LocalDateTime): LocalDateTime = {
                val days = weekdays.get
                val next = Week.adjuster(Week.nextDay, days)
                val end = iterate((t: LocalDateTime) => t.`with`(next),
                                  t.withDayOfMonth(1).minusDays(1),
                                  day.get)
                val nextMonth = t.`with`(Week.firstDayOfNextMonth)
                if (end.isBefore(nextMonth)) {
                  end
                } else {
                  nextMonth.`with`(Week.adjuster(Week.previousDay, days))
                }
              }
              val recEnd: LocalDateTime = rt match {
                case 1 => start.plus(interval * count, unit)
                case 2 => {
                  val perWeek = Week.Count(weekdays.get)
                  val days = weekdays.get
                  iterate((t: LocalDateTime) =>
                      t.plus(Week.nextDay(days, t.get(DAY_OF_WEEK)), DAYS),
                    start.plus(interval * (count / perWeek), unit),
                    count % perWeek)
                }
                case 3 => {
                  val d = day.get
                  if (d < 29) {
                    getEnd(_.withDayOfMonth(day.get))
                  } else {
                    @tailrec
                    def next(t: LocalDateTime): LocalDateTime = {
                      if (d > YearMonth.from(t).lengthOfMonth) {
                        next(t.plusMonths(interval))
                      } else {
                        t.withDayOfMonth(d)
                      }
                    }
                    val skipStart = if (next(start).isBefore(start)) 1 else 0
                    iterate(next, start, count + skipStart)
                  }
                }
                case 4 => getEnd(_.withMonth(month.get).withDayOfMonth(day.get))
                case 5 => getEnd(findDayOfMonth)
                case 6 => getEnd(t => findDayOfMonth(t.withMonth(month.get)))
              }
              recurrence.append(s"e|${recEnd.toEpochSecond(UTC)}000|")
              recurrence.append(s"o|${count}|")
              (recEnd, s";COUNT=${count}")
            } else {
              val recEnd = start.plus(config int ("recurrenceDuration" + rt), unit)
              recurrence.append(s"e|${recEnd.toEpochSecond(UTC)}000|")
              (recEnd, s";UNTIL=${asICal(recEnd)}")
            }
            (recEnd, Some(icalEnd))
          }
          
          appointment ++= List[(Symbol, SQLValue)](
              'timestampfield02 -> recurrenceEnd,
              'intfield02 -> id, // recurrence ID
              'intfield04 -> DAYS.between(start, end),
              'field06 -> recurrence.toString(),
              'field07 -> NULL, // TODO: delete exceptinos (text,NULL,)
              'field08 -> NULL // TODO: change exceptions (text,NULL,)
          )

          if (hasNew) {
            val rrule = new StringBuilder
            rrule append s"FREQ=${freqs(rt - 1)}"
            if (weekdays.isDefined) rrule append s";BYDAY=${iCalDays(weekdays.get)}"
            if (day.isDefined) rrule append s";BYMONTHDAY=${day.get}"
            if (month.isDefined) rrule append s";BYMONTH=${month.get}"
            rrule append s";INTERVAL=${interval}"
            if (iCalEnd.isDefined) rrule append iCalEnd.get

            event ++= List[(Symbol, SQLValue)](
              'series -> eventID.get,
              'rrule -> rrule.toString,
              'rangeUntil -> (if (infinite) Long.MaxValue else recurrenceEnd.atZone(zoneId).toInstant)
            )
          }
        } else {
          appointment += 'timestampfield02 -> (end: SQLValue)
          if (hasNew) {
            event += ('rangeUntil -> end.atZone(zoneId).toInstant)
          }
        }
        
        if (hasOld) 'prg_dates addRow appointment.toMap
        if (hasNew) 'calendar_event addRow event.toMap

        val alarm = 15
        if (hasOld) insert into 'prg_dates_members values (
          'object_id -> id,
          'member_uid -> user_id,
          'confirm -> uniform(0, 4),
          'reason -> (if (probability(0.1)) str(50) else NULL),
          'pfid -> calendarFolder.id,
          'reminder -> alarm
        )

        val reminder_id = 'sequence_reminder.next()
        insert into 'reminder values (
          'object_id -> reminder_id,
          'last_modified -> timestamp,
          'target_id -> id.toString,
          'module -> APPOINTMENT,
          'userid -> user_id,
          'alarm -> (start minusMinutes alarm),
          'recurrence -> 0, // TODO: recurrence id
          'description -> NULL, // varchar(1028),
          'folder -> calendarFolder.id.toString
        )
      }
      
      // Drive
      
      def newFilename(): String = {
        val name = binUUID().bytes
        val hash = name.hashCode().toHexString.padTo(6, '0').grouped(2) // yes, right-padding
        s"hashed/${hash.next}/${hash.next}/${hash.next}/${name}"
      }

      config("filesPerUser") generate {
        
        val id = 'sequence_infostore.next()
        val versions = config int "versionsPerFile"
        
        insert into 'infostore values (
          'id -> id,
          'folder_id -> infostoreFolder.id, // TODO: multiple folders
          'version -> versions,
          'locked_until -> NULL,
          'color_label -> 0,
          'creating_date -> now,
          'last_modified -> now,
          'created_by -> user_id,
          'changed_by -> user_id
        )

        // version 0 has no file
        insert into 'infostore_document values (
          'infostore_id -> id,
          'version_number -> 0,
          'creating_date -> now,
          'last_modified -> now,
          'created_by -> user_id,
          'changed_by -> user_id,
          'title -> str(255),
          'url -> (if (config bool "driveURLs") str(256) else NULL),
          'description -> str(1024),
          'categories -> str(255),
          'filename -> NULL,
          'file_store_location -> NULL,
          'file_size -> 0,
          'file_mimetype -> "application/octet-stream",
          'file_md5sum -> NULL,
          'file_version_comment -> NULL,
          'meta -> NULL
        )

        for (v <- 1 to versions) {
          val name = newFilename()
          val size = (config num "fileSize").toLong
          filestore_usage += size
          val comment: SQLValue = if (config bool "versionComment") str(1024)
                                  else NULL
          
          val meta = fileWriter.random(drivePath(s"${cid}_ctx_store/${name}"),
                                       size, rnd)
          
          insert into 'infostore_document values (
            'infostore_id -> id,
            'version_number -> v,
            'creating_date -> now,
            'last_modified -> now,
            'created_by -> user_id,
            'changed_by -> user_id,
            'title -> str(255),
            'url -> (if (config bool "driveURLs") str(256) else NULL),
            'description -> str(1024),
            'categories -> str(255),
            'filename -> str(255),
            'file_store_location -> name,
            'file_size -> size,
            'file_mimetype -> (config str "fileType"),
            'file_md5sum -> meta.md5.map("%02x" format _).mkString,
            'file_version_comment -> comment,
            'meta -> NULL // TODO: blob,
          )
        }
      }
      
      LoginData(cid, user_id.toInt, oxLogin, email, imapLogin)
    })
    
    insert into 'filestore_usage values ('used -> filestore_usage)
    
    logins
  }
}
