package com.openexchange.datagenerator

import java.time._

package object sql {
  trait SQLValue { def toString(): String }
  implicit class SQLInt(val value: Long) extends SQLValue {
    override def toString() = value.toString
  }
  implicit class SQLFloat(val value: Double) extends SQLValue {
    override def toString() = value.toString
  }
  implicit class SQLString(val value: String) extends SQLValue {
    override def toString() = s"'${Literals.escape(value)}'"
  }
  implicit class SQLTimestamp(val value: Instant) extends SQLValue {
    override def toString() = value.toEpochMilli.toString
  }
  implicit class SQLDate(val value: LocalDate) extends SQLValue {
    override def toString() = s"'${value}'"
  }
  implicit class SQLTime(val value: LocalTime) extends SQLValue {
    override def toString() = s"'${value}'"
  }
  implicit class SQLDateTime(val value: LocalDateTime) extends SQLValue {
    override def toString() = s"'${value}'"
  }
  implicit class SQLBinary(val value: Seq[Byte]) extends SQLValue {
    def bytes: String = value.map("%02X" format _).mkString
    override def toString() = s"X'${bytes}'"
  }
  object NULL extends SQLValue { override def toString() = "NULL" }

  object SQLInt       { def primitiveFrom(v: SQLValue) = v.asInstanceOf[SQLInt].value }
  object SQLFloat     { def primitiveFrom(v: SQLValue) = v.asInstanceOf[SQLFloat].value }
  object SQLString    { def primitiveFrom(v: SQLValue) = v.asInstanceOf[SQLString].value }
  object SQLTimestamp { def primitiveFrom(v: SQLValue) = v.asInstanceOf[SQLTimestamp].value }
  object SQLDate      { def primitiveFrom(v: SQLValue) = v.asInstanceOf[SQLDate].value }
  object SQLTime      { def primitiveFrom(v: SQLValue) = v.asInstanceOf[SQLTime].value }
  object SQLDateTime  { def primitiveFrom(v: SQLValue) = v.asInstanceOf[SQLDateTime].value }
  object SQLBinary    { def primitiveFrom(v: SQLValue) = v.asInstanceOf[SQLBinary].value }

  object Literals {
    val esc = """[\x00\x08\n\r\t\x1a'"\\]""".r
    val escapes = Map("\u0000" -> "\\\\0", "'"  -> "\\\\'", "\"" -> "\\\\\"",
                      "\b" -> "\\\\b", "\n" -> "\\\\n", "\r" -> "\\\\r",
                      "\t" -> "\\\\t", "\u001a" -> "\\\\Z", "\\" -> "\\\\\\\\")
    val unesc = """''|\\.""".r
    val unescapes = Map("''"   -> "'",  "\\0" -> "\u0000", "\\'" -> "'",
                        "\\\"" -> "\"", "\\b" -> "\b", "\\n" -> "\n",
                        "\\r"  -> "\r", "\\t" -> "\t", "\\Z" -> "\u001a",
                        "\\\\" -> "\\\\")
    
    def escape(str: String): String = {
      esc.replaceAllIn(str, m => escapes(m.matched))
    }
    
    def unescape(str: String): String = {
      unesc.replaceAllIn(str, m => unescapes.getOrElse(m.matched, m.matched))
    }
  }
}
