/*
 * Copyright (C) 2018 OX Software GmbH
 *
 * This file is part of OX API-Performance Tests.
 *
 * OX API-Performance Tests is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * OX API-Performance Tests is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OX API-Performance Tests.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.openexchange.testing.performance.ajax.calendar

import com.openexchange.testing.performance.common.ModuleNames
import com.openexchange.testing.performance.common.utils.JSONUtil
import com.openexchange.testing.performance.groupware.{ Appointment, DetailedAppointmentData, TaskAppointmentData }
import io.gatling.commons.util.ClockSingleton
import io.gatling.core.Predef._
import io.gatling.core.session.Expression
import io.gatling.core.structure.ChainBuilder
import io.gatling.http.Predef._

/**
 * [[CalendarAPI]]
 *
 * @author <a href="mailto:nikolaos.tsapanidis@open-xchange.com">Nikolaos Tsapanidis</a>
 */
object CalendarAPI {

  /**
   * Confirms an appointment.
   *
   * @return
   */
  def confirmAppointment = ???

  /**
   * Creates an appointment.
   *
   * @param appointment Appointment object
   */
  def createAppointment(appointment: Expression[Appointment]): ChainBuilder = group(ModuleNames.CALENDAR) {
    exec(new CreateAppointment(appointment).buildHttpRequest()
      .queryParam("timezone", "UTC")
      .check(jsonPath("$.data.id").ofType[String].saveAs(classOf[CreateAppointment].getSimpleName)))

  }

  /**
   * Deletes appointments.
   *
   * @param timestamp
   * @return
   */
  def deleteAppointment(timestamp: Expression[Long]): ChainBuilder = group(ModuleNames.CALENDAR) {
    exec(new DeleteAppointment(timestamp, "")) // TODO
  }

  /**
   * Gets all appointments.
   *
   * @param columns
   * @param start
   * @param end
   * @param optionalQueryParams
   * @return
   */
  def getAllAppointments(columns: String, start: Long, end: Long, optionalQueryParams: Option[Map[String, Any]] = None): ChainBuilder = group(ModuleNames.CALENDAR) {
    exec(new GetAllAppointments(columns, start, end).buildHttpRequest()
      .queryParamMap(optionalQueryParams.getOrElse(Map.empty))
      .check(bodyString.transform(x => JSONUtil().fromJson[GetAllAppointmentsResponse](x))
        .optional.saveAs("getAllAppointmentsResponse")))
  }

  /**
   * Gets an appointment.
   *
   * @param id
   * @param folder
   * @return
   */
  def getAppointment(id: Expression[String], folder: Expression[String]): ChainBuilder = group(ModuleNames.CALENDAR) {
    exec(new GetAppointment(id, folder).buildHttpRequest()
      // TODO https://github.com/FasterXML/jackson-databind/issues/265
      .check(jsonPath("$.data").transform(x => JSONUtil().fromJson[TaskAppointmentData](x)).saveAs("TaskAppointmentData"))
      .check(jsonPath("$.data").transform(x => JSONUtil().fromJson[DetailedAppointmentData](x)).saveAs("DetailedAppointmentData")))
      .exec(session => {
        val taskAppointmentData = session("TaskAppointmentData").as[TaskAppointmentData]
        val detailedAppointmentData = session("DetailedAppointmentData").as[DetailedAppointmentData]

        val app = Appointment(taskAppointmentData, detailedAppointmentData)

        session.set("getAppointment", app)
      })
      .exec(session => {
        session.removeAll("TaskAppointmentData", "DetailedAppointmentData")
      })
  }

  /**
   * Gets a list of appointments.
   *
   * @return
   */
  def getAppointmentList = ???

  /**
   * Gets updated appointments.
   *
   * @param columns
   * @param lastUpdate
   * @param optionalQueryParams
   * @return
   */
  def getUpdatedAppointments(columns: String, lastUpdate: Expression[Long], optionalQueryParams: Option[Map[String, Any]] = None): ChainBuilder = group(ModuleNames.CALENDAR) {
    exec(new GetUpdatedAppointments(
      columns,
      lastUpdate,
      start = ClockSingleton.nowMillis,
      end = ClockSingleton.nowMillis + 5 * 365 * 86400 * 1000L,
      showPrivate = "true").buildHttpRequest()
      .queryParamMap(optionalQueryParams.getOrElse(Map.empty)))
  }

  /**
   * Gets all change exceptions.
   *
   * @return
   */
  def getChangeExceptions = ???

  /**
   * Gets appointments between a specified time range.
   *
   * @return
   */
  def getFreeAndBusy = ???

  /**
   * Gets new appointments.
   *
   * @return
   */
  def getNewAppointments = ???

  /**
   * Requests whether there are appointments on days in a specified time range.
   *
   * @param start
   * @param end
   * @return
   */
  def hasAppointmentsOnDays(start: Long, end: Long): ChainBuilder = group(ModuleNames.CALENDAR) {
    exec(new GetAppointmentInfo(start, end))
  }

  /**
   * Resolves the UID to an OX object ID.
   *
   * @return
   */
  def resolveUID = ???

  /**
   * Searches for appointments.
   *
   * @return
   */
  def searchAppointments = ???

  /**
   * Updates an appointment.
   *
   * @param id
   * @param folder
   * @param lastUpdate
   * @param appointment
   * @return
   */
  def updateAppointment(id: Expression[Int], folder: Expression[String], lastUpdate: Expression[Long], appointment: Expression[Appointment]): ChainBuilder = group(ModuleNames.CALENDAR) {
    exec(new UpdateAppointment(id, folder, lastUpdate, appointment))
  }

}
