/*
 * Copyright (C) 2018 OX Software GmbH
 *
 * This file is part of OX API-Performance Tests.
 *
 * OX API-Performance Tests is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * OX API-Performance Tests is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OX API-Performance Tests.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.openexchange.testing.performance.ajax.chronos

import com.openexchange.testing.performance.ajax.chronos.alarm.{ Ack, Pending, Snooze }
import com.openexchange.testing.performance.ajax.chronos.availability.{ GetAvailability, SetAvailability }
import com.openexchange.testing.performance.ajax.chronos.schedjoules.{ BrowseSchedJoules, CountriesSchedJoules, LanguagesSchedJoules, SearchSchedJoules }
import com.openexchange.testing.performance.common.ModuleNames
import com.openexchange.testing.performance.common.request.OptParamImplicits._
import com.openexchange.testing.performance.common.utils.JSONUtil
import com.openexchange.testing.performance.groupware.chronos._
import io.gatling.core.Predef._
import io.gatling.core.session.Expression
import io.gatling.core.structure.ChainBuilder
import io.gatling.http.Predef._
import org.dmfs.rfc5545.DateTime

import scala.language.implicitConversions

/**
 * [[ChronosAPI]]
 *
 * @author <a href="mailto:nikolaos.tsapanidis@open-xchange.com">Nikolaos Tsapanidis</a>
 */
object ChronosAPI {

  def getAllEvents(
    rangeStart: DateTime, rangeEnd: DateTime, folder: Option[String] = None,
    fields: Option[String] = None, order: Option[String] = None, sort: Option[String] = None,
    expand: Option[Boolean] = None, extendedEntities: Option[Boolean] = None,
    updateCache: Option[Boolean] = None): ChainBuilder = group(ModuleNames.CHRONOS) {
    exec(new GetAllEvents(rangeStart, rangeEnd, Option(Map(
      "folder" -> folder.getOrElse(None),
      "fields" -> fields.getOrElse(None),
      "order" -> order.getOrElse(None),
      "sort" -> sort.getOrElse(None),
      "expand" -> expand.getOrElse(None),
      "extendedEntities" -> extendedEntities.getOrElse(None),
      "updateCache" -> updateCache.getOrElse(None)))))
  }

  def getAllEventsOfMultipleFolders(
    rangeStart: DateTime, rangeEnd: DateTime, body: Expression[ChronosFolderBody], folder: Option[String] = None,
    fields: Option[String] = None, order: Option[String] = None, sort: Option[String] = None, expand: Option[Boolean] = None,
    extendedEntities: Option[Boolean] = None, updateCache: Option[Boolean] = None): ChainBuilder =
    group(ModuleNames.CHRONOS) {
      exec(new GetAllEventsOfFolders(rangeStart, rangeEnd, StringBody(
        session => JSONUtil().toJson(body(session).get)), Option(Map(
        "folder" -> folder.getOrElse(None),
        "fields" -> fields.getOrElse(None),
        "order" -> order.getOrElse(None),
        "sort" -> sort.getOrElse(None),
        "expand" -> expand.getOrElse(None),
        "extendedEntities" -> extendedEntities.getOrElse(None),
        "updateCache" -> updateCache.getOrElse(None)))).buildHttpRequest()
        .check(jsonPath("$.data[*].events[*].id").findAll.optional
          .saveAs(s"${classOf[GetAllEventsOfFolders].getSimpleName}/Ids"))
        .check(jsonPath("$.data[*].events[*]").findAll.transform(_.map(JSONUtil().fromJson[EventData](_)))
          .optional.saveAs(classOf[GetAllEventsOfFolders].getSimpleName)))
    }

  def deleteEvents(
    timestamp:        Long,
    body:             Expression[Array[EventId]],
    rangeStart:       Option[DateTime]           = None,
    rangeEnd:         Option[DateTime]           = None,
    expand:           Option[Boolean]            = None,
    fields:           Option[String]             = None,
    extendedEntities: Option[Boolean]            = None): ChainBuilder = group(ModuleNames.CHRONOS) {
    exec(new DeleteEvents(timestamp, StringBody(session => JSONUtil().toJson(body(session).get)), Option(Map(
      "rangeStart" -> rangeStart.getOrElse(None),
      "rangeEnd" -> rangeEnd.getOrElse(None),
      "expand" -> expand.getOrElse(None),
      "fields" -> fields.getOrElse(None),
      "extendedEntities" -> extendedEntities.getOrElse(None)))))
  }

  def getFreeBusy(
    from: DateTime, until: DateTime, body: Expression[FreeBusyBody], maskId: Option[String] = None,
    merge: Option[Boolean] = None): ChainBuilder = group(ModuleNames.CHRONOS) {
    exec(new FreeBusy(from, until, StringBody(session => JSONUtil().toJson(body(session).get)), Option(Map(
      "maskId" -> maskId.getOrElse(None), "merge" -> merge.getOrElse(None)))))
  }

  def listEvents(
    body: Expression[Array[EventId]], fields: Option[String] = None,
    extendedEntities: Option[Boolean] = None): ChainBuilder = group(ModuleNames.CHRONOS) {
    exec(new ListEvents(StringBody(session => JSONUtil().toJson(body(session).get)), Option(Map(
      "fields" -> fields.getOrElse(None), "extendedEntities" -> extendedEntities.getOrElse(None)))))
  }

  def updateEvent(
    folder: Expression[String], id: Expression[String], timestamp: Long, body: Expression[EventData],
    recurrenceId: Option[String] = None, recurrenceRange: Option[String] = None, checkConflicts: Option[Boolean] = None,
    sendInternalNotifications: Option[Boolean] = None, extendedEntities: Option[Boolean] = None,
    rangeStart: Option[DateTime] = None, rangeEnd: Option[DateTime] = None,
    expand: Option[Boolean] = None): ChainBuilder = group(ModuleNames.CHRONOS) {
    exec(new UpdateEvent(folder, id, timestamp, StringBody(session => JSONUtil().toJson(body(session).get)), Option(Map(
      "recurrenceId" -> recurrenceId.getOrElse(None),
      "recurrenceRange" -> recurrenceRange.getOrElse(None),
      "checkConflicts" -> checkConflicts.getOrElse(None),
      "sendInternalNotifications" -> sendInternalNotifications.getOrElse(None),
      "extendedEntities" -> extendedEntities.getOrElse(None),
      "rangeStart" -> rangeStart.getOrElse(None),
      "rangeEnd" -> rangeEnd.getOrElse(None),
      "expand" -> expand.getOrElse(None)))))
  }

  def updateAttendee(
    folder: Expression[String], id: Expression[String], timestamp: Long, body: Expression[UpdateAttendeeBody],
    recurrenceId: Option[String] = None, sendInternalNotifications: Option[Boolean] = None,
    extendedEntities: Option[Boolean] = None, checkConflicts: Option[Boolean] = None,
    rangeStart: Option[DateTime] = None, rangeEnd: Option[DateTime] = None,
    expand: Option[Boolean] = None): ChainBuilder = group(ModuleNames.CHRONOS) {
    exec(new UpdateAttendee(folder, id, timestamp, StringBody(session => JSONUtil().toJson(body(session).get)),
      Option(Map(
        "recurrenceId" -> recurrenceId.getOrElse(None),
        "checkConflicts" -> checkConflicts.getOrElse(None),
        "sendInternalNotifications" -> sendInternalNotifications.getOrElse(None),
        "extendedEntities" -> extendedEntities.getOrElse(None),
        "rangeStart" -> rangeStart.getOrElse(None),
        "rangeEnd" -> rangeEnd.getOrElse(None),
        "expand" -> expand.getOrElse(None)))))
  }

  def createEvent(folder: Expression[String], body: Expression[EventData], checkConflicts: Option[Boolean] = None,
                  sendInternalNotifications: Option[Boolean] = None, extendedEntities: Option[Boolean] = None,
                  rangeStart: Option[DateTime] = None, rangeEnd: Option[DateTime] = None, expand: Option[Boolean] = None): ChainBuilder = group(ModuleNames.CHRONOS) {
    exec(new CreateEvent(folder, StringBody(session => JSONUtil().toJson(body(session).get)), Option(Map(
      "checkConflicts" -> checkConflicts.getOrElse(None),
      "sendInternalNotifications" -> sendInternalNotifications.getOrElse(None),
      "extendedEntities" -> extendedEntities.getOrElse(None),
      "rangeStart" -> rangeStart.getOrElse(None),
      "rangeEnd" -> rangeEnd.getOrElse(None),
      "expand" -> expand.getOrElse(None)))))
  }

  def getEvent(id: Expression[String], folder: Expression[String]): ChainBuilder = group(ModuleNames.CHRONOS) {
    exec(new GetEvent(id, folder).buildHttpRequest()
      .check(jsonPath("$.data").transform(JSONUtil().fromJson[EventData](_))
        .saveAs(classOf[GetEvent].getSimpleName))
      .check(jsonPath("$.data.attendees[*].entity").findAll.optional
        .saveAs(s"${classOf[GetEvent].getSimpleName}/${classOf[Attendee].getSimpleName}")))
  }

  def needsAction(rangeStart: DateTime, rangeEnd: DateTime): ChainBuilder = group(ModuleNames.CHRONOS) {
    exec(new NeedsAction(rangeStart, rangeEnd))
  }

  def resolve(id: Expression[String]): ChainBuilder = group(ModuleNames.CHRONOS) {
    exec(new Resolve(id))
  }

  def getUpdates(folder: Expression[String], timestamp: Expression[Long]): ChainBuilder = group(ModuleNames.CHRONOS) {
    exec(new Updates(folder, timestamp))
  }

  // Alarm
  def acknowledge(id: Expression[String], folder: Expression[String], alarmId: Expression[Int]): ChainBuilder = {
    group(ModuleNames.CHRONOS) {
      exec(new Ack(id, folder, alarmId))
    }
  }

  def pending(rangeEnd: DateTime, actions: Option[String] = None): ChainBuilder = group(ModuleNames.CHRONOS) {
    group(ModuleNames.ALARM) {
      exec(new Pending(rangeEnd, Option(Map("actions" -> actions.getOrElse(None)))))
    }
  }

  def snooze(id: Expression[String], folder: Expression[String], alarmId: Expression[Int],
             snoozeTime: Long = 300000): ChainBuilder = group(ModuleNames.CHRONOS) {
    exec(new Snooze(id, folder, alarmId, snoozeTime))
  }

  def updateAlarm() = ???

  // Availability
  def getAvailability(): ChainBuilder = group(ModuleNames.CHRONOS) {
    exec(new GetAvailability().buildHttpRequest()
      .check(jsonPath("$.data")
        .transform(json => JSONUtil().fromJson[Seq[Available]](json))
        .saveAs(classOf[GetAvailability].getSimpleName)))
  }

  def setAvailability(body: AvailabilityData): ChainBuilder = group(ModuleNames.CHRONOS) {
    exec(new SetAvailability(StringBody(session => JSONUtil().toJson(body(session).get))))
  }

  // SchedJoules
  def browseSchedJoules(
    id:       Option[Int]    = None,
    language: Option[String] = None,
    country:  Option[String] = None): ChainBuilder = group(ModuleNames.CHRONOS) {
    exec(new BrowseSchedJoules(("id", id), ("language", language), ("country", country)))
  }

  def countriesSchedJoules(language: Option[String] = None): ChainBuilder = group(ModuleNames.CHRONOS) {
    exec(new CountriesSchedJoules(("language", language)))
  }

  def languagesSchedJoules(): ChainBuilder = group(ModuleNames.CHRONOS) {
    exec(new LanguagesSchedJoules())
  }

  def searchSchedJoules(
    query:    String,
    language: Option[String] = None,
    maxRows:  Option[Int]    = None): ChainBuilder = group(ModuleNames.CHRONOS) {
    exec(new SearchSchedJoules(query, ("language", language), ("maxRows", maxRows)))
  }

}
