/*
 * Copyright (C) 2018 OX Software GmbH
 *
 * This file is part of OX API-Performance Tests.
 *
 * OX API-Performance Tests is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * OX API-Performance Tests is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OX API-Performance Tests.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.openexchange.testing.performance.ajax.chronos

import java.time.temporal.ChronoUnit

import com.openexchange.testing.performance.ajax.contacts.{ ContactsAPI, GetContact }
import com.openexchange.testing.performance.ajax.folders.FoldersAPI
import com.openexchange.testing.performance.ajax.halo.{ ContactServices, HaloAPI }
import com.openexchange.testing.performance.ajax.jslob.JSlobAPI
import com.openexchange.testing.performance.ajax.user.UserAPI
import com.openexchange.testing.performance.common.Capabilities
import com.openexchange.testing.performance.common.SessionConstants._
import com.openexchange.testing.performance.common.utils.{ ChronosEventGenerator, DateTimeHelper, RandomHelper }
import com.openexchange.testing.performance.config.Configuration
import com.openexchange.testing.performance.groupware.chronos.{ Attendee, ChronosFolderBody, EventData, EventId }
import io.gatling.core.Predef._
import io.gatling.core.structure.ChainBuilder

/**
 * [[ChronosChain]]
 *
 * @author <a href="mailto:nikolaos.tsapanidis@open-xchange.com">Nikolaos Tsapanidis</a>
 */
object ChronosChain extends Configuration {
  private val getAllEventsOfFoldersClassName = classOf[GetAllEventsOfFolders].getSimpleName
  private val contactServicesClassName = classOf[ContactServices].getSimpleName
  private val getEventClassName = classOf[GetEvent].getSimpleName
  private val attendeeClassName = classOf[Attendee].getSimpleName

  val chronosChain: ChainBuilder = {
    randomSwitch(
      45d -> idleChain(),
      20d -> createChain(),
      5d -> deleteChain(),
      30d -> readChain())
  }

  def idleChain(): ChainBuilder = {
    exec(FoldersAPI.getAllVisibleFolder("event", "1,2,3,4,5,6,20,23,300,301,302,304," +
      "305,306,307,308,309,310,311,312,313,314,315,316,317,318,319,3010,3020,3030,3050,3201,3203,3204,3205,3220", 1))
      .exec(UserAPI.getUserList("1,20,500,501,502,505,524,555,606,614", "[${OX_USER_ID}]"))

      .exec(session => for {
        chronosFolderBody <- session("OX_FID_CHRONOS").validate[String]
      } yield session.set("SELECTED_CHRONOS_FOLDER", ChronosFolderBody(Array(chronosFolderBody, "cal://0/allPublic"))))

      // today() -> +5
      .exec(ChronosAPI.getAllEventsOfMultipleFolders(
        rangeStart = DateTimeHelper.now().startOfDay(),
        rangeEnd = DateTimeHelper.forOrBackward(5, ChronoUnit.DAYS).startOfDay(), body = "${SELECTED_CHRONOS_FOLDER}",
        expand = Some(true), fields = Some("lastModified,color,createdBy,endDate,flags,folder,id,location,recurrenceId," +
          "seriesId,startDate,summary,timestamp,transp"), sort = Some("startDate")))
      // -5 -> today()
      .exec(ChronosAPI.getAllEventsOfMultipleFolders(
        rangeStart = DateTimeHelper.forOrBackward(-5, ChronoUnit.DAYS).startOfDay(),
        rangeEnd = DateTimeHelper.now().startOfDay(), body = "${SELECTED_CHRONOS_FOLDER}",
        expand = Some(true), fields = Some("lastModified,color,createdBy,endDate,flags,folder,id,location,recurrenceId," +
          "seriesId,startDate,summary,timestamp,transp"), sort = Some("startDate")))
      // -10 -> -5
      .exec(ChronosAPI.getAllEventsOfMultipleFolders(
        rangeStart = DateTimeHelper.forOrBackward(-10, ChronoUnit.DAYS).startOfDay(),
        rangeEnd = DateTimeHelper.forOrBackward(-5, ChronoUnit.DAYS).startOfDay(), body = "${SELECTED_CHRONOS_FOLDER}",
        expand = Some(true), fields = Some("lastModified,color,createdBy,endDate,flags,folder,id,location,recurrenceId," +
          "seriesId,startDate,summary,timestamp,transp"), sort = Some("startDate")))
  }

  def createChain(): ChainBuilder = {
    doIf(session => session(CAPABILITIES).as[Seq[String]].contains(
      Capabilities.calendarChronos) && settings.appsuite.modules.contains(Capabilities.calendarChronos)) {
      exec(ChronosAPI.createEvent(
        $(FID_CHRONOS), _ => ChronosEventGenerator.genEvent().get, checkConflicts = Some(true),
        sendInternalNotifications = Some(false), expand = Some(true),
        rangeStart = Some(DateTimeHelper.forOrBackward(-3, ChronoUnit.DAYS).startOfDay()),
        rangeEnd = Some(DateTimeHelper.forOrBackward(2, ChronoUnit.DAYS).startOfDay())))

        .exec(ChronosAPI.pending(DateTimeHelper.forOrBackward(8, ChronoUnit.HOURS), Some("DISPLAY,AUDIO")))

        .exec(session => for {
          chronosFolderBody <- session("OX_FID_CHRONOS").validate[String]
        } yield session.set("SELECTED_CHRONOS_FOLDER", ChronosFolderBody(Array(chronosFolderBody, "cal://0/allPublic"))))
        // -10 -> -5
        .exec(ChronosAPI.getAllEventsOfMultipleFolders(
          rangeStart = DateTimeHelper.forOrBackward(-10, ChronoUnit.DAYS).startOfDay(),
          rangeEnd = DateTimeHelper.forOrBackward(-5, ChronoUnit.DAYS).startOfDay(), body = "${SELECTED_CHRONOS_FOLDER}",
          expand = Some(true), fields = Some("lastModified,color,createdBy,endDate,flags,folder,id,location,recurrenceId," +
            "seriesId,startDate,summary,timestamp,transp"), sort = Some("startDate"), order = Some("asc")))
        // today() -> +5
        .exec(ChronosAPI.getAllEventsOfMultipleFolders(
          rangeStart = DateTimeHelper.now().startOfDay(),
          rangeEnd = DateTimeHelper.forOrBackward(5, ChronoUnit.DAYS).startOfDay(), body = "${SELECTED_CHRONOS_FOLDER}",
          expand = Some(true), fields = Some("lastModified,color,createdBy,endDate,flags,folder,id,location,recurrenceId," +
            "seriesId,startDate,summary,timestamp,transp"), sort = Some("startDate"), order = Some("asc")))

        .exec(FoldersAPI.getFolder($(FID_CHRONOS), tree = 1, altNames = Some(true)))
        .exec(UserAPI.getUserList("1,20,500,501,502,505,524,555,606,614", "[${OX_USER_ID}]"))
    }
  }

  def bareCreateChain(recurring: Boolean = false): ChainBuilder = {
    doIf(session => session(CAPABILITIES).as[Seq[String]].contains(
      Capabilities.calendarChronos) && settings.appsuite.modules.contains(Capabilities.calendarChronos)) {
      exec(ChronosAPI.createEvent(
        $(FID_CHRONOS), _ => ChronosEventGenerator.genEvent(if (recurring) 1 else 2).get, checkConflicts = Some(true),
        sendInternalNotifications = Some(false), expand = Some(true),
        rangeStart = Some(DateTimeHelper.forOrBackward(-3, ChronoUnit.DAYS).startOfDay()),
        rangeEnd = Some(DateTimeHelper.forOrBackward(2, ChronoUnit.DAYS).startOfDay())))

        .exec(ChronosAPI.pending(DateTimeHelper.forOrBackward(8, ChronoUnit.HOURS), Some("DISPLAY,AUDIO")))
    }
  }

  def readChain(): ChainBuilder = {
    doIf(session => session(CAPABILITIES).as[Seq[String]].contains(
      Capabilities.calendarChronos) && settings.appsuite.modules.contains(Capabilities.calendarChronos)) {
      exec(session => for {
        chronosFolderBody <- session("OX_FID_CHRONOS").validate[String]
      } yield session.set("SELECTED_CHRONOS_FOLDER", ChronosFolderBody(Array(chronosFolderBody, "cal://0/allPublic"))))
        .exec(ChronosAPI.getAllEventsOfMultipleFolders(
          rangeStart = DateTimeHelper.forOrBackward(-1, ChronoUnit.DAYS).startOfDay(),
          rangeEnd = DateTimeHelper.forOrBackward(5, ChronoUnit.DAYS).startOfDay(), body = "${SELECTED_CHRONOS_FOLDER}",
          expand = Some(true), fields = Some("lastModified,color,createdBy,endDate,flags,folder,id,location,recurrenceId," +
            "seriesId,startDate,summary,timestamp,transp"), sort = Some("startDate"), order = Some("asc")))

        .doIf("${" + getAllEventsOfFoldersClassName + "/Ids.exists()}") {
          exec(ChronosAPI.getEvent("${" + getAllEventsOfFoldersClassName + "/Ids.random()}", $(FID_CHRONOS)))

            .randomSwitch(50d -> exec(HaloAPI.getHaloServices)
              .doIf("${" + contactServicesClassName + ".exists()}") {
                foreach("${" + contactServicesClassName + "}", "provider") {
                  doSwitch("${provider}")(
                    "com.openexchange.halo.mail" -> exec(HaloAPI.investigateContactHalo(
                      "com.openexchange.halo.mail",
                      "102,600,601,602,603,604,605,606,607,608,609,610,611,612,614,652",
                      limit = Some(10), timezone = Some("utc"))),
                    "com.openexchange.halo.contacts" -> exec(HaloAPI.investigateContactHalo(
                      "com.openexchange.halo.contacts",
                      "1,2,3,4,5,20,100,101,102,104,500,501,502,503,504,505,506,507,508,509,510,511,512,513," +
                        "514,515,516,517,518,519,520,521,522,523,524,525,526,527,528,529,530,531,532,533,534,535,536," +
                        "537,538,539,540,541,542,543,544,545,546,547,548,549,550,551,552,553,554,555,556,557,558,559," +
                        "560,561,562,563,564,565,566,567,568,569,571,572,573,574,575,576,577,578,579,580,581,582,583," +
                        "584,585,586,587,588,589,590,592,594,596,597,598,599,601,602,605,606,607,608,610,611,612",
                      timezone = Some("utc"))),
                    "com.openexchange.halo.events" -> exec(HaloAPI.investigateContactHalo(
                      "com.openexchange.halo.events",
                      "",
                      start = Some(DateTimeHelper.now().getTimestamp),
                      end = Some(DateTimeHelper.forOrBackward(10, ChronoUnit.DAYS).getTimestamp))))
                }
                  .exec(ContactsAPI.getAllUserFromGAB)
                  .exec(FoldersAPI.getFolderPath(
                    "6",
                    "1,2,3,4,5,6,20,23,300,301,302,304,305,306,307,308,309,310,311,312,313,314,315,316,317,318," +
                      "319,3010,3020,3030,3050,3201,3203,3204,3205,3220", 1, altNames = Some(true)))
                  .doIf("${" + getEventClassName + "/" + attendeeClassName + ".exists()}") {
                    exec(JSlobAPI.getJSLobPortal)
                      // TODO add width and height
                      .exec(ContactsAPI.getContactPicture(user_id = Some("${OX_USER_ID}"), context = "${OX_CID}",
                        email = Some("${OX_USER_EMAIL1}"), scaleType = Some("cover")))
                      // -7 -> today()
                      .exec(ChronosAPI.getAllEventsOfMultipleFolders(
                        rangeStart = DateTimeHelper.forOrBackward(-7, ChronoUnit.DAYS).startOfDay(),
                        rangeEnd = DateTimeHelper.now().startOfDay(), body = "${SELECTED_CHRONOS_FOLDER}",
                        expand = Some(true), fields = Some("lastModified,color,createdBy,endDate,flags,folder,id,location," +
                          "recurrenceId,seriesId,startDate,summary,timestamp,transp"), sort = Some("startDate")))
                  }
              })
        }
    }
  }

  def deleteChain(): ChainBuilder = {
    doIf(session => session(CAPABILITIES).as[Seq[String]].contains(Capabilities.calendarChronos) &&
      settings.appsuite.modules.contains(Capabilities.calendarChronos)) {
      exec(idleChain())
        .doIf("${" + getAllEventsOfFoldersClassName + ".exists()}") {
          exec(session => for {
            listOfEvents <- session(getAllEventsOfFoldersClassName).validate[Vector[EventData]]
          } yield {
            val randomEvent = RandomHelper.getRandomElement(listOfEvents)

            if (listOfEvents.nonEmpty)
              session.set("DELETE_EVENT", Array(EventId(randomEvent.folder.get, id = randomEvent.id.get)))
            else session
          })
            .doIf("${DELETE_EVENT.exists()}") {
              exec(ChronosAPI.deleteEvents(System.currentTimeMillis(), "${DELETE_EVENT}", expand = Some(true),
                rangeStart = Some(DateTimeHelper.now().startOfDay()),
                rangeEnd = Some(DateTimeHelper.forOrBackward(1, ChronoUnit.WEEKS).startOfDay()),
                fields = Some("lastModified,color,createdBy,endDate,flags,folder,id,location,recurrenceId,seriesId," +
                  "startDate,summary,timestamp,transp")))
                .exec(ChronosAPI.getAllEventsOfMultipleFolders(
                  rangeStart = DateTimeHelper.forOrBackward(-3, ChronoUnit.DAYS).startOfDay(),
                  rangeEnd = DateTimeHelper.forOrBackward(3, ChronoUnit.DAYS).startOfDay(), body =
                    "${SELECTED_CHRONOS_FOLDER}",
                  expand = Some(true), fields = Some("lastModified,color,createdBy,endDate,flags,folder,id,location,recurrenceId," +
                    "seriesId,startDate,summary,timestamp,transp"), sort = Some("startDate")))
                .exec(ChronosAPI.pending(DateTimeHelper.forOrBackward(8, ChronoUnit.HOURS), Some("DISPLAY,AUDIO")))
                .exec(FoldersAPI.getFolder($(FID_CHRONOS), tree = 1, altNames = Some(true)))

                .exec(cleanUpSession())
            }
        }
    }
  }

  def cleanUpSession(): ChainBuilder = exec(session => {
    session.removeAll(contactServicesClassName, s"$getEventClassName/$attendeeClassName", getEventClassName,
      getAllEventsOfFoldersClassName, "getFolderPathResponse", classOf[GetContact].getSimpleName,
      s"$getAllEventsOfFoldersClassName/Ids", "DELETE_EVENT")
  })

}
