/*
 * Copyright (C) 2018 OX Software GmbH
 *
 * This file is part of OX API-Performance Tests.
 *
 * OX API-Performance Tests is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * OX API-Performance Tests is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OX API-Performance Tests.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.openexchange.testing.performance.ajax.contacts

import com.openexchange.testing.performance.common.Capabilities
import com.openexchange.testing.performance.common.SessionConstants._
import com.openexchange.testing.performance.common.utils.{ CapabilityCheck, ContactGenerator }
import com.openexchange.testing.performance.config.Configuration
import com.openexchange.testing.performance.groupware.contacts.ContactData
import com.typesafe.scalalogging.LazyLogging
import io.gatling.core.Predef._
import io.gatling.core.session.Expression
import io.gatling.core.structure.ChainBuilder

/**
 * [[ContactChain]]
 *
 * @author <a href="mailto:nikolaos.tsapanidis@open-xchange.com">Nikolaos Tsapanidis</a>
 */
object ContactChain extends Configuration with LazyLogging {
  private val contactDataClassName = classOf[ContactData].getSimpleName
  private val createContactClassName = classOf[CreateContact].getSimpleName
  private val getContactResponseName = classOf[GetContact].getSimpleName
  private val getAllContactsClassName = classOf[GetAllContacts].getSimpleName
  private val allContactsBufferName = "allContactsBuffer"
  private val contactListResponseName = "contactListResponse"

  val contactChain: ChainBuilder = {
    randomSwitch(
      40d -> exec(createChain($(FID_CONTACTS))))
    //      20d -> exec(deleteChain()),
    //      10d -> exec(updateChain()))
  }

  def createChain(folder_id: Expression[String]): ChainBuilder = {
    CapabilityCheck(Capabilities.contacts, exec(session => for {
      fid <- folder_id(session)
    } yield {
      val contactData: ContactData = ContactGenerator(fid)
      logger.trace("{}", contactData)
      session.set(contactDataClassName, contactData)
    })
      .doIf(session => session.contains(contactDataClassName)) {
        exec(ContactsAPI.createContact("${" + contactDataClassName + "}"))
          .doIfOrElse(session => session.contains(createContactClassName)) {
            exec(session => for {
              id <- session(createContactClassName).validate[Int]
              fid <- folder_id(session)
            } yield {
              logger.trace("Successfully created contact with id {} in folder {}", id, fid)
              session.remove(contactDataClassName)
            })
              .exec(ContactsAPI.getContact("${" + createContactClassName + "}", folder_id))
              .exec(ContactsAPI.getAllContacts(folder_id, "1,20,101,607", Some(Map("sort" -> "607", "order" -> "asc"))))
              // TODO add width and height
              .exec(ContactsAPI.getContactPicture(
                context = "${OX_CID}", user = Some("${OX_USER_ID}"), folder_id = Some("${OX_FID_CONTACTS}"),
                contact_id = Some("${" + createContactClassName + "}"), scaleType = Some("cover")))

              .doIf(session => session.contains(getAllContactsClassName) &&
                session(getAllContactsClassName).as[GetAllContactsResponse].data.nonEmpty) {
                exec(session => for {
                  list <- session(getAllContactsClassName).validate[GetAllContactsResponse]
                } yield {
                  var allContactsBuffer = collection.mutable.ListBuffer[Map[String, Any]]()
                  list.data.foreach(x => {
                    allContactsBuffer += Map(
                      "id" -> x.head, "folder" -> x(1))
                  })
                  session.set(allContactsBufferName, allContactsBuffer.toArray)
                })
                  .exec(ContactsAPI.getContactsList(
                    "1,2,5,20,101,500,501,502,505,520,524,555,556,557,569,592,602,606,607",
                    "${" + allContactsBufferName + "}"))

                  .doIf(session => session.contains(createContactClassName)) {
                    exec(session => for {
                      id <- session(createContactClassName).validate[Int]
                      fid <- folder_id(session)
                    } yield {
                      session.set(contactListResponseName, Array(Map(
                        "id" -> session(createContactClassName).as[Int],
                        "folder" -> folder_id(session).get.toInt)))
                    })
                      .exec(ContactsAPI.getContactsList(
                        "1,2,5,20,101,500,501,502,505,520,524,555,556,557,569,592,602,606,607",
                        "${" + contactListResponseName + "}"))
                  }

                  .exec(session => {
                    session.removeAll(
                      getAllContactsClassName, allContactsBufferName, getContactResponseName)
                  })
              }

          } {
            exec(session => {
              logger.warn("{} failed.", createContactClassName)
              session
            })
          }
      })
  }

  def bareCreateChain(folder_id: Expression[String]): ChainBuilder = {
    CapabilityCheck(Capabilities.contacts, exec(session => for {
      fid <- folder_id(session)
    } yield {
      val contactData: ContactData = ContactGenerator(fid)
      logger.trace("{}", contactData)
      session.set(contactDataClassName, contactData)
    })
      .doIf(session => session.contains(contactDataClassName)) {
        exec(ContactsAPI.createContact("${" + contactDataClassName + "}"))
          .doIfOrElse(session => session.contains(createContactClassName)) {
            exec(session => for {
              id <- session(createContactClassName).validate[Int]
              fid <- folder_id(session)
            } yield {
              logger.trace("Successfully created contact with id {} in folder {}", id, fid)
              session.remove(contactDataClassName)
            })
          } {
            exec(session => {
              logger.warn("{} failed.", createContactClassName)
              session
            })
          }
      })
  }

  def deleteChain(): ChainBuilder = ???

  def updateChain(): ChainBuilder = ???

}
