/*
 * Copyright (C) 2018 OX Software GmbH
 *
 * This file is part of OX API-Performance Tests.
 *
 * OX API-Performance Tests is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * OX API-Performance Tests is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OX API-Performance Tests.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.openexchange.testing.performance.ajax.contacts

import com.openexchange.testing.performance.common.ModuleNames
import com.openexchange.testing.performance.common.utils.JSONUtil
import com.openexchange.testing.performance.groupware.contacts.{ ContactData, SearchContact }
import io.gatling.core.Predef._
import io.gatling.core.session.Expression
import io.gatling.core.structure.ChainBuilder
import io.gatling.http.Predef._

/**
 * [[ContactsAPI]]
 *
 * @author <a href="mailto:nikolaos.tsapanidis@open-xchange.com">Nikolaos Tsapanidis</a>
 */
object ContactsAPI {
  private val moduleName = ModuleNames.CONTACTS

  /**
   * Retrieve all internal user from the global address book.
   */
  val getAllUserFromGAB: ChainBuilder = group(moduleName) {
    exec(new GetAllContacts("6", "1,20,101,607")
      .buildHttpRequest()
      .queryParam("sort", "607")
      .queryParam("order", "asc")
      .queryParam("admin", "false")
      .check(jsonPath("$.data[*][0]").ofType[Int].findAll.saveAs("ALL_GAB_USER")))
  }

  /**
   * Creates a contact
   */
  def createContact(body: Expression[ContactData]): ChainBuilder = group(moduleName) {
    exec(new CreateContact(body).buildHttpRequest()
      .check(jsonPath("$.data.id").ofType[Int]
        .saveAs(classOf[CreateContact].getSimpleName)))
  }

  /**
   * Creates a contact.
   */
  def createContactAdvanced = ???

  /**
   * Deletes contacts.
   */
  def deleteContacts(timestamp: String, folderId: String, objectId: String): ChainBuilder = group(moduleName) {
    exec(new DeleteContact(timestamp, JSONUtil().toJson(Seq(Map("folder" -> folderId, "id" -> objectId)))))
  }

  /**
   * Auto-complete contacts.
   */
  def doAutoCompleteContacts = ???

  /**
   * Gets all contacts.
   */
  def getAllContacts(
    folder:              Expression[String],
    columns:             String,
    optionalQueryParams: Option[Map[String, Any]] = None): ChainBuilder = group(moduleName) {
    exec(new GetAllContacts(folder, columns).buildHttpRequest()
      .queryParamMap(optionalQueryParams.getOrElse(Map.empty))
      .check(bodyString.transform(JSONUtil().fromJson[GetAllContactsResponse](_))
        .optional
        .saveAs(classOf[GetAllContacts].getSimpleName)))
  }

  /**
   * Gets a contact.
   */
  def getContact(id: Expression[String], folder: Expression[String]): ChainBuilder = group(moduleName) {
    exec(new GetContact(id, folder).buildHttpRequest()
      .check(jsonPath("$.data").transform(x => JSONUtil().fromJson[ContactData](x))
        .optional
        .saveAs(classOf[GetContact].getSimpleName)))
  }

  /**
   * Gets a contact by user ID.
   */
  def getContactByUser = ???

  /**
   * Gets a list of contacts.
   */
  def getContactsList(columns: String, objectIds: Expression[Array[Map[String, Any]]]): ChainBuilder = group(ModuleNames.CONTACTS) {
    exec(new ListContacts(columns, objectIds))
  }

  /**
   * Gets a list of users.
   */
  def getContactListByUsers = ???

  /**
   * Gets updated contacts.
   */
  def getContactUpdates = ???

  /**
   * Search for contacts.
   */
  def searchContacts(columns: String, searchContact: SearchContact, optionalQueryParams: Option[Map[String, Any]] = None): ChainBuilder = group(moduleName) {
    exec(new SearchContacts(columns, JSONUtil().toJson(searchContact)).buildHttpRequest()
      .queryParamMap(optionalQueryParams.getOrElse(Map.empty)))
  }

  /**
   * Search for contacts by filter.
   */
  def searchContactsAdvanced = ???

  /**
   * Search for contacts by anniversary.
   */
  def searchContactsByAnniversary = ???

  /**
   * Search for contacts by birthday.
   */
  def searchContactsByBirthday(columns: String, start: Long, end: Long, optionalQueryParams: Option[Map[String, Any]] = None): ChainBuilder = group(moduleName) {
    exec(new GetBirthdays(columns, start, end).buildHttpRequest()
      .queryParamMap(optionalQueryParams.getOrElse(Map.empty)))
  }

  /**
   * Updates a contact.
   */
  def updateContact(id: Expression[Int], folder: Expression[String], lastUpdate: Expression[Long], contact: Expression[ContactData]): ChainBuilder = group(moduleName) {
    exec(new UpdateContact(id, folder, lastUpdate, contact))
  }

  /**
   * Updates a contact.
   */
  def updateContactAdvanced = ???

  def getContactPicture(
    context:    Expression[Int],
    email:      Option[String]  = None,
    user_id:    Option[String]  = None,
    user:       Option[String]  = None,
    folder_id:  Option[String]  = None,
    contact_id: Option[String]  = None,
    rotate:     Option[Boolean] = None, width: Option[Int] = None, height: Option[Int] = None,
    shrinkOnly: Option[Boolean] = None, scaleType: Option[String] = None, cropWidth: Option[Int] = None,
    cropHeight: Option[Int] = None, cropX: Option[Int] = None,
    cropY: Option[Int] = None): ChainBuilder = group(moduleName) {
    exec(new GetContactPicture(
      Option(Map(
        "rotate" -> rotate.getOrElse(None), "width" -> width.getOrElse(None), "height" -> height.getOrElse(None),
        "shrinkOnly" -> shrinkOnly.getOrElse(None), "scaleType" -> scaleType.getOrElse(None),
        "cropWidth" -> cropWidth.getOrElse(None), "cropHeight" -> cropHeight.getOrElse(None),
        "cropX" -> cropX.getOrElse(None), "cropY" -> cropY.getOrElse(None), "user_id" -> user_id.getOrElse(None),
        "email" -> email.getOrElse(None), "user" -> user.getOrElse(None), "folder_id" -> folder_id.getOrElse(None),
        "contact_id" -> contact_id.getOrElse(None)))).buildHttpRequest()
      .queryParam("context", context))
  }

}
