/*
 * Copyright (C) 2018 OX Software GmbH
 *
 * This file is part of OX API-Performance Tests.
 *
 * OX API-Performance Tests is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * OX API-Performance Tests is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OX API-Performance Tests.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.openexchange.testing.performance.ajax.folders

import com.openexchange.testing.performance.common.ModuleNames
import com.openexchange.testing.performance.common.utils.JSONUtil
import com.openexchange.testing.performance.groupware.folders.FolderBody
import io.gatling.core.Predef._
import io.gatling.core.session.Expression
import io.gatling.core.structure.ChainBuilder
import io.gatling.http.Predef._

/**
 * [[FoldersAPI]]
 *
 * @author <a href="mailto:nikolaos.tsapanidis@open-xchange.com">Nikolaos Tsapanidis</a>
 */
object FoldersAPI {

  /**
   * Clears the content of a list of folders.
   *
   * @param jsonBody A JSON array with object IDs of the folders.
   * @return
   */
  def clearFolders(jsonBody: Array[String]): ChainBuilder = group(ModuleNames.FOLDERS) {
    exec(new FoldersCleanUp(JSONUtil().toJson(jsonBody)))
  }

  /**
   * Creates a new folder.
   */
  def createMailFolder(folder_id: String, subscribed: Int, name: String,
                       optionalQueryParams: Option[Map[String, Any]] = None): ChainBuilder = group(ModuleNames.MAIL) {
    exec(new CreateFolder(folder_id, JSONUtil().toJson(Map("module" -> "mail", "subscribed" -> subscribed,
      "title" -> name)))
      .buildHttpRequest()
      .queryParamMap(optionalQueryParams.getOrElse(Map.empty)))
  }

  def createFolder(folder_id: Expression[String], body: Expression[FolderBody], tree: Int = 0,
                   optionalQueryParams: Option[Map[String, Any]] = None): ChainBuilder = {
    group(ModuleNames.FOLDERS) {
      exec(new CreateFolder(folder_id, body, tree, optionalQueryParams).buildHttpRequest()
        .check(jsonPath("$.data").ofType[String].saveAs(classOf[CreateFolder].getSimpleName)))
    }
  }

  /**
   * Deletes a list of folders.
   */
  def deleteFolder(folderId: String, tree: Int = 0, optionalQueryParams: Option[Map[String, Any]] = None): ChainBuilder =
    group(ModuleNames.FOLDERS) {
      exec(new DeleteFolders(JSONUtil().toJson(Seq(folderId)), tree)
        .buildHttpRequest()
        .queryParamMap(optionalQueryParams.getOrElse(Map.empty)))
    }

  /**
   * Gets a folder.
   */
  def getFolder(id: Expression[String], tree: Int = 0, allowed_modules: Option[String] = None,
                altNames: Option[Boolean] = None): ChainBuilder = {
    group(ModuleNames.FOLDERS) {
      exec(new GetFolder(id, tree, Option(Map(
        "allowed_modules" -> allowed_modules.getOrElse(None), "altNames" -> altNames.getOrElse(None)))))
    }
  }

  /**
   * Gets the new, modified and deleted folders of a given folder.
   */
  def getFolderUpdates(columns: String, tree: Int, timestamp: Long = System.currentTimeMillis()): ChainBuilder = {
    group(ModuleNames.FOLDERS) {
      exec(new GetFolderUpdates(timestamp, columns, tree))
    }
  }

  /**
   * Gets the parent folders above the specified folder.
   */
  def getFolderPath(folderId: Expression[String], columns: String, tree: Int = 0,
                    allowed_modules: Option[String]  = None,
                    altNames:        Option[Boolean] = None): ChainBuilder = group(ModuleNames.FOLDERS) {
    exec(new GetFolderPath(folderId, columns, tree, Option(Map(
      "allowed_modules" -> allowed_modules.getOrElse(None), "altNames" -> altNames.getOrElse(None)))).buildHttpRequest()
      .check(jsonPath("$.data[*][0]").findAll.optional.saveAs("getFolderPathResponse")))
  }

  /**
   * Gets the folders at the root level of the folder structure.
   */
  def getRootFolders(columns: String): ChainBuilder = group(ModuleNames.FOLDERS) {
    exec(new GetRootFolders(columns))
  }

  /**
   * Gets shared folders of a certain module.
   */
  def getSharedFolders = ???

  /**
   * Gets the subfolders of a specified parent folder.
   */
  def getSubFolders(parent: String, columns: String, optionalQueryParams: Option[Map[String, Any]] = None): ChainBuilder = group(ModuleNames.FOLDERS) {
    exec(new GetSubFolders(parent, columns)
      .buildHttpRequest()
      .queryParamMap(optionalQueryParams.getOrElse(Map.empty))
      .check(jsonPath("$.data").transform(x => JSONUtil().fromJson[Seq[Seq[Any]]](x))
        .optional
        .saveAs(classOf[GetSubFolders].getSimpleName)))
  }

  /**
   * Gets all visible folders of a certain module.
   */
  def getAllVisibleFolder(content_type: String, columns: String, tree: Int = 0): ChainBuilder = group(ModuleNames
    .FOLDERS) {
    exec(new GetAllVisibleFolders(content_type, columns, tree))
  }

  /**
   * Notifies users or groups about a shared folder.
   */
  def notifyAboutSharedFolder = ???

  /**
   * Updates a folder.
   */
  def updateFolder = ???

}
