/*
 * Copyright (C) 2018 OX Software GmbH
 *
 * This file is part of OX API-Performance Tests.
 *
 * OX API-Performance Tests is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * OX API-Performance Tests is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OX API-Performance Tests.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.openexchange.testing.performance.ajax.login

import com.openexchange.testing.performance.common.SessionConstants._
import com.openexchange.testing.performance.common.request.{ AbstractAJAXRequest, RequestMethod, RequestParameters }
import io.gatling.core.Predef._
import io.gatling.http.Predef._
import io.gatling.http.check.HttpCheck
import io.gatling.http.check.body.HttpBodyJsonPathCheckBuilder
import io.gatling.http.request.builder.HttpRequestBuilder

/**
 * The [[Login]] is used to obtain a session from the user's login credentials.
 *
 * @author <a href="mailto:nikolaos.tsapanidis@open-xchange.com">Nikolaos Tsapanidis</a>
 * @constructor
 * @param username The login name.
 * @param password The password.
 */
class Login(username: String = USERNAME, password: String = PASSWORD) extends AbstractAJAXRequest {

  import Login._

  override lazy val headers: Map[String, String] = Map(
    "Accept" -> "application/json, text/javascript, */*; q=0.01")
  override lazy val checks: Seq[HttpCheck] = Seq(
    jsonPath("$.code").notExists,
    sessionId.saveAs(SESSION_ID),
    userId.saveAs(USER_ID),
    contextId.saveAs(CONTEXT_ID),
    locale.saveAs(LOCALE),
    // user specific identifier
    userLastname.optional.saveAs(USER_LAST_NAME),
    userFirstname.optional.saveAs(USER_FIRST_NAME),
    email1.optional.saveAs(USER_EMAIL1),
    // default system folder identifier
    fid_calender.optional.saveAs(FID_CALENDAR),
    fid_contacts.optional.saveAs(FID_CONTACTS),
    fid_tasks.optional.saveAs(FID_TASKS),
    fid_infostore.optional.saveAs(FID_INFOSTORE),
    fid_chronos.optional.saveAs(FID_CHRONOS),
    selectedChronosFolder.findAll.optional.saveAs("SELECTED_CHRONOS_FOLDER"),
    // default mail folder identifier
    fid_inbox.optional.saveAs(FID_INBOX),
    fid_sent.optional.saveAs(FID_SENT),
    fid_drafts.optional.saveAs(FID_DRAFTS),
    fid_spam.optional.saveAs(FID_SPAM),
    fid_trash.optional.saveAs(FID_TRASH),
    // default infostore folder identifier
    fid_infostore_pictures.optional.saveAs(FID_INFOSTORE_PICTURES),
    fid_infostore_trash.optional.saveAs(FID_INFOSTORE_TRASH),
    fid_infostore_music.optional.saveAs(FID_INFOSTORE_MUSIC),
    fid_infostore_videos.optional.saveAs(FID_INFOSTORE_VIDEOS),
    fid_infostore_documents.optional.saveAs(FID_INFOSTORE_DOCUMENTS),
    // capabilities
    capabilities.findAll.transform(_.toList).optional.saveAs(CAPABILITIES))
  override val requestMethod: String = RequestMethod.POST
  override val relativeURL: String = "login?action=login"

  override def buildHttpRequest(): HttpRequestBuilder = super.buildHttpRequest()
    .formParam(RequestParameters.NAME, username)
    .formParam(RequestParameters.PASSWORD, password)
}

object Login {
  private[this] val sSESSION_ID = "$.session"
  private[this] val sUSER_ID = "$.user_id"
  private[this] val sCONTEXT_ID = "$.context_id"
  private[this] val sLOCALE = "$.locale"
  private[this] val sUSER_LAST_NAME = "$.rampup.user.last_name"
  private[this] val sUSER_FIRST_NAME = "$.rampup.user.first_name"
  private[this] val sEMAIL1 = "$.rampup.user.email1"
  private[this] val sFID_CALENDAR = "$.rampup.jslobs.*[?(@.id=='io.ox/core')].tree.folder.calendar"
  private[this] val sFID_CHRONOS = "$.rampup.jslobs.*[?(@.id=='io.ox/calendar')].tree.chronos.defaultFolderId"
  private[this] val sSELECTED_CHRONOS_FOLDER = "$.rampup.jslobs.*[?(@.id=='io.ox/calendar')].tree.selectedFolders[*]"
  private[this] val sFID_CONTACTS = "$.rampup.jslobs.*[?(@.id=='io.ox/core')].tree.folder.contacts"
  private[this] val sFID_TASKS = "$.rampup.jslobs.*[?(@.id=='io.ox/core')].tree.folder.tasks"
  private[this] val sFID_INFOSTORE = "$.rampup.jslobs.*[?(@.id=='io.ox/core')].tree.folder.infostore"
  private[this] val sFID_DRAFTS = "$.rampup.jslobs.*[?(@.id=='io.ox/mail')].tree.folder.drafts"
  private[this] val sFID_INBOX = "$.rampup.jslobs.*[?(@.id=='io.ox/mail')].tree.folder.inbox"
  private[this] val sFID_SENT = "$.rampup.jslobs.*[?(@.id=='io.ox/mail')].tree.folder.sent"
  private[this] val sFID_SPAM = "$.rampup.jslobs.*[?(@.id=='io.ox/mail')].tree.folder.spam"
  private[this] val sFID_TRASH = "$.rampup.jslobs.*[?(@.id=='io.ox/mail')].tree.folder.trash"
  private[this] val sFID_INFOSTORE_PICTURES = "$.rampup.jslobs.*[?(@.id=='io.ox/files')].tree.folder.pictures"
  private[this] val sFID_INFOSTORE_TRASH = "$.rampup.jslobs.*[?(@.id=='io.ox/files')].tree.folder.trash"
  private[this] val sFID_INFOSTORE_MUSIC = "$.rampup.jslobs.*[?(@.id=='io.ox/files')].tree.folder.music"
  private[this] val sFID_INFOSTORE_VIDEOS = "$.rampup.jslobs.*[?(@.id=='io.ox/files')].tree.folder.videos"
  private[this] val sFID_INFOSTORE_DOCUMENTS = "$.rampup.jslobs.*[?(@.id=='io.ox/files')].tree.folder.documents"
  private[this] val sCAP = "$.rampup.serverConfig.capabilities[*].id"

  def sessionId: HttpBodyJsonPathCheckBuilder[String] = jsonPath(sSESSION_ID).ofType[String]

  def userId = jsonPath(sUSER_ID)

  def contextId: HttpBodyJsonPathCheckBuilder[Int] = jsonPath(sCONTEXT_ID).ofType[Int]

  def locale: HttpBodyJsonPathCheckBuilder[String] = jsonPath(sLOCALE).ofType[String]

  def userLastname: HttpBodyJsonPathCheckBuilder[String] = jsonPath(sUSER_LAST_NAME).ofType[String]

  def userFirstname: HttpBodyJsonPathCheckBuilder[String] = jsonPath(sUSER_FIRST_NAME).ofType[String]

  def email1: HttpBodyJsonPathCheckBuilder[String] = jsonPath(sEMAIL1).ofType[String]

  def fid_calender: HttpBodyJsonPathCheckBuilder[Int] = jsonPath(sFID_CALENDAR).ofType[Int]

  def fid_chronos = jsonPath(sFID_CHRONOS).ofType[String]

  def selectedChronosFolder = jsonPath(sSELECTED_CHRONOS_FOLDER).ofType[String]

  def fid_contacts: HttpBodyJsonPathCheckBuilder[Int] = jsonPath(sFID_CONTACTS).ofType[Int]

  def fid_tasks: HttpBodyJsonPathCheckBuilder[Int] = jsonPath(sFID_TASKS).ofType[Int]

  def fid_infostore: HttpBodyJsonPathCheckBuilder[Int] = jsonPath(sFID_INFOSTORE).ofType[Int]

  def fid_inbox: HttpBodyJsonPathCheckBuilder[String] = jsonPath(sFID_INBOX).ofType[String]

  def fid_sent: HttpBodyJsonPathCheckBuilder[String] = jsonPath(sFID_SENT).ofType[String]

  def fid_drafts: HttpBodyJsonPathCheckBuilder[String] = jsonPath(sFID_DRAFTS).ofType[String]

  def fid_spam: HttpBodyJsonPathCheckBuilder[String] = jsonPath(sFID_SPAM).ofType[String]

  def fid_trash: HttpBodyJsonPathCheckBuilder[String] = jsonPath(sFID_TRASH).ofType[String]

  def fid_infostore_pictures: HttpBodyJsonPathCheckBuilder[Int] = jsonPath(sFID_INFOSTORE_PICTURES).ofType[Int]

  def fid_infostore_trash: HttpBodyJsonPathCheckBuilder[Int] = jsonPath(sFID_INFOSTORE_TRASH).ofType[Int]

  def fid_infostore_music: HttpBodyJsonPathCheckBuilder[Int] = jsonPath(sFID_INFOSTORE_MUSIC).ofType[Int]

  def fid_infostore_videos: HttpBodyJsonPathCheckBuilder[Int] = jsonPath(sFID_INFOSTORE_VIDEOS).ofType[Int]

  def fid_infostore_documents: HttpBodyJsonPathCheckBuilder[Int] = jsonPath(sFID_INFOSTORE_DOCUMENTS).ofType[Int]

  def capabilities: HttpBodyJsonPathCheckBuilder[String] = jsonPath(sCAP).ofType[String]
}