/*
 * Copyright (C) 2018 OX Software GmbH
 *
 * This file is part of OX API-Performance Tests.
 *
 * OX API-Performance Tests is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * OX API-Performance Tests is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OX API-Performance Tests.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.openexchange.testing.performance.ajax.login

import com.openexchange.testing.performance.ajax.apps.AppsAPI
import com.openexchange.testing.performance.ajax.mailfilter.MailFilterAPI
import com.openexchange.testing.performance.ajax.multiple.MultipleAPI
import com.openexchange.testing.performance.ajax.saml.SamlAPI
import com.openexchange.testing.performance.ajax.system.SystemAPI
import com.openexchange.testing.performance.common.Capabilities
import com.openexchange.testing.performance.common.SessionConstants._
import com.openexchange.testing.performance.common.utils.JSONUtil
import com.openexchange.testing.performance.config.Configuration
import com.typesafe.scalalogging.LazyLogging
import io.gatling.core.Predef._
import io.gatling.core.structure.ChainBuilder
import io.gatling.http.Predef._

/**
 * [[LoginChain]]
 *
 * @author <a href="mailto:nikolaos.tsapanidis@open-xchange.com">Nikolaos Tsapanidis</a>
 */
object LoginChain extends LazyLogging with Configuration {

  def doLogout(authentication: String = settings.middleware.authentication): ChainBuilder = {
    doIf(session => session.contains(SESSION_ID)) {
      exec(authentication match {
        case "saml-dummy" => exec(LoginAPI.logout("samlLogout"))
        case _            => exec(LoginAPI.logout())
      })
        .exec(flushSessionCookies)
        .exec(session => for {
          session_id <- session(SESSION_ID).validate[String]
          user <- session("username").validate[String]
          cid <- session(CONTEXT_ID).validate[Int]
        } yield {
          logger.trace("| user: {}\t| cid: {}\t| session: {} |", user, cid.toString, session_id)
          session
        })
    }
  }

  /**
   * Executes a login request for a specific client.
   *
   * @param authentication The authentication type.
   */
  def doLogin(authentication: String = settings.middleware.authentication): ChainBuilder = {
    exec(session => {
      session.set("modules", settings.appsuite.modules)
    })
      .exec(authentication match {
        case "standard" | "default" => exec(
          LoginAPI.login(queryParams = Option(Map(
            "language" -> settings.simulation.data.locale,
            "client" -> "open-xchange-appsuite",
            "timeout" -> 10000,
            "rampup" -> true,
            "rampUpFor" -> "open-xchange-appsuite"))))
        case "saml-dummy" => exec(SamlAPI.initLoginFlow())
          .doIf("${OX_SAML_AUTH_PARAMS.exists()}") {
            exec(SamlAPI.idpLogin())
              .exec(SamlAPI.authnResponse())
              .exec(SamlAPI.createSession())
              .exec(SamlAPI.cleanUpGatlingSession)
              .exec(AppsAPI.getManifestsConfig())
              .exec(LoginAPI.rampUp())
              .exec(SystemAPI.whoAmI())
          }
        case _ => exec(session => {
          logger.error("""Authentication "{}" not supported.""", authentication)
          session
        })
      })

      .doIf(session => session.contains(CAPABILITIES)) {
        exec(session => {
          var reqBuffer = new scala.collection.mutable.ListBuffer[Map[String, String]]()
          if (session(CAPABILITIES).as[Seq[String]].contains(Capabilities.webmail) &&
            settings.appsuite.modules.getOrElse(Capabilities.webmail, 0.0) > 0) {
            reqBuffer += Map("action" -> "mail", "module" -> "quota")
          }
          if (session(CAPABILITIES).as[Seq[String]].contains(Capabilities.infostore) &&
            settings.appsuite.modules.getOrElse(Capabilities.infostore, 0.0) > 0) {
            reqBuffer += Map("action" -> "filestore", "module" -> "quota")
          }
          session.set("req", JSONUtil().toJson(reqBuffer.toSeq))
        })
          // any reasonable payload will have at least, say, length 5
          // an empty payload (by experiment) is shown as "[]" and has a length of 2
          // e.g.: Debug: req=[{"action":"filestore","module":"quota"}], length=41
          .doIf(session => session.contains("req") && session("req").as[String].length > 5) {
            exec(MultipleAPI.multiple(true, "${req}"))
              .exec(session => {
                session.remove("req")
              })
          }
      }
      .doIf(session => session.contains(CAPABILITIES) &&
        session(CAPABILITIES).validate[Seq[String]].get.contains(Capabilities.mailfilterV2)) {
        exec(MailFilterAPI.getAllMailFilterRules(Some("vacation")))
          .exec(MailFilterAPI.getAllMailFilterRules(Some("autoforward")))
          .exec(MailFilterAPI.getMailFilterConfig())
      }

      .exec(session => for {
        user <- session("username").validate[String]
        cid <- session(CONTEXT_ID).validate[Int]
        caps <- session(CAPABILITIES).validate[Seq[String]]
      } yield {
        logger.trace("| user: {}\t| cid: {},\t| capabilities: {} |", user, cid.toString, caps.mkString("[", ", ", "]"))
        session
      })
  }

}