/*
 * Copyright (C) 2018 OX Software GmbH
 *
 * This file is part of OX API-Performance Tests.
 *
 * OX API-Performance Tests is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * OX API-Performance Tests is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OX API-Performance Tests.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.openexchange.testing.performance.ajax.mail

import java.text.SimpleDateFormat
import java.util.{ Date, Locale }

import com.openexchange.testing.performance.common.ModuleNames
import com.openexchange.testing.performance.common.utils.JSONUtil
import com.openexchange.testing.performance.groupware.mail.Mail
import io.gatling.core.Predef._
import io.gatling.core.session.Expression
import io.gatling.core.structure.ChainBuilder
import io.gatling.http.Predef._

/**
 * [[MailAPI]]
 *
 * @author <a href="mailto:nikolaos.tsapanidis@open-xchange.com">Nikolaos Tsapanidis</a>
 */
object MailAPI {

  /**
   * Clears the content of mail folders.
   *
   * @return
   */
  def clearMailFolders = ???

  /**
   * Copies a mail to another folder.
   *
   * @param objectId
   * @param srcFolder
   * @param dstFolderId
   * @return
   */
  def copyMail(objectId: String, srcFolder: String, dstFolderId: String): ChainBuilder = group(ModuleNames.MAIL) {
    exec(new CopyMail(objectId, srcFolder, JSONUtil().toJson(Map("folder_id" -> dstFolderId))))
  }

  /**
   * Deletes mails.
   *
   * @param timestamp
   * @param requestBody
   * @return
   */
  def deleteMails(timestamp: Any, requestBody: String): ChainBuilder = group(ModuleNames.MAIL) {
    exec(new DeleteMails(timestamp, requestBody))
  }

  /**
   * Examines a specifed mail folder and provides its basic status information.
   *
   * @return
   */
  def examine = ???

  /**
   * Forwards a mail.
   *
   * @return
   */
  def forwardMail = ???

  /**
   * Gets all mails.
   */
  def getAllMails(folder: String, columns: String, headers: Option[String] = None, unseen: Option[Boolean] = None,
                  deleted: Option[Boolean] = None, sort: Option[String] = None, order: Option[String] = None,
                  left_hand_limit: Option[Int] = None, right_hand_limit: Option[Int] = None,
                  limit: Option[String] = None, categoryid: Option[String] = None): ChainBuilder = {
    group(ModuleNames.MAIL) {
      exec(new GetAllMails(folder, columns, Option(Map(
        "headers" -> headers.getOrElse(None), "unseen" -> unseen.getOrElse(None), "deleted" -> deleted.getOrElse(None),
        "sort" -> sort.getOrElse(None), "order" -> order.getOrElse(None),
        "left_hand_limit" -> left_hand_limit.getOrElse(None), "right_hand_limit" -> right_hand_limit.getOrElse(None),
        "limit" -> limit.getOrElse(None), "categoryid" -> categoryid.getOrElse(None)))).buildHttpRequest()
        .check(bodyString.transform(x => JSONUtil().fromJson[GetAllMailsResponse](x).data)
          .optional.saveAs("getAllMailsResponse")))
    }
  }

  /**
   * Gets multiple mail attachments as a ZIP file.
   *
   * @return
   */
  def getAttachmentsAsZip = ???

  /**
   * Gets a mail.
   *
   * @param id
   * @param folder
   * @param optionalQueryParams
   * @return
   */
  def getMail(id: String, folder: String, optionalQueryParams: Option[Map[String, Any]] = None): ChainBuilder = group(ModuleNames.MAIL) {
    exec(new GetMail(folder, id).buildHttpRequest()
      .queryParamMap(optionalQueryParams.getOrElse(Map.empty))
      .check(jsonPath("$.data.security_info.encrypted").ofType[Boolean].optional.saveAs("MAIL_SECURITY_INFO")))
  }

  /**
   * Gets a mail attachment.
   *
   * @return
   */
  def getMailAttachment = ???

  /**
   * Gets all mail conversations.
   *
   * @return
   */
  def getMailConversations = ???

  /**
   * Gets the mail count.
   *
   * @return
   */
  def getMailCount = ???

  /**
   * Gets the message headers as plain text.
   *
   * @return
   */
  def getMailHeaders = ???

  /**
   * Gets a list of mails.
   *
   * @param columns
   * @param objectIds
   * @return
   */
  def getMailList(columns: String, objectIds: Expression[Seq[Map[String, Any]]]): ChainBuilder = group(ModuleNames.MAIL) {
    exec(new ListMails(columns, objectIds))
  }

  /**
   * Gets the complete message source as plain text.
   *
   * @return
   */
  def getMailSource = ???

  /**
   * Gets updated mails (not IMAP).
   *
   * @return
   */
  def getMailUpdates = ???

  /**
   * Gets multiple mails as a ZIP file.
   *
   * @return
   */
  def getMailsAsZip = ???

  /**
   * Import of mails as MIME data block ([RFC822](https://tools.ietf.org/html/rfc822)).
   *
   * @param folder
   * @param fileName
   * @param optionalQueryParams
   * @return
   */
  def importMail(folder: String, fileName: String, optionalQueryParams: Option[Map[String, Any]] = None): ChainBuilder = group(ModuleNames.MAIL) {
    exec(session => {
      session.setAll("UUID" -> java.util.UUID.randomUUID().toString, "MAIL_DATE" -> new SimpleDateFormat("E, d MMM yyyy HH:mm:ss Z", Locale.US).format(new Date()))
    })
      .exec(new ImportMails(folder, fileName).buildHttpRequest()
        .queryParamMap(optionalQueryParams.getOrElse(Map.empty)))

      // clean up
      .exec(session => {
        session.removeAll("UUID", "MAIL_DATE")
      })
  }

  /**
   * Marks all mails of a folder as seen.
   *
   * @return
   */
  def markAllMailsAsRead = ???

  /**
   * Stores a mail attachment to the drive.
   *
   * @return
   */
  def putMailAttachment = ???

  /**
   * Requests a delivery receipt for a priviously sent mail.
   *
   * @return
   */
  def receiptMailAck = ???

  /**
   * Replies a mail to all.
   *
   * @return
   */
  def replyAllMail = ???

  /**
   * Replies a mail.
   *
   * @return
   */
  def replyMail = ???

  /**
   * Resolves a given share reference.
   *
   * @return
   */
  def resolveShareReference = ???

  /**
   * Searches for mails.
   *
   * @return
   */
  def searchMails = ???

  /**
   * Sends a mail.
   *
   * @param mail
   * @return
   */
  def sendMail(mail: Expression[Mail]): ChainBuilder = group(ModuleNames.MAIL) {
    exec(new SendMail(mail).buildHttpRequest()
      .check(regex("data.*?(default[0-9]+.*[0-9])").saveAs("sendMailResponse")))
  }

  /**
   * Sends or saves a mail as MIME data block ([RFC822](https://tools.ietf.org/html/rfc822)).
   *
   * @return
   */
  def sendOrSaveMail = ???

  /**
   * Updates a mail or a folder's messages and/or moves a mail to another folder.
   *
   * @param srcFolder
   * @param dstFolderId
   * @param optionalQueryParams
   * @return
   */
  def updateMail(srcFolder: String, dstFolderId: String, optionalQueryParams: Option[Map[String, Any]] = None): ChainBuilder =
    group(ModuleNames.MAIL) {
      exec(new UpdateMail(srcFolder, JSONUtil().toJson(Map("folder_id" -> dstFolderId)))
        .buildHttpRequest()
        .queryParamMap(optionalQueryParams.getOrElse(Map.empty)))
    }

  //TODO not documented?!
  def getAllMailConversations(folder: String, columns: String, optionalQueryParams: Option[Map[String, Any]] = None) =
    group(ModuleNames.MAIL) {
      exec(new GetAllMailConversations(folder, columns)
        .buildHttpRequest()
        .queryParamMap(optionalQueryParams.getOrElse(Map.empty)))
    }

}
