/*
 * Copyright (C) 2018 OX Software GmbH
 *
 * This file is part of OX API-Performance Tests.
 *
 * OX API-Performance Tests is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * OX API-Performance Tests is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OX API-Performance Tests.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.openexchange.testing.performance.ajax.tasks

import java.util.concurrent.ThreadLocalRandom

import com.openexchange.testing.performance.common.Capabilities
import com.openexchange.testing.performance.common.SessionConstants._
import com.openexchange.testing.performance.common.utils.{ JSONUtil, TaskGenerator }
import com.openexchange.testing.performance.config.Configuration
import com.openexchange.testing.performance.groupware.Task
import com.typesafe.scalalogging.LazyLogging
import io.gatling.core.Predef._
import io.gatling.core.session.Expression
import io.gatling.core.structure.ChainBuilder

/**
 * [[TaskChain]]
 *
 * @author <a href="mailto:nikolaos.tsapanidis@open-xchange.com">Nikolaos Tsapanidis</a>
 */
object TaskChain extends LazyLogging with Configuration {
  private val random = ThreadLocalRandom.current()
  private val taskDataClassName = classOf[Task].getSimpleName
  private val createTaskClassName = classOf[CreateTask].getSimpleName

  val taskChain: ChainBuilder = randomSwitch(
    50d -> createChain($(FID_TASKS)),
    50d -> deleteChain())

  def createChain(folderId: Expression[String]): ChainBuilder = {
    doIf(session => session(CAPABILITIES).as[Seq[String]].contains(Capabilities.tasks) &&
      settings.appsuite.modules.contains(Capabilities.tasks)) {
      exec(TasksAPI.getAllTasks($(FID_TASKS), "1,5,20,101,200,203,220,300,301,317,401", Some(Map(
        "timezone" -> "UTC", "sort" -> "317", "order" -> "asc", "cache" -> true))))
        .exec(session => for {
          fid <- folderId(session)
        } yield {
          val taskData = TaskGenerator.generateRandomTask(fid)
          logger.trace("{}", taskData)
          session.set(taskDataClassName, taskData)
        })
        .doIf(session => session.contains(taskDataClassName)) {
          exec(TasksAPI.createTask("${" + taskDataClassName + "}"))
            .doIfOrElse(session => session.contains(createTaskClassName)) {
              exec(exec(session => for {
                id <- session(createTaskClassName).validate[Int]
              } yield {
                logger.trace("Successfully created task with id {}", id)
                session
              }))
            } {
              exec(session => {
                logger.warn("{} failed.", createTaskClassName)
                session
              })
            }
        }
    }
  }

  def bareCreateChain(folderId: Expression[String]): ChainBuilder = {
    doIf(session => session(CAPABILITIES).as[Seq[String]].contains(Capabilities.tasks) &&
      settings.appsuite.modules.contains(Capabilities.tasks)) {
      exec(session => for {
        fid <- folderId(session)
      } yield {
        val taskData = TaskGenerator.generateRandomTask(fid)
        logger.trace("{}", taskData)
        session.set(taskDataClassName, taskData)
      })
        .doIf(session => session.contains(taskDataClassName)) {
          exec(TasksAPI.createTask("${" + taskDataClassName + "}"))
            .doIfOrElse(session => session.contains(createTaskClassName)) {
              exec(exec(session => for {
                id <- session(createTaskClassName).validate[Int]
              } yield {
                logger.trace("Successfully created task with id {}", id)
                session
              }))
            } {
              exec(session => {
                logger.warn("{} failed.", createTaskClassName)
                session
              })
            }
        }
    }
  }

  def deleteChain(): ChainBuilder = {
    doIf(session => session(CAPABILITIES).as[Seq[String]].contains(Capabilities.tasks) &&
      settings.appsuite.modules.contains(Capabilities.tasks)) {
      exec(TasksAPI.getAllTasks($(FID_TASKS), "1,5,20,101,200,203,220,300,301,317,401", Some(Map(
        "timezone" -> "UTC", "sort" -> "317", "order" -> "asc", "cache" -> true))))

        .doIf(session => session.contains("getAllTasksResponse") &&
          session("getAllTasksResponse").as[GetAllTasksResponse].data.nonEmpty) {
          exec(session => for {
            idx <- ThreadLocalRandom.current().nextInt(session("getAllTasksResponse").as[GetAllTasksResponse].data.size)
            lastUpdate <- session("getAllTasksResponse").as[GetAllTasksResponse].timestamp
            randomElement = session("getAllTasksResponse").as[GetAllTasksResponse].data(idx)
            objectId <- randomElement.head
            folderId <- randomElement(2)
          } yield {
            session.setAll("objectId" -> objectId, "folderId" -> folderId, "lastUpdate" -> lastUpdate,
              "requestBody" -> JSONUtil().toJson(Seq(Map("id" -> objectId, "folder" -> folderId))))
          })
            .exec(TasksAPI.getTask("${objectId}", "${folderId}"))
            .exec(TasksAPI.deleteTasks("${lastUpdate}", "${requestBody}"))
            .exec(session => for {
              folderId <- session("folderId").validate[String]
              objectId <- session("objectId").validate[String]
            } yield {
              logger.debug("Successfully deleted task in folder {} with id {}", folderId, objectId)
              session.removeAll("objectId", "folderId", "lastUpdate", "getAllTasksResponse", "getTask", "requestBody")
            })
        }

    }
  }

  def updateChain(): ChainBuilder = ???

  def cleanUpSession(): ChainBuilder = exec(session => {
    session.removeAll(createTaskClassName, taskDataClassName)
    // TODO session.removeAll("objectId", "folderId", "lastUpdate", "getAllTasksResponse", "getTask", "requestBody")
  })

}