/*
 * Copyright (C) 2018 OX Software GmbH
 *
 * This file is part of OX API-Performance Tests.
 *
 * OX API-Performance Tests is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * OX API-Performance Tests is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OX API-Performance Tests.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.openexchange.testing.performance.simulations

import com.openexchange.testing.performance.ajax.config.ConfigAPI
import com.openexchange.testing.performance.ajax.folders.{ FoldersAPI, GetSubfoldersResponse }
import com.openexchange.testing.performance.ajax.login.LoginChain
import com.openexchange.testing.performance.common.SessionConstants._
import com.openexchange.testing.performance.config.{ Configuration, HTTPConfig }
import com.typesafe.scalalogging.LazyLogging
import io.gatling.core.Predef._
import io.gatling.core.structure.ChainBuilder

/**
 * [[CleanUpSimulation]]
 *
 * @author <a href="mailto:nikolaos.tsapanidis@open-xchange.com">Nikolaos Tsapanidis</a>
 */
class CleanUpSimulation extends Simulation with Configuration with LazyLogging {

  val feederSize = {
    csv(settings.simulation.userFeeder).records.size
  }

  val scn = scenario(getClass.getSimpleName)
    .repeat((_: Session) => feederSize / 10) {
      feed(csv(settings.simulation.userFeeder))
        .exec(LoginChain.doLogin())

        .doIfOrElse(session => session.contains(SESSION_ID)) {
          foreach(settings.appsuite.modules.keys.toList, "module") {
            doSwitch("${module}")(
              "webmail" -> doIf(session => session(CAPABILITIES).as[Seq[String]].contains("webmail")) {
                exec(
                  Mail.getAllDefaultMailFolderIds,
                  Mail.deleteFolderChain($(FID_INBOX)),
                  Mail.deleteFolderChain($(FID_TRASH)),
                  Mail.clearDefaultFoldersChain)
              },
              "calendar" -> doIf(session => session(CAPABILITIES).as[Seq[String]].contains("calendar")) {
                exec(ConfigAPI.getDefaultFolderCalendar, FoldersAPI.clearFolders(Array($(FID_CALENDAR))))
              },
              "contacts" -> doIf(session => session(CAPABILITIES).as[Seq[String]].contains("contacts")) {
                exec(ConfigAPI.getDefaultFolderContacts, FoldersAPI.clearFolders(Array($(FID_CONTACTS))))
              },
              "infostore" -> doIf(session => session(CAPABILITIES).as[Seq[String]].contains("infostore")) {
                exec(
                  ConfigAPI.getDefaultFolderInfostore,
                  ConfigAPI.getDefaultFolderInfostoreTrash,
                  FoldersAPI.clearFolders(Array($(FID_INFOSTORE), $(FID_INFOSTORE_TRASH))))
              },
              "tasks" -> doIf(session => session(CAPABILITIES).as[Seq[String]].contains("tasks")) {
                exec(ConfigAPI.getDefaultFolderTasks, FoldersAPI.clearFolders(Array($(FID_TASKS))))
              })
          }
            .exec(LoginChain.doLogout())
        } {
          exec(session => {
            logger.error("Could not get session key from {}", settings.middleware.baseUrl)
            session
          })
        }
    }

  setUp(scn.inject(atOnceUsers(settings.simulation.injection.users)))
    .protocols(HTTPConfig.appsuiteConf
      .acceptCharsetHeader("application/json, text/javascript, */*; q=0.01"))
    .assertions(global.successfulRequests.percent.gte(settings.simulation.assertions.successfulRequests))

}

object Mail {
  val getAllDefaultMailFolderIds: ChainBuilder = {
    exec(ConfigAPI.getAllDefaultMailFolderIds)
  }

  val clearDefaultFoldersChain: ChainBuilder = {
    tryMax(3) {
      exec(FoldersAPI.clearFolders(Array($(FID_INBOX), $(FID_SENT), $(FID_DRAFTS), $(FID_TRASH))))
    }
  }

  def deleteFolderChain(folder: String): ChainBuilder = {
    exec(FoldersAPI.getSubFolders(folder, "1"))

      .doIf(session => session.contains("getSubfoldersResponse") && session("getSubfoldersResponse").as[GetSubfoldersResponse].data.nonEmpty) {
        exec(session => for {
          response <- session("getSubfoldersResponse").validate[GetSubfoldersResponse]
        } yield {
          session.set("folders2delete", response.data.flatten.filter(_.toString.toLowerCase.contains("oxgatling-")))
        })

          .doIf(session => session.contains("folders2delete")) {
            foreach("${folders2delete}", "id") {
              tryMax(3) {
                exec(FoldersAPI.deleteFolder("${id}", 0))
              }
            }
              .exec(session => {
                session.removeAll("folders2delete", "getSubfoldersResponse")
              })
          }
      }
  }
}