/*
 * Copyright (C) 2018 OX Software GmbH
 *
 * This file is part of OX API-Performance Tests.
 *
 * OX API-Performance Tests is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * OX API-Performance Tests is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OX API-Performance Tests.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.openexchange.testing.performance.simulations

import java.util.concurrent.{ ThreadLocalRandom, TimeUnit }

import com.openexchange.testing.performance.ajax.chronos.ChronosChain
import com.openexchange.testing.performance.ajax.config.ConfigAPI
import com.openexchange.testing.performance.ajax.contacts.ContactChain
import com.openexchange.testing.performance.ajax.login.LoginChain
import com.openexchange.testing.performance.ajax.mail.MailAPI
import com.openexchange.testing.performance.ajax.tasks.TaskChain
import com.openexchange.testing.performance.common.SessionConstants._
import com.openexchange.testing.performance.config.{ Configuration, HTTPConfig }
import com.typesafe.scalalogging.LazyLogging
import io.gatling.core.Predef._
import io.gatling.core.session.el._
import io.gatling.core.structure.ScenarioBuilder

import scala.concurrent.duration._

/**
 * [[DataGenSimulation]]
 *
 * Fill accounts with random data so that reasonable simulations can run later on non-empty users
 *
 * One user will be filled with a configurable(*) amount of PIM items. These are created in random
 * order to hit the platform subsystems in a most equalized manner. Each of these users is configured such that
 * it will create one object per second.
 *
 * We assume the platform has a performance capacity, i.e. a number of parallel users which can be filled
 * simultaneously. By experience as of writing this text one (ox node, database instance) can handle some 200
 * of these users. Since the number of accounts to be filled is typically much larger, an overall repeat loop
 * is put around this, so the total number of filled accounts will be in the end
 *
 * settings.simulation.injection.users * settings.simulation.injection.repeatCounter
 *
 * (*) configurable: TODO. For the moment adjust the code for the itemsToCreate instantiation below
 *
 * @author <a href="mailto:nikolaos.tsapanidis@open-xchange.com">Nikolaos Tsapanidis</a>
 * @author <a href="mailto:dominik.epple@open-xchange.com">Dominik Epple</a>
 */
class DataGenSimulation extends Simulation with Configuration with LazyLogging {
  private val mail: String = settings.simulation.data.locale match {
    case "ja" => "lorem_jp"
    case _    => "lorem"
  }

  val scn: ScenarioBuilder = scenario(getClass.getSimpleName)
    .repeat(settings.simulation.injection.repeatCounter) {
      feed(csv(settings.simulation.userFeeder))
        .exec(LoginChain.doLogin())
        .exec(session => {
          println("logged in: " + session("username").as[String])
          session
        })
        .exec(session => {
          val itemsToCreate = Array(
            settings.simulation.data.mail.numberOfMails.getOrElse(100),
            settings.simulation.data.contacts.numberOfContacts.getOrElse(100),
            settings.simulation.data.calendar.numberOfAppointments.getOrElse(195),
            settings.simulation.data.calendar.numberOfRecurringAppointments.getOrElse(5),
            settings.simulation.data.tasks.numberOfTasks.getOrElse(10))
          session.set("itemsToCreate", itemsToCreate)
        })
        .asLongAs(session => session("itemsToCreate").as[Array[Int]].sum > 0) {
          pace(new FiniteDuration(1000, TimeUnit.MILLISECONDS))
            .exec(session => {
              val itemsToCreate = session("itemsToCreate").as[Array[Int]]
              val itemToCreate = takeOne(itemsToCreate)
              session.setAll("itemsToCreate" -> itemsToCreate, "itemToCreate" -> itemToCreate)
            })
            // TODO evaluate whether we can use more stripped-down chains, who do not read all the stuff, just write it
            .doSwitch("${itemToCreate}")(
              0 -> // create a mail
                exec(ConfigAPI.getAllDefaultMailFolderIds).exec(MailAPI.importMail($(FID_INBOX), s"$mail.eml")),
              1 -> // create contact
                // why the ContactChain not wrapped in exec() in the original code?
                exec(ConfigAPI.getDefaultFolderContacts).exec(ContactChain.bareCreateChain($(FID_CONTACTS).el[String])),
              2 -> // create appt w.o. recurring
                exec(ChronosChain.bareCreateChain(recurring = false)),
              3 -> // create appt w. recurring
                exec(ChronosChain.bareCreateChain(recurring = true)),
              4 -> //create task
                exec(ConfigAPI.getDefaultFolderTasks).exec(TaskChain.bareCreateChain($(FID_TASKS).el[String])))
        }
        .exec(session => {
          session.removeAll("itemsToCreate", "itemToCreate")
        })
        .exec(LoginChain.doLogout())
    }

  setUp(scn.inject(
    rampUsers(settings.simulation.injection.users) over new FiniteDuration((settings.simulation.injection.users + settings.simulation.injection.usersPerSec - 1) / settings.simulation.injection.usersPerSec, TimeUnit.SECONDS)))
    .protocols(HTTPConfig.appsuiteConf)

  // l to be an array with some elements
  // e.g. Array(100, 195, 5, 10)
  // which represents (generate 100 mails, 195 appointments w.o. recurring, 5 appointments with recurring, 10 contacts)
  // take one random element of this list and return it
  // and modify the list to look like e.g. (99, 195, 5, 10) if returned value is 1
  private def takeOne(l: Array[Int]): Int = {
    val i = ThreadLocalRandom.current().nextInt(l.sum)
    var sum = 0
    var j = 0
    while (j < l.length && i >= sum + l(j)) {
      sum = sum + l(j)
      j = j + 1
    }
    l(j) = l(j) - 1
    j
  }
}
