/*
 * Copyright (C) 2018 OX Software GmbH
 *
 * This file is part of OX API-Performance Tests.
 *
 * OX API-Performance Tests is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * OX API-Performance Tests is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OX API-Performance Tests.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.openexchange.testing.performance.simulations

import java.time.Instant
import java.time.temporal.ChronoUnit
import java.util.Locale
import java.util.concurrent.ThreadLocalRandom

import com.openexchange.testing.performance.ajax.attachments.AttachmentsAPI
import com.openexchange.testing.performance.ajax.calendar.{ CalendarAPI, CalendarChain, CreateAppointment, GetAllAppointmentsResponse }
import com.openexchange.testing.performance.ajax.contacts.{ ContactChain, ContactsAPI, CreateContact, GetAllContactsResponse }
import com.openexchange.testing.performance.ajax.folders.FoldersAPI
import com.openexchange.testing.performance.ajax.guardjson.GuardJSONAPI
import com.openexchange.testing.performance.ajax.halo.HaloAPI
import com.openexchange.testing.performance.ajax.infostore.{ GetAllInfoItemsResponse, InfostoreAPI, InfostoreChain }
import com.openexchange.testing.performance.ajax.jslob.JSlobAPI
import com.openexchange.testing.performance.ajax.mail.MailAPI
import com.openexchange.testing.performance.ajax.multiple.MultipleAPI
import com.openexchange.testing.performance.ajax.oauth.OAuthAPI
import com.openexchange.testing.performance.ajax.reminder.ReminderAPI
import com.openexchange.testing.performance.ajax.snippet.SnippetAPI
import com.openexchange.testing.performance.ajax.tasks.{ CreateTask, GetAllTasksResponse, TaskChain, TasksAPI }
import com.openexchange.testing.performance.common.Capabilities._
import com.openexchange.testing.performance.common.SessionConstants._
import com.openexchange.testing.performance.common.utils.DateTimeHelper._
import com.openexchange.testing.performance.common.utils._
import com.openexchange.testing.performance.config.Configuration
import com.openexchange.testing.performance.groupware.contacts.ContactData
import com.openexchange.testing.performance.groupware.mail.MailSecurity
import com.openexchange.testing.performance.groupware.{ Appointment, Module, Task }
import com.typesafe.scalalogging.LazyLogging
import io.gatling.core.Predef._
import io.gatling.core.session.Expression
import io.gatling.core.session.el._
import io.gatling.core.structure.ChainBuilder

import scala.util.Random

/**
 * [[OXAppSuiteChains]]
 *
 * @author <a href="mailto:nikolaos.tsapanidis@open-xchange.com">Nikolaos Tsapanidis</a>
 */
object OXAppSuiteChains extends LazyLogging with Configuration {

  // default locale of the JVM
  private val locale = Locale.getDefault

  private val timezone = "UTC"

  private val col1 = "1,2,3,4,5,6,20,23,300,301,302,304,305,306,307,308,309,310,311,312,313,314,315,316,317,3010,3020,3030,3050"

  private val pagination = Seq(0, 30)

  private val continue = true

  //================================================================================
  // Portal chain
  //================================================================================

  val portalChain = {
    exec(session => {
      logger.debug("Executing portal chain...")
      session
    })

      .exec(JSlobAPI.getJSLobPortal)
      .exec(JSlobAPI.getJSLobDriveClients)

      .exec(OAuthAPI.getAllOAuthServices)
      .exec(OAuthAPI.getAllOAuthAccounts)

      .exec(MultipleAPI.multiple(true, JSONUtil().toJson(Seq(
        Map("action" -> "all", "module" -> "fileservice"), Map("action" -> "all", "module" -> "fileaccount")))))

      .exec(CalendarAPI.getAllAppointments(
        "1,20,5,101,102,206,207,201,200,202,400,401,402,221,224,227,2,209,212,213,214,215,222,216,220",
        Instant.now().toEpochMilli,
        DateTimeHelper.forOrBackward(1, ChronoUnit.MONTHS),
        Some(Map(
          "showPrivate" -> true,
          "recurrence_master" -> false,
          "sort" -> 201,
          "order" -> "asc",
          "timezone" -> "UTC"))))

      // TODO
      // .exec(TasksAPI.searchTasks("1,2,20,101,200,203,220,221,300,301,309,317", SearchTasks("", end = Some(Tools.todayNow().getMillis))))
      .exec(ContactsAPI.searchContactsByBirthday(
        "1,20,500,501,502,503,504,505,511",
        DateTimeHelper.now().startOfDay(),
        DateTimeHelper.forOrBackward(3, ChronoUnit.MONTHS).startOfDay(),
        Some(Map("timezone" -> timezone, "right_hand_limit" -> 25))))
      .exec(ReminderAPI.getReminderRange(end = Some(DateTimeHelper.forOrBackward(1, ChronoUnit.HOURS).getTimestamp)))
      .exec(CalendarAPI.getUpdatedAppointments(
        "1,20,101,102,206,207,201,200,202,400,401,402,221,224,227,2,209,212,213,214,215,222,216,220",
        session => 0L, Some(Map(
          "start" -> DateTimeHelper.forOrBackward(2, ChronoUnit.HOURS).toAllDay,
          "end" -> DateTimeHelper.forOrBackward(5, ChronoUnit.YEARS).toAllDay,
          "showPrivate" -> true,
          "recurrence_master" -> true,
          "ignore" -> "deleted",
          "sort" -> 201,
          "order" -> "asc",
          "timezone" -> timezone))))

  }

  //================================================================================
  // Folder chains
  //================================================================================

  // TODO this chain has to be completely reviewed
  def createFolderChain(folderId: String, module: String): ChainBuilder = {
    exec(session => {
      logger.debug("Executing create folder chain...")
      session
    })
      .exec(session => {
        session.set("tmpFolderName", s"oxgatling-${Random.alphanumeric.take(8).mkString}")
      })

      .exec(FoldersAPI.createFolder("${OX_FID_INFOSTORE}", "infostore"))

      .doIf(session => session.contains("createFolderResponse")) {
        exec(session => for {
          id <- session("createFolderResponse").validate[String]
          tmpFolderName <- session("tmpFolderName").validate[String]
        } yield {
          logger.debug("Created {} folder {} with id {}", module, tmpFolderName, id)
          session.set("createdFolderId", id)
        })

          .exec(FoldersAPI.getFolder("${createdFolderId}", tree = 1, altNames = Some(true)))
          .exec(FoldersAPI.getAllVisibleFolder(
            module,
            col1,
            1))
          .exec(session => {
            session.removeAll("tmpFolderName", "createFolderResponse", "createdFolderId")
          })
      }
  }

  def deleteFolderChain(folderId: String, module: String) = {
    exec(session => {
      logger.debug("Executing delete folder chain...")
      session
    })

      .exec(FoldersAPI.getSubFolders(folderId, "1,2,3,4,5,6,20,300,301,302,304,305,306,307,308,309,310,311,312,313,314,315,316,3010,3020,3030"))

    //      .doIf(session => session.contains("allSubfoldersById") &&
    //        session("allSubfoldersById").as[Seq[String]].nonEmpty || session.contains("allSubfoldersByName") &&
    //        session("allSubfoldersByName").as[Seq[String]].nonEmpty) {
    //        exec(session => {
    //          val allSubfolderById = session("allSubfoldersById").as[Seq[String]].filter(x => x.toLowerCase.contains("oxgatling-"))
    //          val allSubfolderByName = session("allSubfoldersByName").as[Seq[String]].filter(x => x.toLowerCase.contains("oxgatling-"))
    //          logger.trace("ALL SUBFOIL {}", allSubfolderById)
    //
    //          if (allSubfolderById.nonEmpty) {
    //            session.set("allSubfolder", allSubfolderById.head)
    //          } else {
    //            session
    //          }
    //        })
    //      }

    //      .doIf(session => session.contains("allSubfolder") && session("allSubfolder").as[String].nonEmpty) {
    //        exec(AjaxAPI.deleteFolder("${allSubfolder}"))
    //          .exec(session => for {
    //            folderName <- session("allSubfolder").validate[String]
    //          } yield {
    //            logger.debug("Successfully deleted {} folder with id {}.", module, folderName)
    //            if (listOfCreatedObjects.contains(folderName)) {
    //              listOfCreatedObjects -= Tuple3(folderName, folderId, module) // TODO None?!
    //              logger.trace("Removed folder with id {} from {}", folderName, listOfCreatedObjects)
    //            }
    //            session.removeAll("allSubfolder", "getSubfoldersResponse")
    //          })
    //      }
  }

  //================================================================================
  // Calendar chains
  //================================================================================

  val calendarChain = {
    exec(session => {
      logger.debug("Executing calendar chain...")
      session
    })

      .doIf(session => session(CAPABILITIES).as[Seq[String]].contains("gab")) {
        exec(ContactsAPI.getAllUserFromGAB)
      }

      .randomSwitch(
        //    5d-> createFolderChain($(FID_CALENDAR), "calendar"), TODO
        //    5d-> deleteFolderChain($(FID_CALENDAR), "calendar"), TODO
        50d -> createAppointmentChain($(FID_CALENDAR).el[String], "calendar"),
        15d -> deleteAppointmentChain(FID_CALENDAR, "calendar"),
        25d -> updateAppointmentChain(FID_CALENDAR, "calendar"))

      .doIf(session => session(CAPABILITIES).as[Seq[String]].contains("gab")) {
        exec(session => {
          session.remove("ALL_GAB_USER")
        })
      }

  }

  def createAppointmentChain(folder: Expression[String], module: String) = {
    exec(CalendarAPI.getAllAppointments(
      "1,20,5,101,102,206,207,201,200,202,400,401,402,221,224,227,2,209,212,213,214,215,222,216,220",
      DateTimeHelper.forOrBackward(-2, ChronoUnit.MONTHS).startOfDay(),
      DateTimeHelper.forOrBackward(6, ChronoUnit.DAYS).startOfDay(),
      Some(Map("recurrence_master" -> false, "sort" -> 201, "order" -> "asc", "timezone" -> timezone, "folder" -> 0))))
      .exec(CalendarChain.createAppointment(folder))

      .doIf(session => session.contains(classOf[CreateAppointment].getSimpleName)) {
        exec(ReminderAPI.getReminderRange(end = Some(DateTimeHelper.forOrBackward(1, ChronoUnit.HOURS).getTimestamp)))
          .exec(CalendarAPI.hasAppointmentsOnDays(
            DateTimeHelper.forOrBackward(-4, ChronoUnit.WEEKS).startOfDay(),
            DateTimeHelper.forOrBackward(4, ChronoUnit.WEEKS).startOfDay()))
          .exec(CalendarAPI.getAllAppointments(
            "1,20,207,206,2",
            DateTimeHelper.now().startOfDay(),
            DateTimeHelper.forOrBackward(2, ChronoUnit.DAYS).startOfDay()))
      }
      .doIf(session => session.contains(classOf[CreateAppointment].getSimpleName)) {
        exec(session => for {
          folderId <- folder(session)
          id <- session(classOf[CreateAppointment].getSimpleName).validate[String]
        } yield {
          logger.debug("Successfully created appointment with id {}.", id)
          session.removeAll(classOf[CreateAppointment].getSimpleName, "classOf[CreateAppointment].getSimpleName", "getAllAppointmentsResponse")
        })
      }
  }

  def deleteAppointmentChain(folder: String, module: String) = {
    exec(CalendarAPI.getAllAppointments(
      "1,20,5,101,102,206,207,201,200,202,400,401,402,221,224,227,2,209,212,213,214,215,222,216,220",
      DateTimeHelper.forOrBackward(-2, ChronoUnit.MONTHS).startOfDay(),
      DateTimeHelper.forOrBackward(6, ChronoUnit.DAYS).startOfDay(),
      Some(Map("recurrence_master" -> false, "sort" -> 201, "order" -> "asc", "timezone" -> timezone, "folder" -> 0))))

      .doIf(session => session.contains("getAllAppointmentsResponse") &&
        session("getAllAppointmentsResponse").as[GetAllAppointmentsResponse].data.nonEmpty) {
        exec(session => for {
          idx <- ThreadLocalRandom.current().nextInt(session("getAllAppointmentsResponse").as[GetAllAppointmentsResponse].data.size)
          lastUpdate <- session("getAllAppointmentsResponse").as[GetAllAppointmentsResponse].timestamp
          randomElement = session("getAllAppointmentsResponse").as[GetAllAppointmentsResponse].data(idx)
          objectId <- randomElement.head
          folderId <- randomElement(1)
        } yield session.setAll("objectId" -> objectId, "folderId" -> folderId, "lastUpdate" -> lastUpdate))
          .exec(MultipleAPI.multiple(continue, JSONUtil().toJson(Seq(Map(
            "action" -> "delete",
            "module" -> module,
            "timestamp" -> "${lastUpdate}",
            "data" -> Map("id" -> "${objectId}", "folder" -> "${folderId}"))))))
          .exec(session => for {
            folderId <- session("folderId").validate[String]
            objectId <- session("objectId").validate[String]
          } yield {
            logger.debug("Successfully deleted appointment in folder {} with id {}", folderId, objectId)
            session.removeAll("objectId", "folderId", "lastUpdate", "getAllAppointmentsResponse")
          })
      }
  }

  def updateAppointmentChain(folder: String, module: String) = {
    exec(CalendarAPI.getAllAppointments(
      "1,20,5,101,102,206,207,201,200,202,400,401,402,221,224,227,2,209,212,213,214,215,222,216,220",
      DateTimeHelper.forOrBackward(-2, ChronoUnit.MONTHS).startOfDay(),
      DateTimeHelper.forOrBackward(6, ChronoUnit.DAYS).startOfDay(),
      Some(Map("recurrence_master" -> false, "sort" -> 201, "order" -> "asc", "timezone" -> timezone, "folder" -> 0))))

      .doIf(session => session.contains("getAllAppointmentsResponse") &&
        session("getAllAppointmentsResponse").as[GetAllAppointmentsResponse].data.nonEmpty) {
        exec(session => for {
          idx <- ThreadLocalRandom.current().nextInt(session("getAllAppointmentsResponse").as[GetAllAppointmentsResponse].data.size)
          lastUpdate <- session("getAllAppointmentsResponse").as[GetAllAppointmentsResponse].timestamp
          randomElement = session("getAllAppointmentsResponse").as[GetAllAppointmentsResponse].data(idx)
          objectId <- randomElement.head
          folderId <- randomElement(1)
        } yield session.setAll("objectId" -> objectId, "folderId" -> folderId, "lastUpdate" -> lastUpdate))
          .exec(CalendarAPI.getAppointment("${objectId}", "${folderId}"))
          .exec(AttachmentsAPI.getAllAttachments("${objectId}", "${folderId}", Module.Appointment.moduleType, "1,800,801,802,803,804,805"))

          .exec(session => for {
            appointment <- session("getAppointment").as[Appointment]
            appData <- appointment.task_appointment_data
            detData <- appointment.detailed_appointment_data
            newAppData <- appData.copy(note = Some(Random.alphanumeric.take(128).mkString))
            app <- Appointment(newAppData, detData)
          } yield session.set("updatedAppointment", app))

          .exec(CalendarAPI.updateAppointment("${objectId}", "${folderId}", "${lastUpdate}", "${updatedAppointment}"))
          .exec(ReminderAPI.getReminderRange(end = Some(DateTimeHelper.forOrBackward(1, ChronoUnit.HOURS).getTimestamp)))

          .exec(session => for {
            folderId <- session("folderId").validate[String]
            objectId <- session("objectId").validate[String]
          } yield {
            logger.debug("Successfully updated appointment in folder {} with id {}", folderId, objectId)
            session.removeAll("getAppointment", "updatedAppointment", "getAllAppointmentsResponse", "objectId", "folderId", "lastUpdate")
          })
      }
  }

  //================================================================================
  // Task chains
  //================================================================================

  val taskChain = {
    exec(session => {
      logger.debug("Executing task chain...")
      session
    })

      .randomSwitch(
        //        10d -> createFolderChain($(FID_TASKS), "tasks"), TODO
        //        15d -> deleteFolderChain($(FID_TASKS), "tasks"), TODO
        30d -> createTaskChain($(FID_TASKS).el[String], "tasks"),
        25d -> deleteTaskChain(FID_TASKS, "tasks"),
        20d -> updateTaskChain(FID_TASKS, "tasks"))

  }

  def createTaskChain(folder: Expression[String], module: String) = {
    exec(TasksAPI.getAllTasks($(FID_TASKS), "1,20,101,200,203,220,300,301,317", Some(
      Map("sort" -> "317", "timezone" -> timezone, "order" -> "asc"))))
      .exec(TaskChain.createChain(folder))

      .doIf(session => session.contains(classOf[CreateTask].getSimpleName)) {
        exec(session => for {
          objectId <- session(classOf[CreateTask].getSimpleName).validate[Int]
          folderId <- folder(session)
        } yield {
          session.setAll("taskArray" -> Array(Map("id" -> objectId, "folder" -> folderId)))
        })

          .exec(TasksAPI.getTaskList("1,2,20,101,200,203,220,221,300,301,309,317,401", "${taskArray}"))
          .exec(session => for {
            folderId <- folder(session)
            id <- session(classOf[CreateTask].getSimpleName).validate[Int]
          } yield {
            logger.debug("Successfully created task with id {}", id)
            session.removeAll("newTask", "createTaskResponse", "getAllTasksResponse", "taskArray")
          })
      }
  }

  def updateTaskChain(folder: String, module: String) = {
    exec(TasksAPI.getAllTasks($(FID_TASKS), "1,20,101,200,203,220,300,301,317", Some(
      Map("sort" -> "202", "timezone" -> timezone, "order" -> "asc"))))

      .doIf(session => session.contains("getAllTasksResponse") &&
        session("getAllTasksResponse").as[GetAllTasksResponse].data.nonEmpty) {
        exec(session => for {
          idx <- ThreadLocalRandom.current().nextInt(session("getAllTasksResponse").as[GetAllTasksResponse].data.size)
          lastUpdate <- session("getAllTasksResponse").as[GetAllTasksResponse].timestamp
          randomElement = session("getAllTasksResponse").as[GetAllTasksResponse].data(idx)
          objectId <- randomElement.head
          folderId <- randomElement(1)
        } yield session.setAll("objectId" -> objectId, "folderId" -> folderId, "lastUpdate" -> lastUpdate))

          .exec(AttachmentsAPI.getAllAttachments("${objectId}", "${folderId}", Module.Task.moduleType, "1,800,801,802,803,804,805"))
          .exec(TasksAPI.getTask("${objectId}", "${folderId}"))

          .exec(session => for {
            taskData <- session("getTask").as[Task].task_appointment_data
            detData <- session("getTask").as[Task].detailed_task_data
            newTaskData <- taskData.copy(note = Some(Random.alphanumeric.take(128).mkString))
            task <- Task(newTaskData, detData)
          } yield session.set("updatedTask", task))

          // TODO
          //          .exec(TasksAPI.updateTask("${objectId}", "${folderId}", "${lastUpdate}", "${updatedTask}"))
          .exec(session => for {
            folderId <- session("folderId").validate[String]
            objectId <- session("objectId").validate[String]
          } yield {
            logger.debug("Successfully updated task in folder {} with id {}", folderId, objectId)
            session.removeAll("getTask", "updatedTask", "getAllTasksResponse", "objectId", "folderId", "lastUpdate")
          })
      }

  }

  def deleteTaskChain(folder: String, module: String) = {
    // TODO removed column (sort) 317 since this only available for appsuite >= 7.8
    exec(TasksAPI.getAllTasks($(FID_TASKS), "1,20,101,200,203,220,300,301", Some(
      Map("sort" -> "202", "timezone" -> timezone, "order" -> "asc"))))

      .doIf(session => session.contains("getAllTasksResponse") &&
        session("getAllTasksResponse").as[GetAllTasksResponse].data.nonEmpty) {
        exec(session => for {
          idx <- ThreadLocalRandom.current().nextInt(session("getAllTasksResponse").as[GetAllTasksResponse].data.size)
          lastUpdate <- session("getAllTasksResponse").as[GetAllTasksResponse].timestamp
          randomElement = session("getAllTasksResponse").as[GetAllTasksResponse].data(idx)
          objectId <- randomElement.head
          folderId <- randomElement(1)
        } yield {
          session.setAll("objectId" -> objectId, "folderId" -> folderId, "lastUpdate" -> lastUpdate,
            "requestBody" -> JSONUtil().toJson(Seq(Map("id" -> objectId, "folder" -> folderId))))
        })
          .exec(TasksAPI.getTask("${objectId}", "${folderId}"))

          .exec(TasksAPI.deleteTasks("${lastUpdate}", "${requestBody}"))
          .exec(session => for {
            folderId <- session("folderId").validate[String]
            objectId <- session("objectId").validate[String]
          } yield {
            logger.debug("Successfully deleted task in folder {} with id {}", folderId, objectId)
            session.removeAll("objectId", "folderId", "lastUpdate", "getAllTasksResponse", "getTask", "requestBody")
          })
      }
  }

  //================================================================================
  // Contact chains
  //================================================================================

  val contactChain = {
    exec(session => {
      logger.debug("Executing contact chain...")
      session
    })

      .randomSwitch(
        //        10d -> createFolderChain($(FID_CONTACTS), "contacts"), TODO
        //        20d -> deleteFolderChain($(FID_CONTACTS), "contacts"), TODO
        40d -> createContactChain($(FID_CONTACTS), "contacts"),
        20d -> deleteContactChain($(FID_CONTACTS), "contacts"),
        10d -> updateContactChain($(FID_CONTACTS), "contacts"))
  }

  def createContactChain(folder: Expression[String], module: String) = {
    exec(ContactChain.createChain(folder))

      .doIf(session => session.contains(classOf[CreateContact].getSimpleName)) {
        exec(session => for {
          objectId <- session(classOf[CreateContact].getSimpleName).validate[String]
          folderId <- folder(session)
        } yield {
          session.setAll("objectId" -> objectId.toInt, "folderId" -> folderId.toInt)
        })

          .exec(ContactsAPI.getContact("${objectId}", "${folderId}"))
          .exec(ContactsAPI.getAllContacts("${folderId}", "1,20,101,607", Some(Map("timezone" -> timezone, "sort" -> 607, "order" -> "asc", "admin" -> false))))
          .exec(session => for {
            id <- session(classOf[CreateContact].getSimpleName).validate[Int]
            folderId <- folder(session)
          } yield {
            logger.debug("Successfully created contact with id {}", id)
            session.removeAll("newContact", classOf[CreateContact].getSimpleName, "objectId", "folderId", "getAllContactsResponse", "getContactResponse")
          })
      }
  }

  def deleteContactChain(folder: Expression[String], module: String) = {
    exec(ContactsAPI.getAllContacts(folder, "1,20,101,607", Some(Map(
      "timezone" -> timezone, "sort" -> 607, "order" -> "asc", "admin" -> false))))

      .doIf(session => session.contains("getAllContactsResponse") &&
        session("getAllContactsResponse").as[GetAllContactsResponse].data.nonEmpty) {
        exec(session => {
          val idx = ThreadLocalRandom.current().nextInt(session("getAllContactsResponse").as[GetAllContactsResponse].data.size)
          val randomElement = session("getAllContactsResponse").as[GetAllContactsResponse].data(idx)
          val objectId = randomElement.head
          val folderId = randomElement(1)
          val lastUpdate = session("getAllContactsResponse").as[GetAllContactsResponse].timestamp
          session.setAll("objectId" -> objectId, "folderId" -> folderId, "lastUpdate" -> lastUpdate)
        })
          .exec(ContactsAPI.deleteContacts("${lastUpdate}", "${folderId}", "${objectId}"))

          .exec(session => for {
            objectId <- session("objectId").validate[String]
            folderId <- session("folderId").validate[String]
          } yield {
            logger.debug("Successfully deleted contact in folder {} with id {}", folderId, objectId)
            session.removeAll("objectId", "folderId", "lastUpdate", "getAllContactsResponse")
          })
      }
  }

  def updateContactChain(folder: Expression[String], module: String) = {
    exec(ContactsAPI.getAllContacts(folder, "1,20,101,607", Some(Map(
      "timezone" -> timezone, "sort" -> "607", "order" -> "asc", "admin" -> false))))

      .doIf(session => session.contains("getAllContactsResponse") &&
        session("getAllContactsResponse").as[GetAllContactsResponse].data.nonEmpty) {
        exec(session => {
          val idx = ThreadLocalRandom.current().nextInt(session("getAllContactsResponse").as[GetAllContactsResponse].data.size)
          val randomElement = session("getAllContactsResponse").as[GetAllContactsResponse].data(idx)
          val objectId = randomElement.head
          val folderId = randomElement(1)
          val lastUpdate = session("getAllContactsResponse").as[GetAllContactsResponse].timestamp
          session.setAll("objectId" -> objectId, "folderId" -> folderId, "lastUpdate" -> lastUpdate)
        })
          .exec(AttachmentsAPI.getAllAttachments("${objectId}", "${folderId}", Module.Contact.moduleType, "1,800,801,802,803,804,805"))
          .exec(ContactsAPI.getContact("${objectId}", "${folderId}"))
          .exec(session => {
            val contactNew = new ContactData(folder_id = session("folderId").as[Int].toString, nickname = Some(Random.alphanumeric.take(16).mkString))
            session.set("updatedContact", contactNew)
          })
          .exec(ContactsAPI.updateContact("${objectId}", "${folderId}", "${lastUpdate}", "${updatedContact}"))

          .exec(session => for {
            objectId <- session("objectId").validate[String]
            folderId <- session("folderId").validate[String]
          } yield {
            logger.debug("Successfully updated contact in folder {} with id {}", folderId, objectId)
            session.removeAll("getContactResponse", "updatedContact", "getAllContactsResponse", "lastUpdate", "folderId", "objectId")
          })
      }
  }

  //================================================================================
  // Mail chains
  //================================================================================

  private val rndSwitchBuilder: Seq[(Double, ChainBuilder)] =
    if (settings.simulation.data.mail.conversationView.getOrElse(false)) {
      Seq(
        (85d, exec(
          MailAPI.getAllMails($(FID_INBOX), "102,600,601,602,603,604,605,606,607,608,610,611,614,652,656," +
            "X-Open-Xchange-Share-URL,662,664", sort = Some("610"), order = Some("desc"), deleted = Some(false), limit = Some("0,50")))),
        (15d, exec(
          MailAPI.getAllMailConversations($(FID_INBOX), "102,600,601,602,603,604,605,606,607,608,610,611,614,652,656,X-Open-Xchange-Share-URL,662,664",
            Some(Map(
              "sort" -> 610,
              "order" -> "desc",
              "deleted" -> true,
              "timezone" -> timezone,
              "limit" -> "0,50",
              "includeSent" -> true,
              "max" -> 300))))))
    } else {
      Seq((100d, exec(
        MailAPI.getAllMails($(FID_INBOX), "102,600,601,602,603,604,605,606,607,608,610,611,614,652,656," +
          "X-Open-Xchange-Share-URL,662,664", sort = Some("610"), order = Some("desc"), deleted = Some(false), limit = Some("0,50")))))
    }

  val mailChain: ChainBuilder = {
    exec(session => {
      logger.debug("Executing mail chain...")
      session
    })

      .randomSwitch(rndSwitchBuilder: _*)

      .exec(FoldersAPI.getFolder($(FID_INBOX), altNames = Some(true)))
      .exec(FoldersAPI.getFolderPath($(FID_INBOX), col1, altNames = Some(true)))

      .exec(FoldersAPI.getSubFolders("1", "1,2,3,4,5,6,20,23,300,301,302,304,305,306,307,308,309,310,311," +
        "312,313,314,315,316,317,318,319,3010,3020,3030,3050,3201,3203,3204,3205,3220",
        Some(Map("all" -> 0, "altNames" -> true, "timezone" -> timezone, "tree" -> 0))))

      .exec(FoldersAPI.getSubFolders($(FID_INBOX), "1,2,3,4,5,6,20,23,300,301,302,304,305,306,307,308,309,310,311," +
        "312,313,314,315,316,317,318,319,3010,3020,3030,3050,3201,3203,3204,3205,3220",
        Some(Map("all" -> 0, "altNames" -> true, "timezone" -> timezone, "tree" -> 0))))

      .doIf(session => session.contains("getAllMailsResponse") &&
        session("getAllMailsResponse").as[Seq[Seq[Any]]].nonEmpty) {
        exec(session =>
          for {
            list <- session("getAllMailsResponse").validate[Seq[Seq[Any]]]
          } yield {
            val mailBuffer = list.map(x => Map(
              "action" -> "get",
              "timezone" -> timezone,
              "embedded" -> true,
              "unseen" -> true,
              "id" -> x(1),
              "folder" -> x(2),
              "view" -> "html",
              "max_size" -> 102400,
              "module" -> "mail"))
            session.set("mailBuffer", mailBuffer)
          })

          // TODO workaround for multiple request...
          .exec(session => {
            session.set("multipleBody", JSONUtil().toJson(session("mailBuffer").as[Seq[Map[String, String]]]))
          })

          .exec(MultipleAPI.multiple(continue, "${multipleBody}"))

          .randomSwitch(
            15d -> readMailChain(),
            50d -> sendMailChain(),
            35d -> deleteMailChain())

          .exec(session => {
            session.removeAll("mailBuffer", "multipleBody", "getAllMailsResponse", "getFolderPathResponse", "getSubfoldersResponse")
          })
      }
  }

  def readMailChain(): ChainBuilder = {
    exec(MailAPI.getAllMails($(FID_INBOX), "102,600,601,602,603,604,605,606,607,608,610,611,614,652,656," +
      "X-Open-Xchange-Share-URL,662,664", deleted = Some(true), limit = Some("0,50"), order = Some("desc"), sort = Some("610")))

      .doIf("${getAllMailsResponse.exists()}") {
        doIf(session => session("getAllMailsResponse").as[List[List[Any]]].nonEmpty) {
          // pick a random email and save it in the session
          exec(session => for {
            data <- session("getAllMailsResponse").validate[List[List[Any]]]
          } yield session.set("RANDOM_MAIL", RandomHelper.getRandomElement(data)).remove("getAllMailsResponse"))

            .exec(MailAPI.getMail("${RANDOM_MAIL(1)}", "${RANDOM_MAIL(2)}", Some(
              Map(
                "embedded" -> false,
                "unseen" -> true,
                "sanitize" -> false,
                "view" -> "noimg",
                "process_plain_text" -> false,
                "timezone" -> timezone,
                "view" -> "html",
                "max_size" -> 102400))))
            // check if email is encrypted
            .doIfEqualsOrElse("${MAIL_SECURITY_INFO}", true) {
              exec(GuardJSONAPI.hasAuthToken())
                .exec(GuardJSONAPI.getAuthToken("${password}", -1))
                .exec(MailAPI.getMail("${RANDOM_MAIL(1)}", "${RANDOM_MAIL(2)}", Option(Map(
                  "timezone" -> timezone,
                  "embedded" -> false,
                  "sanitize" -> false,
                  "id" -> "${RANDOM_MAIL(1)}",
                  "decrypt" -> true,
                  "view" -> "noimg",
                  "max_size" -> 102400,
                  "process_plain_text" -> false,
                  "folder" -> "${RANDOM_MAIL(0)}"))))
            } {
              exec(MultipleAPI.multiple(true, JSONUtil().toJson(Seq(Map(
                "action" -> "update",
                "id" -> "${RANDOM_MAIL(1)}",
                "folder" -> "${RANDOM_MAIL(2)}",
                "timestamp" -> "2116800000000",
                "module" -> "mail",
                "data" -> Map(
                  "flags" -> 32,
                  "value" -> true,
                  "collect_addresses" -> true))))))
            }
            .exec(FoldersAPI.getFolder(id = $(FID_INBOX), altNames = Some(true)))
        }
      }

      .exec(session => {
        session.removeAll("MAIL_SECURITY_INFO", "RANDOM_MAIL")
      })
  }

  def sendMailChain(): ChainBuilder = {
    exec(SnippetAPI.getAllSnippets())
      .exec(MultipleAPI.multiple(continue, JSONUtil().toJson(Seq(Map(
        "action" -> "autocomplete",
        "query" -> Random.nextPrintableChar(),
        "admin" -> false,
        "email" -> true,
        "sort" -> 609,
        "columns" -> "1,2,5,20,101,500,501,502,505,520,524,555,556,557,569,592,602,606,607,551,552",
        "right_hand_limit" -> 0,
        "module" -> "contacts")))))

      .exec(HaloAPI.getHaloContactPicture(width = Some(54), height = Some(54), scaleType = Some("cover"),
        internal_userid = Some("${OX_USER_ID}"), context = Some("${OX_CID}"), user = Some("${OX_USER_ID}"),
        email = Some($(USER_EMAIL1))))

      .exec(HaloAPI.getHaloContactPicture(width = Some(16), height = Some(16), scaleType = Some("contain"),
        internal_userid = Some("${OX_USER_ID}"), context = Some("${OX_CID}"), user = Some("${OX_USER_ID}"),
        email = Some($(USER_EMAIL1))))

      .doIfOrElse(session => session(CAPABILITIES).as[Seq[String]].contains(guard_mail)) {
        randomSwitch(
          // Send mail
          70d ->
            exec(generateMail()),
          // Send signed mail
          5d ->
            exec(generateMail(Some(MailSecurity(false, false, true))))
            .doIf(session => session.contains("newMail")) {
              exec(GuardJSONAPI.getAuthToken(PASSWORD, -1))
            },
          // Send pgp mime mail
          10d ->
            exec(generateMail(Some(MailSecurity(true, false, false)))),
          // Send pgp mime signed mail
          10d ->
            exec(generateMail(Some(MailSecurity(true, false, true))))
            .doIf(session => session.contains("newMail")) {
              exec(GuardJSONAPI.getAuthToken(PASSWORD, -1))
            },
          // Send pgp inline mail
          2.5d ->
            exec(generateMail(Some(MailSecurity(true, true, false)))),
          // Send pgp inline signed mail
          2.5d ->
            exec(generateMail(Some(MailSecurity(true, true, true))))
            .doIf(session => session.contains("newMail")) {
              exec(GuardJSONAPI.getAuthToken(PASSWORD, -1))
            })
      } {
        exec(generateMail())
      }

      .doIf(session => session.contains("newMail")) {
        exec(MailAPI.sendMail("${newMail}"))

          .exec(MultipleAPI.multiple(true, JSONUtil().toJson(Seq(
            Map(
              "action" -> "get",
              "altNames" -> true,
              "id" -> $(FID_INBOX),
              "timezone" -> "UTC",
              "tree" -> 0,
              "module" -> "folders"),
            Map(
              "action" -> "get",
              "altNames" -> true,
              "id" -> $(FID_SENT),
              "timezone" -> "UTC",
              "tree" -> 0,
              "module" -> "folders"),
            Map(
              "action" -> "get",
              "altNames" -> true,
              "id" -> $(FID_DRAFTS),
              "timezone" -> "UTC",
              "tree" -> 0,
              "module" -> "folders")))))
          .exec(MailAPI.getAllMails($(FID_INBOX), "102,600,601,602,603,604,605,606,607,608,610,611,614,652,656," +
            "X-Open-Xchange-Share-URL,662,664", deleted = Some(true), order = Some("desc"), sort = Some("610"), limit =
            Some("0,50")))
          .exec(FoldersAPI.getFolder($(FID_SENT), altNames = Some(true)))
          .exec(FoldersAPI.getFolder($(FID_INBOX), altNames = Some(true)))
      }

      .doIf(session => session.contains("sendMailResponse")) {
        exec(session => for {
          id <- session("sendMailResponse").validate[String]
        } yield {
          logger.debug("Email successfully sent {}", id)
          session.removeAll("newMail", "sendMailResponse")
        })
      }
  }

  def deleteMailChain(): ChainBuilder = {
    exec(MailAPI.getAllMails($(FID_INBOX), "102,600,601,602,603,604,605,606,607,608,610,611,614,652",
      sort = Some("610"), order = Some("asc"), limit = Some(pagination.mkString(","))))

      .doIf(session => session.contains("getAllMailsResponse") &&
        session("getAllMailsResponse").as[List[List[Any]]].nonEmpty) {
        exec(session => for {
          idx <- ThreadLocalRandom.current().nextInt(session("getAllMailsResponse").as[List[List[Any]]].size)
          randomElement = session("getAllMailsResponse").as[List[List[Any]]].apply(idx)
          objectId <- randomElement(1)
          folderId <- randomElement(2)
          lastUpdate <- randomElement(10)
        } yield session.setAll("objectId" -> objectId, "folderId" -> folderId, "lastUpdate" -> lastUpdate))

          .exec(MailAPI.deleteMails("${lastUpdate}", JSONUtil().toJson(Seq(Map("id" -> "${objectId}", "folder" -> "${folderId}")))))

          .exec(session => for {
            folderId <- session("folderId").validate[String]
            objectId <- session("objectId").validate[String]
          } yield {
            logger.debug("Successfully deleted mail in folder {} with id {}", folderId, objectId)
            session.removeAll("objectId", "folderId", "lastUpdate", "getAllMailsResponse")
          })
      }
  }

  private def generateMail(security: Option[MailSecurity] = None) = {
    exec(session => for {
      firstName <- session(USER_FIRST_NAME).validate[String]
      lastName <- session(USER_LAST_NAME).validate[String]
      mailAddress <- session(USER_EMAIL1).validate[String]
      mail <- MailGenerator.generateAppSuiteMail(firstName, lastName, mailAddress, security)
    } yield {
      session.set("newMail", mail)
    })
  }

  //================================================================================
  // Drive chains
  //================================================================================

  val driveChain = {
    exec(FoldersAPI.getFolder($(FID_INFOSTORE), 1, altNames = Some(true)))

      .exec(FoldersAPI.getFolderPath($(FID_INFOSTORE), col1))
      .exec(FoldersAPI.getSubFolders($(FID_INFOSTORE), col1, Some(Map(
        "all" -> 0, "altNames" -> true, "timezone" -> timezone, "tree" -> 1))))

      .exec(MultipleAPI.multiple(continue, JSONUtil().toJson(Seq(
        Map(
          "action" -> "all",
          "module" -> "fileservice"),
        Map(
          "action" -> "all",
          "module" -> "fileaccount")))))

      // TODO removed column 108 since this only available for appsuite >= 7.8
      .exec(InfostoreAPI.getAllInfoItems($(FID_INFOSTORE), "1,2,3,5,20,23,700,702,703,704,705,707", Some(Map(
        "sort" -> 702,
        "order" -> "asc",
        "pregenerate_previews" -> true,
        "timezone" -> timezone))))

      .randomSwitch(
        30d -> createDriveChain($(FID_INFOSTORE), "drive"),
        50d -> previewDriveChain($(FID_INFOSTORE), "drive"),
        20d -> deleteDriveChain($(FID_INFOSTORE), "drive"))

      .exec(session => {
        session.removeAll("getFolderPathResponse", "getSubfoldersResponse", "getAllInfoitemsResponse")
      })
  }

  def createDriveChain(folder: String, module: String) = {
    doIfOrElse(session => session(CAPABILITIES).as[Seq[String]].contains(guard_drive)) {
      doIf(session => session.contains("newInfoitem")) {
        randomSwitch(
          100d -> exec(InfostoreChain.createInfoItemAdvanced(folder)))
        /* FIXME replace previous line with the following two once fixed.
          50d -> exec(InfostoreChain.createInfoItemAdvanced(folder)),
          50d -> exec(GuardFileAPI.uploadFile("${newInfoitemPath}", $(CONTEXT_ID), $(USER_ID), $(FID_INFOSTORE))))
          */
      }
    } {
      exec(InfostoreChain.createInfoItemAdvanced(folder))
    }

      .doIf(session => session.contains("createInfoitemResponse")) {
        exec(session => for {
          folderId <- session(folder).validate[String]
          id <- session("createInfoitemResponse").validate[String]
        } yield {
          logger.debug("Successfully created infoitem with id {}", id)
          session.removeAll("newInfoitem", "newInfoitemFilename", "newInfoitemMediaType", "newInfoitemPath", "createInfoitemResponse")
        })
      }
  }

  def deleteDriveChain(folder: String, module: String) = {
    doIf(session => session.contains("getAllInfoitemsResponse") &&
      session("getAllInfoitemsResponse").as[GetAllInfoItemsResponse].data.nonEmpty) {
      exec(session => for {
        idx <- ThreadLocalRandom.current().nextInt(session("getAllInfoitemsResponse").as[GetAllInfoItemsResponse].data.size)
        randomElement = session("getAllInfoitemsResponse").as[GetAllInfoItemsResponse].data(idx)
        objectId <- randomElement.head
        folderId <- randomElement(4)
      } yield session.setAll("objectId" -> objectId, "folderId" -> folderId))

        .exec(InfostoreAPI.deleteInfoItems(Instant.now.toEpochMilli, JSONUtil().toJson(
          Seq(Map("id" -> "${objectId}", "folder" -> "${folderId}"))), Some(Map("hardDelete" -> false))))

        .exec(session => for {
          folderId <- session("folderId").validate[String]
          objectId <- session("objectId").validate[String]
        } yield {
          logger.debug("Successfully deleted infoitem in folder {} with id {}", folderId, objectId)
          session.removeAll("objectId", "folderId")
        })
    }
  }

  def previewDriveChain(folder: String, module: String) = {
    exec(InfostoreAPI.getAllInfoItems($(FID_INFOSTORE), "1,2,3,5,20,23,700,702,703,704,705,707", Some(Map(
      "sort" -> 702,
      "order" -> "asc",
      "pregenerate_previews" -> true,
      "timezone" -> timezone))))
      .doIf(session => session.contains("getAllInfoitemsResponse") &&
        session("getAllInfoitemsResponse").as[GetAllInfoItemsResponse].data.nonEmpty) {
        exec(session => for {
          idx <- ThreadLocalRandom.current().nextInt(session("getAllInfoitemsResponse").as[GetAllInfoItemsResponse].data.size)
          randomElement = session("getAllInfoitemsResponse").as[GetAllInfoItemsResponse].data(idx)
          objectId <- randomElement.head
          folderId <- randomElement(4)
          metaData <- randomElement(5)
          fileName <- randomElement(6)
        } yield session.setAll("objectId" -> objectId, "folderId" -> folderId, "metaData" -> metaData, "fileName" -> fileName))

          .exec(InfostoreAPI.getInfoItem("${objectId}", "${folderId}"))
          .exec(InfostoreAPI.getInfoItemVersions("${objectId}", "1,2,3,4,5,20,23,100,102,108,109,700,701,702,703,704,705,706,707,708,709,710,711,7040"))
          .randomSwitch(
            50d -> exec(InfostoreAPI.getInfoItemDocument("${objectId}", "${folderId}", Some(Map("delivery" -> "view")))),
            50d -> exec(InfostoreAPI.downloadInfoItemDocument("${objectId}", "${folderId}", Some(Map("delivery" ->
              "download", "callback" -> "yell")))))

          .exec(session => {
            session.removeAll("objectId", "folderId", "metaData", "fileName")
          })

      }
  }

  def updateDriveChain() = ???
}
