/*
 * Copyright (C) 2018 OX Software GmbH
 *
 * This file is part of OX API-Performance Tests.
 *
 * OX API-Performance Tests is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * OX API-Performance Tests is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OX API-Performance Tests.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.openexchange.testing.performance.simulations

import java.time.temporal.ChronoUnit
import java.util.concurrent.ThreadLocalRandom

import com.openexchange.testing.performance.ajax.calendar.CalendarAPI
import com.openexchange.testing.performance.ajax.contacts._
import com.openexchange.testing.performance.ajax.folders.{ FoldersAPI, GetSubFolders }
import com.openexchange.testing.performance.ajax.halo.HaloAPI
import com.openexchange.testing.performance.ajax.jslob.JSlobAPI
import com.openexchange.testing.performance.ajax.mail.account.{ GetAllAccounts, MailAccountAPI }
import com.openexchange.testing.performance.ajax.mail.{ GetAllMailsResponse, MailAPI }
import com.openexchange.testing.performance.ajax.multiple.{ Multiple, MultipleAPI }
import com.openexchange.testing.performance.ajax.reminder.ReminderAPI
import com.openexchange.testing.performance.ajax.tasks._
import com.openexchange.testing.performance.ajax.user.UserAPI
import com.openexchange.testing.performance.common.ModuleNames
import com.openexchange.testing.performance.common.SessionConstants._
import com.openexchange.testing.performance.common.utils.DateTimeHelper._
import com.openexchange.testing.performance.common.utils._
import com.openexchange.testing.performance.config.Configuration
import com.openexchange.testing.performance.groupware._
import com.openexchange.testing.performance.groupware.contacts.ContactData
import com.openexchange.testing.performance.groupware.mail.{ MailAttachment, MailData }
import com.typesafe.scalalogging.LazyLogging
import io.gatling.core.Predef._
import io.gatling.http.Predef._

/**
 * [[PSOXAppSuiteChains]]
 *
 * @author <a href="mailto:dominik.epple@open-xchange.com">Dominik Epple</a>
 *
 *         Structure based on OXAppSuite chains. Individual chains created by doing stuff in App Suite WebUI and modelling it here.
 */
object PSOXAppSuiteChains extends LazyLogging with Configuration {

  val idleChain = {
    exec(session => {
      logger.debug("Executing idle chain...")
      session
    })

      .exec(
        MailAccountAPI.getAllAccounts(
          "1001,1002,1003,1004,1005,1006,1007,1008,1009,1010,1011,1012,1013,1014,1015,1016,1017,1018,1019,1020,1021,1022,1023,1024,1025,1026,1027,1028,1029,1030,1031,1032,1033,1034,1035,1036,1037,1038,1039,1040,1041,1042,1043,1044,1045,1046,1047,1048"))
      .exec(session => session.remove(classOf[GetAllAccounts].getSimpleName))

      .exec(
        MultipleAPI.multiple(
          true,
          "[{\"action\":\"list\",\"all\":\"0\",\"altNames\":true,\"parent\":\"default0\",\"timezone\":\"UTC\",\"tree\":0,\"columns\":\"1,2,3,4,5,6,20,23,300,301,302,304,305,306,307,308,309,310,311,312,313,314,315,316,317,318,3010,3020,3030,3050\",\"module\":\"folders\"},{\"action\":\"list\",\"all\":\"0\",\"altNames\":true,\"parent\":\"1\",\"timezone\":\"UTC\",\"tree\":0,\"columns\":\"1,2,3,4,5,6,20,23,300,301,302,304,305,306,307,308,309,310,311,312,313,314,315,316,317,318,3010,3020,3030,3050\",\"module\":\"folders\"},{\"action\":\"list\",\"all\":\"0\",\"altNames\":true,\"parent\":\"default0/INBOX\",\"timezone\":\"UTC\",\"tree\":0,\"columns\":\"1,2,3,4,5,6,20,23,300,301,302,304,305,306,307,308,309,310,311,312,313,314,315,316,317,318,3010,3020,3030,3050\",\"module\":\"folders\"},{\"action\":\"allVisible\",\"altNames\":true,\"content_type\":\"calendar\",\"timezone\":\"UTC\",\"tree\":1,\"columns\":\"1,2,3,4,5,6,20,23,300,301,302,304,305,306,307,308,309,310,311,312,313,314,315,316,317,318,3010,3020,3030,3050\",\"module\":\"folders\"}]"))

      .exec(
        MultipleAPI.multiple(
          true,
          "[{\"action\":\"mail\",\"module\":\"quota\"}]"))

      .exec(
        MailAPI.getAllMails(
          "default0/INBOX",
          "610,600,601,611",
          unseen = Some(true), deleted = Some(false), sort = Some("610"), order = Some("desc"), limit = Some("100")))

      .exec(session => session.remove("getAllMailsResponse"))

      .exec(session => {
        val now: Long = System.currentTimeMillis
        val t5y: Long = 5L * 365L * 86400L * 1000L
        val minus5y: Long = now - t5y
        //val plus1h: Long = now + 86400L * 1000L
        //val minus2h: Long = now - 2*3600*1000L
        //val plus2y: Long = now + 2*365*24*86400*1000L
        session.setAll("minus5y" -> minus5y)
      })

      .exec(
        CalendarAPI.getUpdatedAppointments(
          "1,20,101,102,206,207,201,200,202,400,401,402,221,224,227,2,209,212,213,214,215,222,216,220",
          "${minus5y}",
          Some(Map("showPrivate" -> "true", "recurrence_master" -> "true", "ignore" -> "deleted", "sort" -> "201", "order" -> "asc", "timezone" -> "UTC"))))
      .exec(session => session.remove("minus5y"))

      .exec(
        TasksAPI.getAllTasks(
          "${OX_FID_TASKS}",
          "1,20,200,203,221,300,309,317",
          Some(Map("sort" -> "317", "order" -> "asc", "timezone" -> "UTC"))))

      .exec(session => {
        val r = session("getAllTasksResponse").as[GetAllTasksResponse]
        val now: Long = System.currentTimeMillis
        var OD: Array[Map[String, Int]] = Array()
        for (
          t <- r.data
        ) {
          val c317 = t(7).asInstanceOf[Long]
          if (now > c317) {
            OD = OD ++ Array(Map("id" -> t(0).asInstanceOf[Int], "folder" -> t(1).asInstanceOf[Int]))
          }
        }
        session.set("overdueTasks", OD).remove("getAllTasksResponse")
      })

      .exec(ReminderAPI.getReminderRange(end = Some(DateTimeHelper.forOrBackward(1, ChronoUnit.HOURS).getTimestamp)))

      .exec(
        FoldersAPI.getSubFolders(
          "1",
          "1,2,3,4,5,6,20,23,300,301,302,304,305,306,307,308,309,310,311,312,313,314,315,316,317,318,3010,3020,3030,3050",
          Some(Map("all" -> "0", "altNames" -> "true", "timezone" -> "UTC", "tree" -> "0"))))
      .exec(session => session.remove(classOf[GetSubFolders].getSimpleName))

      .exec(
        MailAPI.getAllMails(
          "default0/INBOX",
          "102,600,601,602,603,604,605,606,607,608,610,611,614,652,656,X-Open-Xchange-Share-URL",
          sort = Some("610"), order = Some("desc"), categoryid = Some("general"), limit = Some("0,50")))
      .exec(session => session.remove("getAllMailsResponse"))

      .exec(
        TasksAPI.getTaskList(
          "1,2,5,20,101,200,203,220,221,300,301,309,316,317,401",
          "${overdueTasks}"))
      .exec(session => session.remove("overdueTasks"))

      .exec(
        UserAPI.getUserList("1,20,500,501,502,505,524,555,606,614", "[${OX_USER_ID}]"))

      .exec(
        CalendarAPI.getAllAppointments(
          "1,20,5,101,102,206,207,201,200,202,400,401,402,221,224,227,2,209,212,213,214,215,222,216,220",
          DateTimeHelper.forOrBackward(-1, ChronoUnit.MONTHS).startOfDay(),
          DateTimeHelper.forOrBackward(1, ChronoUnit.MONTHS).startOfDay(),
          Some(Map("showPrivate" -> "true", "recurrence_master" -> "false", "sort" -> "201", "order" -> "asc", "timezone" -> "UTC", "folder" -> "0"))))
      .exec(session => session.remove("getAllAppointmentsResponse"))

      .exec(
        FoldersAPI.getSubFolders(
          "default0",
          "1,2,3,4,5,6,20,23,300,301,302,304,305,306,307,308,309,310,311,312,313,314,315,316,317,318,3010,3020,3030,3050",
          Some(Map("all" -> "0", "altNames" -> "true", "timezone" -> "UTC", "tree" -> "0"))))
      .exec(session => session.remove(classOf[GetSubFolders].getSimpleName))

      .exec(
        FoldersAPI.getSubFolders(
          "default0/INBOX",
          "1,2,3,4,5,6,20,23,300,301,302,304,305,306,307,308,309,310,311,312,313,314,315,316,317,318,3010,3020,3030,3050",
          Some(Map("all" -> "0", "altNames" -> "true", "timezone" -> "UTC", "tree" -> "0"))))
      .exec(session => session.remove(classOf[GetSubFolders].getSimpleName))

  }

  val realisticTasksChain = {
    exec(session => {
      logger.debug("Executing realistic tasks chain...")
      session
    })

      .exec(
        TasksAPI.getAllTasks(
          "${OX_FID_TASKS}",
          "1,5,20,101,200,203,220,300,301,317",
          Some(Map("timezone" -> "UTC", "sort" -> "317", "order" -> "asc", "cache" -> "true"))))
      .exec(session => {
        val getAllTasksResponse = session("getAllTasksResponse").as[GetAllTasksResponse]
        val listTasksRequestBody = getAllTasksResponse.data.map(x => Map("id" -> x(0), "folder" -> x(2)))
        session.setAll("listTasksRequestBody" -> listTasksRequestBody.toArray)
      })
      .exec(session => session.remove("getAllTasksResponse"))

      // TODO: missing: Some(Map("altNames" -> "true", "timezone" -> "UTC"))
      .exec(
        FoldersAPI.getAllVisibleFolder(
          "tasks",
          "1,2,3,4,5,6,20,23,300,301,302,304,305,306,307,308,309,310,311,312,313,314,315,316,317,318,3010,3020,3030,3050",
          1))

      .group(ModuleNames.TASKS) {
        // missing: Some(Map("timezone"->"utc"))
        exec(
          new GetTaskList("1,2,5,20,101,200,203,220,221,300,301,309,316,317,401", "${listTasksRequestBody}").buildHttpRequest()
            .check(jsonPath("$.data").transform(x => JSONUtil().fromJson[Seq[Seq[Any]]](x)).optional.saveAs("listTasksResponseBody")))
      }
      .exec(session => session.remove("listTasksRequestBody"))

      .repeat(3, "rc") {
        exec(session => {
          val listTasksResponseBody = session("listTasksResponseBody").as[Seq[Seq[Any]]]
          val idx = if (session("rc").as[Int] == 0) 0 else ThreadLocalRandom.current().nextInt(listTasksResponseBody.length)

          val task = listTasksResponseBody(idx)
          // id: 0, folder: 3, email: 11

          session.setAll("getTaskId" -> task(0), "getTaskFolder" -> task(3))
        })
          .exec(TasksAPI.getTask("${getTaskId}", "${getTaskFolder}"))
          .exec(session => session.remove("getTask"))
          .doIf(session => session("rc").as[Int] == 0) {
            exec(
              JSlobAPI.setJSlob(
                "{\"interval\":30,\"notifyNewModifiedDeleted\":true,\"notifyAcceptedDeclinedAsCreator\":true,\"notifyAcceptedDeclinedAsParticipant\":true,\"folderview\":{\"visible\":{\"large\":true}}}",
                "io.ox/tasks"))
          }
      }
      .exec(
        JSlobAPI.setJSlob(
          "{\"interval\":30,\"notifyNewModifiedDeleted\":true,\"notifyAcceptedDeclinedAsCreator\":true,\"notifyAcceptedDeclinedAsParticipant\":true,\"folderview\":{\"visible\":{\"large\":true}}}",
          "io.ox/tasks"))
      .exec(session => session.removeAll("getTaskId", "getTaskFolder"))
      .exec(session => session.remove("listTasksReponseBody"))

      .exec(TaskChain.createChain("${OX_FID_TASKS}"))
      .exec(session => session.remove(classOf[Task].getSimpleName))
      .exec(session => {
        val myApptReqData = Array(
          Map(
            "id" -> session(classOf[CreateTask].getSimpleName).as[Int],
            "folder" -> session("OX_FID_TASKS").as[Int]))
        session.set("myApptReqData", myApptReqData)
      })

      .exec(
        JSlobAPI.setJSlob(
          "{\"interval\":30,\"notifyNewModifiedDeleted\":true,\"notifyAcceptedDeclinedAsCreator\":true,\"notifyAcceptedDeclinedAsParticipant\":true,\"folderview\":{\"visible\":{\"large\":true}}}",
          "io.ox/tasks"))
      .exec(
        TasksAPI.getAllTasks(
          "${OX_FID_TASKS}",
          "1,5,20,101,200,203,220,300,301,317",
          Some(Map("timezone" -> "UTC", "sort" -> "317", "order" -> "asc", "cache" -> "true"))))
      .exec(session => {
        val getAllTasksResponse = session("getAllTasksResponse").as[GetAllTasksResponse]
        val listTasksRequestBody = getAllTasksResponse.data.map(x => Map("id" -> x(0), "folder" -> x(2)))
        session.setAll("listTasksRequestBody" -> listTasksRequestBody.toArray)
      })
      .exec(session => session.remove("getAllTasksResponse"))

      .group(ModuleNames.TASKS) {
        // missing: Some(Map("timezone"->"utc"))
        exec(
          new GetTaskList("1,2,5,20,101,200,203,220,221,300,301,309,316,317,401", "${myApptReqData}").buildHttpRequest()
            .check(jsonPath("$.data").transform(x => JSONUtil().fromJson[Seq[Seq[Any]]](x)).optional.saveAs("listTasksResponseBody")))
      }
      .exec(session => session.removeAll("listTasksResponseBody", "myApptReqData"))

      .group(ModuleNames.TASKS) {
        // missing: Some(Map("timezone"->"utc"))
        exec(
          new GetTaskList("1,2,5,20,101,200,203,220,221,300,301,309,316,317,401", "${listTasksRequestBody}").buildHttpRequest()
            .check(jsonPath("$.data").transform(x => JSONUtil().fromJson[Seq[Seq[Any]]](x)).optional.saveAs("listTasksResponseBody")))
      }
      .exec(session => session.remove("listTasksRequestBody"))
      .exec(session => session.remove("listTasksResponseBody"))

      .exec(
        JSlobAPI.setJSlob(
          "{\"interval\":30,\"notifyNewModifiedDeleted\":true,\"notifyAcceptedDeclinedAsCreator\":true,\"notifyAcceptedDeclinedAsParticipant\":true,\"folderview\":{\"visible\":{\"large\":true}}}",
          "io.ox/tasks"))
  }

  val realisticContactsChain = {
    exec(session => {
      logger.debug("Executing realistic contacts chain...")
      session
    })

      .exec(
        JSlobAPI.setJSlob(
          // originally there was also a \"vgrid\":{\"editable\":false,\"previous\":{\"22\":\"22.58\"}} key, skipping since I dont have a "previous" at hand.
          "{\"showAdmin\":false,\"fullNameFormat\":\"auto\",\"startInGlobalAddressbook\":true,\"mapService\":\"google\",\"folderview\":{\"visible\":{\"small\":true,\"medium\":true}},\"search\":{\"minimumQueryLength\":0,\"autocompleteFields\":\"first_name,last_name,display_name,email1,email2,email3\"},\"maxImageSize\":4194304}",
          "io.ox/contacts"))

      .exec(
        ContactsAPI.getAllContacts(
          "${OX_FID_CONTACTS}",
          "1,20,101,607",
          Some(Map("timezone" -> "utc", "sort" -> "607", "order" -> "asc", "admin" -> "false"))))
      .exec(session => {
        val getAllContactsResponse = session(classOf[GetAllContacts].getSimpleName).as[GetAllContactsResponse]
        val listContactsRequestBody = getAllContactsResponse.data.map(x => Map("id" -> x(0), "folder" -> x(1)))
        session.setAll("listContactsRequestBody" -> listContactsRequestBody.toArray)
      })
      .exec(session => session.remove(classOf[GetAllContacts].getSimpleName))

      // TODO: missing: Some(Map("altNames" -> "true", "timezone" -> "UTC"))
      .exec(
        FoldersAPI.getAllVisibleFolder(
          "contacts",
          "1,2,3,4,5,6,20,23,300,301,302,304,305,306,307,308,309,310,311,312,313,314,315,316,317,318,3010,3020,3030,3050",
          1))

      .exec(FoldersAPI.getFolder("${OX_FID_CONTACTS}", 1, altNames = Some(true)))

      .group(ModuleNames.CONTACTS) {
        // missing: Some(Map("timezone"->"utc"))
        exec(
          new ListContacts("1,2,5,20,101,500,501,502,505,520,524,555,556,557,569,592,602,606,607", "${listContactsRequestBody}").buildHttpRequest()
            .check(jsonPath("$.data").transform(x => JSONUtil().fromJson[Seq[Seq[Any]]](x)).optional.saveAs("listContactsResponseBody")))
      }
      .exec(session => session.remove("listContactsRequestBody"))

      .repeat(3, "rc") {
        exec(session => {
          val listContactsResponseBody = session("listContactsResponseBody").as[Seq[Seq[Any]]]
          val idx = if (session("rc").as[Int] == 0) 0 else ThreadLocalRandom.current().nextInt(listContactsResponseBody.length)

          val contact = listContactsResponseBody(idx)
          // id: 0, folder: 1, email: 11

          session.setAll("getContactId" -> contact(0), "getContactFolder" -> contact(3), "getContactEmail" -> contact(11))
        })
          .exec(ContactsAPI.getContact("${getContactId}", "${getContactFolder}"))
          .exec(session => session.remove("getContactResponse"))
          .doIf(session => session("rc").as[Int] == 0) {
            // only for first pass
            // not a copy-paste error -- the UI really calls this two times in my logging.
            exec(ContactsAPI.getContact("${getContactId}", "${getContactFolder}"))
              .exec(JSlobAPI.getJSLobPortal)
            // only for first pass end

          }
          .exec(HaloAPI.getHaloContactPicture(width = Some(128), height = Some(128), scaleType = Some("cover"),
            internal_userid = Some("${OX_USER_ID}"), context = Some("${OX_CID}"), user = Some("${OX_USER_ID}"),
            email = Some("${getContactEmail}")))

          .exec(
            JSlobAPI.setJSlob(
              // originally there was also a \"vgrid\":{\"editable\":false,\"previous\":{\"22\":\"22.58\"}} key, skipping since I dont have a "previous" at hand.
              "{\"showAdmin\":false,\"fullNameFormat\":\"auto\",\"startInGlobalAddressbook\":true,\"mapService\":\"google\",\"folderview\":{\"visible\":{\"small\":true,\"medium\":true}},\"search\":{\"minimumQueryLength\":0,\"autocompleteFields\":\"first_name,last_name,display_name,email1,email2,email3\"},\"maxImageSize\":4194304}",
              "io.ox/contacts"))

      }
      .exec(session => session.removeAll("getContactId", "getContactFolder", "getContactEmail"))
      .exec(session => session.remove("listContactsResponseBody"))

      .exec(ContactChain.createChain("${OX_FID_CONTACTS}"))
      .exec(session => session.remove("ContactData"))

      .exec(session => {
        val myContactReqData = Array(
          Map(
            "id" -> session(classOf[CreateContact].getSimpleName).as[Int],
            "folder" -> session("OX_FID_CONTACTS").as[Int]))
        session.set("myContactReqData", myContactReqData)
      })

      .exec(ContactsAPI.getContact("${CreateContact}", "${OX_FID_CONTACTS}"))

      .exec(session => {
        val email1 = session(classOf[GetContact].getSimpleName).as[ContactData].email1
        session.set("rcemail1", email1).remove(classOf[GetContact].getSimpleName)
      })

      .exec(
        ContactsAPI.getAllContacts(
          "${OX_FID_CONTACTS}",
          "1,20,101,607",
          Some(Map("timezone" -> "utc", "sort" -> "607", "order" -> "asc", "admin" -> "false"))))
      .exec(session => {
        val getAllContactsResponse = session(classOf[GetAllContacts].getSimpleName).as[GetAllContactsResponse]
        val listContactsRequestBody = getAllContactsResponse.data.map(x => Map("id" -> x(0), "folder" -> x(1)))
        session.setAll("listContactsRequestBody" -> listContactsRequestBody.toArray)
      })
      .exec(session => session.remove(classOf[GetAllContacts].getSimpleName))
      .exec(HaloAPI.getHaloContactPicture(width = Some(128), height = Some(128), scaleType = Some("cover"),
        internal_userid = Some("${OX_USER_ID}"), context = Some("${OX_CID}"), user = Some("${OX_USER_ID}"),
        email = Some("${getContactEmail}")))
      .exec(session => session.removeAll("rcemail1", "CreateContact"))

      .exec(
        JSlobAPI.setJSlob(
          "{\"showAdmin\":false,\"fullNameFormat\":\"auto\",\"startInGlobalAddressbook\":true,\"mapService\":\"google\",\"folderview\":{\"visible\":{\"small\":true,\"medium\":true}},\"search\":{\"minimumQueryLength\":0,\"autocompleteFields\":\"first_name,last_name,display_name,email1,email2,email3\"},\"maxImageSize\":4194304}",
          "io.ox/contacts"))

      .group(ModuleNames.CONTACTS) {
        // missing: Some(Map("timezone"->"utc"))
        exec(
          new ListContacts("1,2,5,20,101,500,501,502,505,520,524,555,556,557,569,592,602,606,607", "${listContactsRequestBody}").buildHttpRequest()
            .check(jsonPath("$.data").transform(x => JSONUtil().fromJson[Seq[Seq[Any]]](x)).optional.saveAs("listContactsResponseBody")))
      }
      .exec(session => session.remove("listContactsRequestBody"))
      .exec(session => session.remove("listContactsResponseBody"))

      .group(ModuleNames.CONTACTS) {
        // missing: Some(Map("timezone"->"utc"))
        exec(
          new ListContacts("1,2,5,20,101,500,501,502,505,520,524,555,556,557,569,592,602,606,607", "${myContactReqData}").buildHttpRequest()
            .check(jsonPath("$.data").transform(x => JSONUtil().fromJson[Seq[Seq[Any]]](x)).optional.saveAs("listContactsResponseBody")))
      }
      .exec(session => session.removeAll("listContactsResponseBody", "myContactReqData"))

      .exec(FoldersAPI.getFolder("${OX_FID_CONTACTS}", 1, altNames = Some(true)))

      .exec(
        JSlobAPI.setJSlob(
          "{\"showAdmin\":false,\"fullNameFormat\":\"auto\",\"startInGlobalAddressbook\":true,\"mapService\":\"google\",\"folderview\":{\"visible\":{\"small\":true,\"medium\":true}},\"search\":{\"minimumQueryLength\":0,\"autocompleteFields\":\"first_name,last_name,display_name,email1,email2,email3\"},\"maxImageSize\":4194304}",
          "io.ox/contacts"))
  }

  val realisticCalendarChain = {
    exec(session => {
      logger.debug("Executing realistic calendar chain...")
      session
    })

      .exec(
        FoldersAPI.getAllVisibleFolder(
          "calendar",
          "1,2,3,4,5,6,20,23,300,301,302,304,305,306,307,308,309,310,311,312,313,314,315,316,317,318,3010,3020,3030,3050",
          1
        /* missing: altNames=true&timezone=UTC */
        ))

      .exec(
        CalendarAPI.getAllAppointments(
          "1,20,5,101,102,206,207,201,200,202,400,401,402,221,224,227,2,209,212,213,214,215,222,216,220",
          DateTimeHelper.forOrBackward(-1, ChronoUnit.HOURS),
          DateTimeHelper.forOrBackward(2, ChronoUnit.MONTHS),
          Some(Map("showPrivate" -> "true", "recurrence_master" -> "false", "sort" -> "201", "order" -> "asc", "timezone" -> "UTC", "folder" -> "0"))))
      .exec(session => session.remove("getAllAppointmentsResponse"))

      .exec(
        UserAPI.getUserList("1,20,500,501,502,505,524,555,606,614", "[${OX_USER_ID}]"))
      .exec(HaloAPI.getHaloContactPicture(width = Some(128), height = Some(128), scaleType = Some("cover"),
        internal_userid = Some("${OX_USER_ID}"), context = Some("${OX_CID}"), user = Some("${OX_USER_ID}"),
        email = Some("${getContactEmail}")))
      .exec(
        new Multiple(
          true,
          "[{\"action\":\"autocomplete\",\"query\":\"q\",\"admin\":false,\"email\":true,\"sort\":\"609\",\"columns\":\"1,2,5,20,101,500,501,502,505,519,520,524,555,556,557,569,592,602,606,607,551,552\",\"right_hand_limit\":10,\"module\":\"contacts\"},{\"action\":\"search\",\"timezone\":\"utc\",\"admin\":false,\"emailAutoComplete\":false,\"limit\":10,\"module\":\"resource\",\"data\":{\"pattern\":\"q\"}},{\"action\":\"search\",\"timezone\":\"utc\",\"columns\":\"1,20,500,524\",\"sort\":\"500\",\"order\":\"asc\",\"admin\":false,\"emailAutoComplete\":false,\"limit\":10,\"module\":\"group\",\"data\":{\"pattern\":\"q\"}}]")
          .buildHttpRequest()
          .check(bodyString.transform(x => JSONUtil().fromJson[Object](x)).optional.saveAs("autocompleteResponse")))
      .exec(session => {
        // TODO
        // Now get some data out of that List(Map(timestamp->..., data->List(List))) thing
        // any help how to do that nicer without compiler warnings appreciated
        var s = session
        val ar = session("autocompleteResponse").as[Object]
        ar match {
          case l: List[Any] => {
            if (l.nonEmpty) l.head match {
              case m: Map[String, Any] => {
                m.get("data").get match {
                  case l2: List[Any] => {
                    if (l2.nonEmpty) l2(0) match {
                      case l4: List[Any] => s = s.setAll("rcemail1" -> l4(12), "rcemailfid1" -> l4(3))
                    }
                    if (l2.size > 1) l2(1) match {
                      case l5: List[Any] => s = s.setAll("rcemail2" -> l5(12), "rcemailfid2" -> l5(3))
                    }
                    if (l2.nonEmpty) for (i <- 1 to 4) {
                      val rc = l2(ThreadLocalRandom.current().nextInt(l2.length))
                      rc match {
                        case l3: List[Any] => s = s.setAll(("rcid" + i) -> l3(0), ("rcfid" + i) -> l3(3))
                      }
                    }
                  }
                }
              }
            }
          }
        }
        s.remove("autocompleteResponse")
      })

      .exec(
        MultipleAPI.multiple(
          true,
          "[{\"action\":\"search\",\"timezone\":\"utc\",\"admin\":false,\"emailAutoComplete\":false,\"limit\":10,\"module\":\"resource\",\"data\":{\"pattern\":\"qw\"}},{\"action\":\"search\",\"timezone\":\"utc\",\"columns\":\"1,20,500,524\",\"sort\":\"500\",\"order\":\"asc\",\"admin\":false,\"emailAutoComplete\":false,\"limit\":10,\"module\":\"group\",\"data\":{\"pattern\":\"qw\"}}]"))

      .doIfOrElse(session => session.contains("rcid1")) {
        exec(ContactsAPI.getContact("${rcid1}", "${rcfid1}"))
      } {
        exec(session => {
          println("Warning, skipping stuff because not defined: rcid1")
          session
        })
      }
      .doIfOrElse(session => session.contains("rcid2")) {
        exec(ContactsAPI.getContact("${rcid2}", "${rcfid2}"))
      } {
        exec(session => {
          println("Warning, skipping stuff because not defined: rcid2")
          session
        })
      }
      .doIfOrElse(session => session.contains("rcid3")) {
        exec(ContactsAPI.getContact("${rcid3}", "${rcfid3}"))
      } {
        exec(session => {
          println("Warning, skipping stuff because not defined: rcid3")
          session
        })
      }
      .doIfOrElse(session => session.contains("rcid4")) {
        exec(ContactsAPI.getContact("${rcid4}", "${rcfid4}"))
      } {
        exec(session => {
          println("Warning, skipping stuff because not defined: rcid4")
          session
        })
      }

      .doIfOrElse(session => session.contains("rcemail1")) {
        exec(HaloAPI.getHaloContactPicture(width = Some(128), height = Some(128), scaleType = Some("cover"),
          internal_userid = Some("${OX_USER_ID}"), context = Some("${OX_CID}"), user = Some("${OX_USER_ID}"),
          email = Some("${getContactEmail}")))
      } {
        exec(session => {
          println("Warning, skipping stuff because not defined: rcemail1")
          session
        })
      }
      .doIfOrElse(session => session.contains("rcemail2")) {
        exec(HaloAPI.getHaloContactPicture(width = Some(128), height = Some(128), scaleType = Some("cover"),
          internal_userid = Some("${OX_USER_ID}"), context = Some("${OX_CID}"), user = Some("${OX_USER_ID}"),
          email = Some("${getContactEmail}")))
      } {
        exec(session => {
          println("Warning, skipping stuff because not defined: rcemail2")
          session
        })
      }
      .doIfOrElse(session => session.contains("rcemail1")) {
        exec(session => {
          val folderid = session("OX_FID_CALENDAR").as[String]
          val taskAppData = TaskAppointmentData(
            Some(folderid),
            title = Some("PS Lunch"),
            start_date = DateTimeHelper.forOrBackward(12, ChronoUnit.HOURS).startOfDay(),
            end_date = DateTimeHelper.forOrBackward(13, ChronoUnit.HOURS).startOfDay(),
            note = Some("as usual"),
            notification = Some(false),
            private_flag = Some(ThreadLocalRandom.current().nextBoolean()),
            participants = Some(Array(Participant(2, 1), Participant(0, 5, Some(session("rcemail1").as[String])))))
          val detailedAppData = DetailedAppointmentData(
            shown_as = Some(ThreadLocalRandom.current().nextInt(1, 4 + 1)),
            full_time = Some(ThreadLocalRandom.current().nextBoolean()),
            ignore_conflicts = Some(true))
          val a: Appointment = new Appointment(taskAppData, detailedAppData)
          session.set("newAppt", a)
        })
          .exec(CalendarAPI.createAppointment("${newAppt}"))
          .exec(CalendarAPI.getAppointment("${CreateAppointment}", "${OX_FID_CALENDAR}"))
      } {
        exec(session => {
          println("Warning, skipping stuff because not defined: rcemail1")
          session
        })
      }

      .exec(session => session.removeAll("rcid1", "rcfid1", "rcid2", "rcfid2", "rcid3", "rcfid3", "rcid4", "rcfid4", "rcemail1", "rcemail2", "newAppt", "getContactResponse"))

      .exec(ReminderAPI.getReminderRange(end = Some(DateTimeHelper.forOrBackward(1, ChronoUnit.HOURS).getTimestamp)))

      .exec(
        CalendarAPI.getAllAppointments(
          "1,20,5,101,102,206,207,201,200,202,400,401,402,221,224,227,2,209,212,213,214,215,222,216,220",
          DateTimeHelper.forOrBackward(-1, ChronoUnit.HOURS).startOfDay(),
          DateTimeHelper.forOrBackward(2, ChronoUnit.MONTHS).startOfDay(),
          Some(Map("showPrivate" -> "true", "recurrence_master" -> "false", "sort" -> "201", "order" -> "asc", "timezone" -> "UTC", "folder" -> "0"))))
      .exec(session => session.remove("getAllAppointmentsResponse"))

      .exec(
        FoldersAPI.getFolder("21", 1, altNames = Some(true)))

      .exec(
        UserAPI.getUserList("1,20,500,501,502,505,524,555,606,614", "[${OX_USER_ID}]"))

  }

  val realisticMailChain = {
    exec(session => {
      logger.debug("Executing realistic mail chain...")
      session
    })

      .exec(
        MailAPI.getAllMails(
          "default0/INBOX",
          "102,600,601,602,603,604,605,606,607,608,610,611,614,652,656,X-Open-Xchange-Share-URL",
          sort = Some("610"), order = Some("desc"), categoryid = Some("general"), limit = Some("0,50")))

      .repeat(4, "rc") {
        exec(session => {
          val idx = if (session("rc").as[Int] == 0) 0 else ThreadLocalRandom.current().nextInt(session("getAllMailsResponse").as[GetAllMailsResponse].data.size)
          val randomElement = session("getAllMailsResponse").as[GetAllMailsResponse].data(idx)
          val objectId = randomElement(1)
          val folderId = randomElement(2)
          val contentType = randomElement(14)
          val sender = randomElement(4)
          val email = sender.asInstanceOf[List[List[String]]](0)(1)
          session.setAll("objectId" -> objectId, "folderId" -> folderId, "contentType" -> contentType, "email" -> email)
        })

          .exec(MailAPI.getMail("${objectId}", "${folderId}", Some(Map("embedded" -> 0, "unseen" -> true, "timezone" -> "UTC", "view" -> "html", "max_size" -> 102400))))

          .exec(
            MultipleAPI.multiple(
              true,
              "[{\"action\":\"update\",\"id\":\"${objectId}\",\"folder\":\"${folderId}\",\"timestamp\":2116800000000,\"module\":\"mail\",\"data\":{\"flags\":32,\"value\":true}}]"))
          .exec(HaloAPI.getHaloContactPicture(width = Some(128), height = Some(128), scaleType = Some("cover"),
            internal_userid = Some("${OX_USER_ID}"), context = Some("${OX_CID}"), user = Some("${OX_USER_ID}"),
            email = Some("${getContactEmail}")))
      }
      .exec(session => session.remove("getAllMailsResponse"))

      .exec(session => {
        session.removeAll("objectId", "folderId", "contentType", "email")
      })

      .exec(FoldersAPI.getFolder("default0/INBOX", 0, altNames = Some(true)))

      .exec(
        new Multiple(
          true,
          "[{\"action\":\"autocomplete\",\"query\":\"q\",\"admin\":false,\"email\":true,\"sort\":\"609\",\"columns\":\"1,2,5,20,101,500,501,502,505,519,520,524,555,556,557,569,592,602,606,607,551,552\",\"right_hand_limit\":10,\"module\":\"contacts\"}]")
          .buildHttpRequest()
          .check(jsonPath("$[0].data").transform(x => JSONUtil().fromJson[Seq[Seq[Any]]](x)).optional.saveAs("autocompleteResponse")))
      .exec(session => {
        val a = session("autocompleteResponse").as[Seq[Seq[Any]]]
        val b = if (a.nonEmpty) scala.util.Random.shuffle(
          a.map(x => Map("id" -> x(0), "folderid" -> x(3), "email" -> x(12), "first_name" -> x(6), "last_name" -> x(7)))).take(if (a.size > 4) 4 else a.size)
        else List()
        session.remove("autocompleteResponse").set("autocompleteContacts", b)
      })

      .repeat(session => session("autocompleteContacts").as[Seq[Any]].length, "rc") {
        exec(session => {
          val l = session("autocompleteContacts").as[Seq[Map[String, Any]]]
          val m = l(session("rc").as[Int])

          session.setAll("id" -> m.get("id").get, "folderid" -> m.get("folderid").get, "email" -> m.get("email").get, "first_name" -> m.get("first_name").get, "last_name" -> m.get("last_name").get)
        })
          .exec(ContactsAPI.getContact("${id}", "${folderid}"))
          .exec(session => session.remove("getContactResponse"))

          .exec(HaloAPI.getHaloContactPicture(width = Some(128), height = Some(128), scaleType = Some("cover"),
            internal_userid = Some("${OX_USER_ID}"), context = Some("${OX_CID}"), user = Some("${OX_USER_ID}"),
            email = Some("${getContactEmail}")))
      }
      .exec(session => session.remove("autocompleteContacts"))

      .doIfOrElse(session => session.contains("email")) {
        exec(session => {
          val mail = com.openexchange.testing.performance.groupware.mail.Mail(MailData(
            from = Array(Array[String](session(USER_FIRST_NAME).as[String] + " " + session(USER_LAST_NAME).as[String], session(USER_EMAIL1).as[String])),
            to = Array(Array[String](session("first_name").as[String] + " " + session("last_name").as[String], session("email").as[String])),
            subject = java.util.UUID.randomUUID().toString,
            priority = ThreadLocalRandom.current().nextInt(0, 6),
            attachments = Some(Array(MailAttachment(content_type = "text/html", content = "Plain stupid content"))),
            vcard = Some(ThreadLocalRandom.current().nextInt(0, 2)),
            contacts_ids = Some(Array.empty[String]),
            infostore_ids = Some(Array.empty[String]),
            dataSources = Some(Array.empty[String])))
          session.set("newMail", mail)
        })
          .exec(MailAPI.sendMail("${newMail}"))
      } {
        exec(session => {
          println("Warning: not sending new Mail as not found any hits from autocomplete")
          session
        })
      }

      .exec(session => session.removeAll("newMail", "first_name", "last_name", "email"))

    // TODO: cleanup session; create getallmails here in this chain, no x-dep to idlechain
  }

}
