/*
 * Copyright (C) 2018 OX Software GmbH
 *
 * This file is part of OX API-Performance Tests.
 *
 * OX API-Performance Tests is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * OX API-Performance Tests is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OX API-Performance Tests.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.openexchange.testing.performance.simulations

import java.util.concurrent.TimeUnit

import com.openexchange.testing.performance.ajax.config.ConfigAPI
import com.openexchange.testing.performance.ajax.folders.FoldersAPI
import com.openexchange.testing.performance.ajax.login.LoginChain
import com.openexchange.testing.performance.common.Capabilities._
import com.openexchange.testing.performance.common.SessionConstants._
import com.openexchange.testing.performance.config.{ Configuration, HTTPConfig }
import com.openexchange.testing.performance.simulations.GuardChains._
import com.openexchange.testing.performance.simulations.OXAppSuiteChains._
import com.typesafe.scalalogging.LazyLogging
import io.gatling.core.Predef._

import scala.concurrent.duration._

/**
 * [[QAOXAppSuiteSoakSimulation]]
 *
 * @author <a href="mailto:nikolaos.tsapanidis@open-xchange.com">Nikolaos Tsapanidis</a>
 */
class QAOXAppSuiteSoakSimulation extends Simulation with Configuration with LazyLogging {

  val scn = scenario(getClass.getSimpleName)

    .feed(csv(settings.simulation.userFeeder).circular)
    .exec(LoginChain.doLogin())

    .doIf(session => session.contains(SESSION_ID)) {
      pause(10 seconds, 25 seconds)

        .foreach("${modules}", "module") {
          doSwitch("${module}")(
            webmail -> exec(
              ConfigAPI.getDefaultFolderInbox,
              ConfigAPI.getDefaultFolderDrafts,
              ConfigAPI.getDefaultFolderSent,
              ConfigAPI.getDefaultFolderSpam,
              ConfigAPI.getDefaultFolderTrash),
            calendar -> ConfigAPI.getDefaultFolderCalendar,
            contacts -> ConfigAPI.getDefaultFolderContacts,
            infostore -> exec(ConfigAPI.getDefaultFolderInfostore, ConfigAPI.getDefaultFolderInfostoreTrash),
            tasks -> ConfigAPI.getDefaultFolderTasks)
        }
        .doIf(session => session(CAPABILITIES).as[Seq[String]].contains(guard)) {
          exec(loginChainGuard)
        }

        .during(new FiniteDuration(settings.simulation.injection.loopTime, TimeUnit.SECONDS), exitASAP = false) {
          pace(
            new FiniteDuration(10, TimeUnit.SECONDS),
            new FiniteDuration(settings.simulation.injection.loopTime, TimeUnit.SECONDS))
            .randomSwitch(
              5d -> doIf(session => session("modules").as[Seq[String]].contains(portal) && session(CAPABILITIES).as[Seq[String]].contains(portal)) {
                portalChain
                  .pause(10 seconds, 45 seconds)
              },
              30d -> doIf(session => session("modules").as[Seq[String]].contains(webmail) && session(CAPABILITIES).as[Seq[String]].contains(webmail)) {
                mailChain
                  .pause(10 seconds, 45 seconds)
              },
              20d -> doIf(session => session("modules").as[Seq[String]].contains(calendar) && session(CAPABILITIES).as[Seq[String]].contains(calendar)) {
                calendarChain
                  .pause(10 seconds, 45 seconds)
              },
              5d -> doIf(session => session("modules").as[Seq[String]].contains(tasks) && session(CAPABILITIES).as[Seq[String]].contains(tasks)) {
                taskChain
                  .pause(10 seconds, 45 seconds)
              },
              25d -> doIf(session => session("modules").as[Seq[String]].contains(contacts) && session(CAPABILITIES).as[Seq[String]].contains(contacts)) {
                contactChain
                  .pause(10 seconds, 45 seconds)
              },
              15d -> doIf(session => session("modules").as[Seq[String]].contains(infostore) && session(CAPABILITIES).as[Seq[String]].contains(infostore)) {
                driveChain
                  .pause(10 seconds, 45 seconds)
              })
        }

        // cleanup
        .doIf(settings.simulation.data.keepData) {
          foreach(settings.appsuite.modules.keys.toList, "module") {
            doSwitch("${module}")(
              "webmail" -> FoldersAPI.clearFolders(Array($(FID_SENT), $(FID_DRAFTS), $(FID_TRASH))),
              "calendar" -> FoldersAPI.clearFolders(Array($(FID_CALENDAR))),
              "contacts" -> FoldersAPI.clearFolders(Array($(FID_CONTACTS))),
              "infostore" -> FoldersAPI.clearFolders(Array($(FID_INFOSTORE), $(FID_INFOSTORE_TRASH))),
              "tasks" -> FoldersAPI.clearFolders(Array($(FID_TASKS))))
          }
        }

        .doIf(session => session(CAPABILITIES).as[Seq[String]].contains(guard)) {
          exec(logoutChainGuard)
        }
        .exec(LoginChain.doLogout())
    }

  setUp(scn.inject(
    rampUsersPerSec(1) to 10 during (30 seconds),
    constantUsersPerSec(settings.simulation.injection.usersPerSec).during(new FiniteDuration(settings.simulation
      .injection.duration, TimeUnit.SECONDS))))
    .protocols(HTTPConfig.appsuiteConf)
    .assertions(global.successfulRequests.percent.gte(settings.simulation.assertions.successfulRequests))
}

