/*
 * Copyright (C) 2018 OX Software GmbH
 *
 * This file is part of OX API-Performance Tests.
 *
 * OX API-Performance Tests is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * OX API-Performance Tests is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OX API-Performance Tests.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.openexchange.testing.performance.aws

import java.io.File
import java.util

import com.amazonaws.auth.{ AWSCredentials, AWSStaticCredentialsProvider, BasicAWSCredentials }
import com.amazonaws.client.builder.AwsClientBuilder.EndpointConfiguration
import com.amazonaws.services.s3.model.DeleteObjectsRequest.KeyVersion
import com.amazonaws.services.s3.model.{ AbortMultipartUploadRequest, CompleteMultipartUploadRequest, InitiateMultipartUploadRequest, PartETag, UploadPartRequest, _ }
import com.amazonaws.services.s3.{ AmazonS3, AmazonS3ClientBuilder }
import com.amazonaws.util.StringUtils
import com.amazonaws.{ ClientConfiguration, Protocol }
import com.openexchange.testing.performance.config.Configuration
import com.typesafe.scalalogging.LazyLogging

import scala.collection.JavaConverters._
import scala.collection.mutable
import scala.collection.mutable.ArrayBuffer

/**
 * [[AwsS3Client]]
 *
 * @author <a href="mailto:nikolaos.tsapanidis@open-xchange.com">Nikolaos Tsapanidis</a>
 * @author <a href="mailto:philipp.schumacher@open-xchange.com">Philipp Schumacher</a>
 */

class AwsS3Client(accessKey: String, secretKey: String, endpoint: String, region: String, bucket: String, pathStyleAccess: Boolean, signer: Option[String]) extends LazyLogging {
  val credentials: AWSCredentials = new BasicAWSCredentials(accessKey, secretKey)
  val clientConfiguration: ClientConfiguration = new ClientConfiguration()
  if (signer != None) clientConfiguration.setSignerOverride(signer.get)
  clientConfiguration.setProtocol(Protocol.HTTP)
  val endpointConfiguration: EndpointConfiguration = new EndpointConfiguration(endpoint, region)

  val client: AmazonS3 = AmazonS3ClientBuilder.standard()
    .withCredentials(new AWSStaticCredentialsProvider(credentials))
    .withClientConfiguration(clientConfiguration)
    .withEndpointConfiguration(endpointConfiguration)
    .withPathStyleAccessEnabled(pathStyleAccess)
    .build()

  def listBuckets(): mutable.Seq[Bucket] = client.listBuckets().asScala

  def listBucketsContent(bucketName: String = bucket): mutable.Seq[S3ObjectSummary] = {
    val objects: ObjectListing = client.listObjects(bucketName)
    val objectSummaries: mutable.Seq[S3ObjectSummary] = objects.getObjectSummaries.asScala

    logger.info("Listing content of bucket \"{}\" ", bucketName)
    for (objectSummary: S3ObjectSummary <- objectSummaries) {
      logger.debug(s"${objectSummary.getKey}\t${objectSummary.getSize}\t${StringUtils.fromDate(objectSummary.getLastModified)}")
    }
    objectSummaries
  }

  def createObject(bucketName: String = bucket, destinationPath: String, file: File): Unit = {
    // Perform multipart upload if files are bigger than 10MB
    if (file.length() > 10485760) {
      val partETags = new util.ArrayList[PartETag]
      val initRequest = new InitiateMultipartUploadRequest(bucketName, destinationPath)
      val initResponse = client.initiateMultipartUpload(initRequest)
      val contentLength = file.length
      var partSize = 10485760L
      try {
        var filePosition = 0L
        var i = 1
        while ({
          filePosition < contentLength
        }) {
          partSize = Math.min(partSize, contentLength - filePosition)
          val uploadRequest = new UploadPartRequest().withBucketName(bucketName).withKey(destinationPath).withUploadId(initResponse.getUploadId).withPartNumber(i).withFileOffset(filePosition).withFile(file).withPartSize(partSize)
          partETags.add(client.uploadPart(uploadRequest).getPartETag)
          filePosition += partSize

          {
            i += 1;
            i - 1
          }
        }
        val compRequest = new CompleteMultipartUploadRequest(bucketName, destinationPath, initResponse.getUploadId, partETags)
        client.completeMultipartUpload(compRequest)
      } catch {
        case _: Exception =>
          client.abortMultipartUpload(new AbortMultipartUploadRequest(bucketName, destinationPath, initResponse.getUploadId))
      }
    } else {
      val putObjectRequest = new PutObjectRequest(bucketName, destinationPath, file)
      val putObjectResult: PutObjectResult = client.putObject(putObjectRequest)
      logger.info("Successfully created object {}", putObjectResult.getMetadata)
    }
  }

  def getObject(bucketName: String = bucket, key: String): S3ObjectInputStream = {
    try {
      client.getObject(bucketName, key).getObjectContent
    } catch {
      case _: Exception => return null
    }

  }

  def deleteBucketsContent(bucketName: String = bucket): Unit = {
    val s3objects = listBucketsContent(bucketName)
    val multiObjectDeleteRequest: DeleteObjectsRequest = new DeleteObjectsRequest(bucketName)
    val keys: ArrayBuffer[KeyVersion] = ArrayBuffer[KeyVersion]()

    for (s3Object <- s3objects) {
      keys += new KeyVersion(s3Object.getKey)
      logger.debug("Successfully deleted object with key {} from bucket {}.", s3Object.getKey, bucketName)
    }

    if (keys.nonEmpty) {
      multiObjectDeleteRequest.setKeys(keys.asJava)
      val delObjRes: DeleteObjectsResult = client.deleteObjects(multiObjectDeleteRequest)
      logger.info("Successfully deleted {} object(s).", delObjRes.getDeletedObjects.size())
    } else {
      logger.info("No objects to delete")
    }
  }

}
