/*
 * Copyright (C) 2018 OX Software GmbH
 *
 * This file is part of OX API-Performance Tests.
 *
 * OX API-Performance Tests is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * OX API-Performance Tests is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OX API-Performance Tests.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.openexchange.testing.performance.common

import java.util.concurrent.TimeUnit

import com.openexchange.testing.performance.config.{ Configuration, HTTPConfig }
import io.gatling.core.Predef._
import io.gatling.core.controller.inject.InjectionStep
import io.gatling.core.structure.ScenarioBuilder
import io.gatling.http.protocol.HttpProtocolBuilder
import org.asynchttpclient
import org.asynchttpclient.Dsl.asyncHttpClient
import org.asynchttpclient._

import scala.concurrent.duration.FiniteDuration

/**
 * [[BaseSimulation]]
 *
 * @author <a href="mailto:nikolaos.tsapanidis@open-xchange.com">Nikolaos Tsapanidis</a>
 */
trait BaseSimulation extends Simulation with Configuration {
  private val perfType: String = settings.simulation.performanceType.getOrElse("normal")
  // TODO can we agree to change rampup to just a duration, not a divisor?
  private val rampUp: Long = perfType match {
    // effectively integer division rounded up
    case "webui" => (settings.simulation.injection.users + settings.simulation.injection.usersPerSec - 1) / settings.simulation.injection.usersPerSec
    case _       => settings.simulation.injection.duration / settings.simulation.injection.rampup
  }
  private val durationTime: Long = settings.simulation.injection.duration
  private val concurrency: Int = settings.simulation.injection.users
  val finiteDuration: FiniteDuration = perfType match {
    // TODO can we agree to just use FiniteDuration(durationTime, TimeUnit.SECONDS) also for peak? this arithmetic
    // is just over-complicating it for me, but okay...
    case "peak"                                 => new FiniteDuration(durationTime - rampUp, TimeUnit.SECONDS)
    case "webui"                                => new FiniteDuration(durationTime + rampUp, TimeUnit.SECONDS)
    case "normal" | "soak" | "stress" | "spike" => new FiniteDuration(durationTime, TimeUnit.SECONDS)
    case _                                      => throw new UnsupportedOperationException("Performance type not supported.")
  }
  lazy val httpConf: HttpProtocolBuilder = HTTPConfig.appsuiteConf
  lazy val serverVersion: String = {
    val config = new DefaultAsyncHttpClientConfig.Builder().setAcceptAnyCertificate(true).build()
    asyncHttpClient(config).prepareGet(s"${settings.middleware.baseUrl}/servlet/webdav" +
      s".version").execute().toCompletableFuture.join().getResponseBody.toLowerCase.trim
  }

  before {
    if (!settings.influx.endpoint.isEmpty) {
      sendEvent("Start of simulation.", serverVersion, perfType, getClass.getSimpleName)
    }
  }

  setUp(scn.inject(getInjectionStep())).protocols(httpConf)

  after {
    if (!settings.influx.endpoint.isEmpty) {
      sendEvent("End of simulation.", serverVersion, perfType, getClass.getSimpleName)
    }
  }

  def scn: ScenarioBuilder

  def getInjectionStep(): InjectionStep = {
    perfType match {
      case "soak" | "normal" => atOnceUsers(concurrency)
      case "peak" | "webui"  => rampUsers(concurrency) over new FiniteDuration(rampUp, TimeUnit.SECONDS)
      case "spike"           => throw new NotImplementedError("Performance type \"spike\" not implemented.")
      case "stress" => constantUsersPerSec(settings.simulation.injection.usersPerSec) during
        new FiniteDuration(durationTime, TimeUnit.SECONDS) randomized
      case _ => throw new UnsupportedOperationException("Performance type not supported.")
    }
  }

  def sendEvent(title: String, tags: String*): ListenableFuture[asynchttpclient.Response] = {
    asyncHttpClient
      .preparePost(s"${settings.influx.endpoint}&precision=s")
      .setBody(s"""events title="$title",tags="${tags.mkString(",")}"""")
      .execute()
  }
}
