/*
 * Copyright (C) 2018 OX Software GmbH
 *
 * This file is part of OX API-Performance Tests.
 *
 * OX API-Performance Tests is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * OX API-Performance Tests is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OX API-Performance Tests.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.openexchange.testing.performance.common.utils

import java.time.LocalDate
import java.util.concurrent.ThreadLocalRandom

import com.openexchange.testing.performance.common.utils.DateTimeHelper._
import com.openexchange.testing.performance.common.utils.FakerHelper.faker
import com.openexchange.testing.performance.groupware.{ Appointment, DetailedAppointmentData, Participant, TaskAppointmentData }
import org.dmfs.rfc5545.{ DateTime, Duration }

/**
 * [[AppointmentGenerator]]
 *
 * @author <a href="mailto:nikolaos.tsapanidis@open-xchange.com">Nikolaos Tsapanidis</a>
 */
object AppointmentGenerator {
  private val tlRandom: ThreadLocalRandom = ThreadLocalRandom.current()

  def generateRandomAppointment(folderId: String, participants: Seq[Participant], rangeDays: Int = 0, allowRecurring: Boolean = true): Appointment = {

    var startDate: DateTime = null
    var endDate: DateTime = null
    while (startDate == null || endDate == null) {
      startDate = DateTimeHelper.forward()
      val durationToAdd = Duration.parse(
        s"+PT${tlRandom.nextInt(0, 24)}H${tlRandom.nextInt(0, 60)}M")
      try {
        endDate = startDate.addDuration(durationToAdd)
      } catch {
        case e: java.lang.IllegalArgumentException => {
          println("Warning: AppointmentGenerator.generateRandomAppointment: caught java.lang.IllegalArgumentException, Message: " + e.getMessage)
          println("Warning: DEBUG: startDate=" + startDate.toString() + ", DEBUG: durationToAdd=" + durationToAdd.toString())
          println("Warning: trying again.")
          //e.printStackTrace
        }
      }
    }
    val recType: Int = if (allowRecurring) ThreadLocalRandom.current().nextInt(0, 4) else 0

    val taskAppData = TaskAppointmentData(
      Some(folderId),
      title = Some(s"${faker.name().firstName()}-${faker.name().lastName()}-${faker.address().city()}"),
      start_date = startDate,
      end_date = endDate,
      note = Some(faker.lorem().paragraph()),
      notification = Some(false),
      private_flag = Some(tlRandom.nextBoolean()),
      participants = Some(participants))
    val detailedAppData = DetailedAppointmentData(
      location = Some(faker.address().city()),
      shown_as = Some(tlRandom.nextInt(1, 4 + 1)),
      full_time = Some(tlRandom.nextBoolean()),
      ignore_conflicts = Some(true))

    recType match {
      case 0 => Appointment(taskAppData, detailedAppData)
      case 1 => Appointment(
        taskAppData.copy(
          recurrence_type = Some(recType),
          interval = Some(1),
          occurrences = Some(ThreadLocalRandom.current().nextInt(1, 99 + 1))),
        detailedAppData)
      case 2 => Appointment(
        taskAppData.copy(
          recurrence_type = Some(recType),
          interval = Some(1),
          occurrences = Some(ThreadLocalRandom.current().nextInt(1, 99 + 1)),
          days = Some(ThreadLocalRandom.current().nextInt(1, 127 + 1))),
        detailedAppData)
      case 3 => Appointment(
        taskAppData.copy(
          recurrence_type = Some(recType),
          interval = Some(1),
          occurrences = Some(ThreadLocalRandom.current().nextInt(1, 99 + 1)),
          day_in_month = Some(LocalDate.now.getDayOfMonth) //TODO .get
        ),
        detailedAppData)
      case _ => throw new IllegalArgumentException(s"Wrong recurrence type $recType")
    }
  }

}
