/*
 * Copyright (C) 2018 OX Software GmbH
 *
 * This file is part of OX API-Performance Tests.
 *
 * OX API-Performance Tests is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * OX API-Performance Tests is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OX API-Performance Tests.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.openexchange.testing.performance.common.utils

import java.time.temporal.ChronoUnit
import java.util.TimeZone
import java.util.concurrent.ThreadLocalRandom

import com.openexchange.testing.performance.groupware.chronos.{ Alarm, DateTimeData, EventData, Trigger }
import org.dmfs.rfc5545.Duration
import org.dmfs.rfc5545.recur.RecurrenceRule.{ Part, WeekdayNum }
import org.dmfs.rfc5545.recur.{ Freq, RecurrenceRule }
import org.scalacheck.Gen

import scala.collection.JavaConverters._

/**
 * [[ChronosEventGenerator]]
 *
 * @author <a href="mailto:nikolaos.tsapanidis@open-xchange.com">Nikolaos Tsapanidis</a>
 */
object ChronosEventGenerator {
  private val tlRandom = ThreadLocalRandom.current()

  private val genBool: Gen[Boolean] = Gen.oneOf(true, false)

  private val genTimeZones: Gen[String] = Gen.oneOf("Pacific/Samoa", "Pacific/Honolulu", "America/Anchorage",
    "America/Tijuana", "America/Chihuahua", "America/Chicago", "America/Kentucky/Louisville", "America/New_York",
    "America/Puerto_Rico", "America/Argentina/Buenos_Aires", "Africa/Casablanca", "Europe/London",
    "Europe/Berlin", "Europe/Athens", "Asia/Qatar", "Europe/Moscow", "Asia/Tokyo", "Australia/Sydney", "Pacific/Fiji")

  private val genFrequency: Gen[Freq] = Gen.frequency((3, Freq.DAILY), (2, Freq.MONTHLY), (3, Freq.WEEKLY),
    (1, Freq.YEARLY))

  private val genRuleByDay: Gen[Seq[WeekdayNum]] = Gen.atLeastOne(List[WeekdayNum](
    WeekdayNum.valueOf("SU"), WeekdayNum.valueOf("MO"), WeekdayNum.valueOf("TU"),
    WeekdayNum.valueOf("WE"), WeekdayNum.valueOf("TH"), WeekdayNum.valueOf("FR"),
    WeekdayNum.valueOf("SA")))

  private val rrule: Option[String] = for {
    freq <- genFrequency.sample
    byDay <- genRuleByDay.sample
    byMonth <- Gen.choose(1, 12).sample // Gen.atLeastOne(Range(1, 12)).sample
    byMonthDay <- Gen.choose(1, 28).sample // Gen.atLeastOne(Range(1, 28)).sample
    count <- Gen.choose(1, 99).sample
    setpos <- Gen.oneOf(-1, 1).sample
    freqMonthly <- Gen.oneOf(Part.BYDAY, Part.BYMONTHDAY).sample
  } yield {
    val rule = new RecurrenceRule(freq)

    rule.getFreq match {
      case Freq.DAILY | Freq.WEEKLY => rule.setByDayPart(byDay.asJava)
      case Freq.MONTHLY => freqMonthly match {
        case Part.BYDAY =>
          rule.setByDayPart(byDay.asJava)
          rule.setByPart(Part.BYSETPOS, setpos)
        case Part.BYMONTHDAY => rule.setByPart(Part.BYMONTHDAY, byMonth)
        case _               => throw new IllegalArgumentException(s"Wrong part $freqMonthly")
      }
      case Freq.YEARLY =>
        rule.setByPart(Part.BYMONTHDAY, byMonthDay)
        rule.setByPart(Part.BYMONTH, byMonth)
      // rule.setByPart(Part.BYMONTHDAY, byMonthDay.map(x => Integer.valueOf(x)).sorted.asJava)
      // rule.setByPart(Part.BYMONTH, byMonth.map(x => Integer.valueOf(x)).sorted.asJava)
      case _ => throw new IllegalArgumentException(s"Wrong frequency $freq")
    }
    rule.setCount(count)
    rule.toString
  }

  private val genAlarmDuration: Gen[Option[Int]] = {
    val range = Seq(0, 5, 10, 15, 30, 45, 120)
    Gen.oneOf(Gen.oneOf(range).sample, Gen.oneOf(range).sample.map(n => -n))
  }

  // recurrence: 0 == auto: previous behavior: random 9:1 recurrent vs non-recurrent
  //             1 == recurrent
  //             _ == non-recurrent
  def genEvent(recurrence: Int = 0): Option[EventData] = for {
    transp <- Gen.oneOf("OPAQUE", "TRANSPARENT").sample
    clazz <- Gen.oneOf("PUBLIC", "PRIVATE", "CONFIDENTIAL").sample
    allDayFlag <- genBool.sample
    timeZone <- genTimeZones.sample
    summary <- Gen.alphaNumStr.sample
    color <- Gen.option("#%06x".format(tlRandom.nextInt(1 << 25)).toUpperCase).sample
    alarmAction <- Gen.frequency((9, "DISPLAY"), (1, "AUDIO")).sample
    triggerRelated <- Gen.oneOf("START", "END").sample
    triggerDuration <- genAlarmDuration.sample
    recurrenceRule <- recurrence match {
      case 0 => Gen.frequency((9, Gen.const(None)), (1, Gen.some(rrule.get))).sample
      // WTH... how to syntactically get this right without the Gen. thing?
      case 1 => Gen.frequency((1, Gen.some(rrule.get))).sample
      case _ => Gen.frequency((1, Gen.const(None))).sample
    }
  } yield {
    val start = DateTimeHelper.forward(tz = TimeZone.getTimeZone(timeZone), allDay = allDayFlag)
    val end = start.addDuration(Duration.parse(
      s"+PT${tlRandom.nextInt(0, 24)}H${tlRandom.nextInt(0, 60)}M"))
    EventData(
      startDate = Some(DateTimeData(Some(start.toString), Some(TimeZone.getDefault))),
      endDate = Some(DateTimeData(Some(end.toString), Some(TimeZone.getDefault))),
      clazz = Some(clazz),
      transp = Some(transp),
      alarms = Some(Array(Alarm(action = Some(alarmAction), description = Some("reminder"), trigger = Some(Trigger(
        Some(DateTimeHelper.getDurationString(triggerDuration.get, ChronoUnit.MINUTES)), related = Some(triggerRelated)))))),
      summary = Some(summary),
      color = color,
      rrule = recurrenceRule)
  }

}