/*
 * Copyright (C) 2018 OX Software GmbH
 *
 * This file is part of OX API-Performance Tests.
 *
 * OX API-Performance Tests is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * OX API-Performance Tests is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OX API-Performance Tests.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.openexchange.testing.performance.common.utils

import java.time.temporal.ChronoUnit
import java.util.TimeZone
import java.util.concurrent.ThreadLocalRandom

import org.dmfs.rfc5545.{ DateTime, Duration }

import scala.language.implicitConversions

/**
 * [[DateTimeHelper]]
 *
 * @author <a href="mailto:nikolaos.tsapanidis@open-xchange.com">Nikolaos Tsapanidis</a>
 */
object DateTimeHelper {
  private val tlRandom = ThreadLocalRandom.current()
  private val timezone = DateTime.UTC
  private val calendarMetrics = DateTime.GREGORIAN_CALENDAR_SCALE

  def now(tz: TimeZone = timezone): DateTime = DateTime.now(tz)

  def birthday(minAge: Int = 18, maxAge: Int = 65): DateTime = {
    val now = DateTime.now(timezone)
    val bounds = prepareBounds(now, minAge, maxAge)

    between(
      new DateTime(calendarMetrics, bounds._1, now.getMonth, now.getDayOfMonth),
      new DateTime(calendarMetrics, bounds._2, now.getMonth, now.getDayOfMonth)).toAllDay
  }

  def forOrBackward(n: Int, unit: ChronoUnit, tz: TimeZone = timezone): DateTime = now(tz)
    .addDuration(Duration.parse(getDurationString(n, unit)))

  /** Generates a random date in the future. */
  def forward(days: Int = 365, tz: TimeZone = timezone, allDay: Boolean = false): DateTime = {
    val from = DateTime.now(tz)
    val to = DateTime.now(tz).addDuration(randDuration(days, allDay))

    between(from, to)
  }

  /** Generates a random date in the past. */
  def backward(days: Int = -365, tz: TimeZone = timezone, allDay: Boolean = false): DateTime = {
    val from = DateTime.now(tz).addDuration(randDuration(days, allDay))
    val to = DateTime.now(tz)

    between(from, to)
  }

  def randDuration(days: Int, allDay: Boolean): Duration = {
    if (allDay)
      new Duration(days.signum, Math.abs(days), 0, 0, 0)
    else
      new Duration(days.signum, Math.abs(days), tlRandom.nextInt(0, 24),
        tlRandom.nextInt(0, 60), tlRandom.nextInt(0, 60))
  }

  def between(from: DateTime, to: DateTime): DateTime = randInRange(from, to)

  def getDurationString(n: Int, unit: ChronoUnit): String = {
    val sign = if (n.signum == -1) "-" else "+"

    val durationString = unit match {
      case ChronoUnit.YEARS   => s"${sign}P${365 * Math.abs(n)}D"
      case ChronoUnit.MONTHS  => s"${sign}P${30 * Math.abs(n)}D"
      case ChronoUnit.WEEKS   => s"${sign}P${4 * Math.abs(n)}D"
      case ChronoUnit.DAYS    => s"${sign}P${Math.abs(n)}D"
      case ChronoUnit.HOURS   => s"${sign}PT${Math.abs(n)}H"
      case ChronoUnit.MINUTES => s"${sign}PT${Math.abs(n)}M"
      case ChronoUnit.SECONDS => s"${sign}PT${Math.abs(n)}S"
      case _                  => throw new IllegalArgumentException(s"Unsupported ChronoUnit. $unit")
    }
    durationString
  }

  private def prepareBounds(now: DateTime, minAge: Int, maxAge: Int): (Int, Int) = (now.getYear - minAge, now
    .getYear - maxAge)

  private def randInRange(from: DateTime, to: DateTime): DateTime = {
    val month: Int = tlRandom.nextInt(1, 13)

    val dayOfMonth = month match {
      case 1 | 3 | 5 | 7 | 8 | 10 | 12 => tlRandom.nextInt(1, 32)
      case 2 => tlRandom.nextInt(1, DateTime.GREGORIAN_CALENDAR_SCALE
        .getDaysPerPackedMonth(from.getYear, month - 1))
      case 4 | 6 | 9 | 11 => tlRandom.nextInt(1, 31)
    }

    if (from.getYear == to.getYear)
      new DateTime(calendarMetrics, timezone,
        from.getYear, month - 1, dayOfMonth, from.getHours, from.getMinutes, to.getSeconds)
    else if (to.before(from))
      new DateTime(calendarMetrics, timezone,
        tlRandom.nextInt(to.getYear, from.getYear), month - 1, dayOfMonth, to.getHours, to.getMinutes, to.getSeconds)
    else
      new DateTime(calendarMetrics, timezone,
        tlRandom.nextInt(from.getYear, to.getYear), month - 1, dayOfMonth, from.getHours, from.getMinutes, to.getSeconds)
  }

  implicit def localDateTime2Long(date: DateTime): Long = date.getTimestamp
}
