/*
 * Copyright (C) 2018 OX Software GmbH
 *
 * This file is part of OX API-Performance Tests.
 *
 * OX API-Performance Tests is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * OX API-Performance Tests is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OX API-Performance Tests.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.openexchange.testing.performance.common.utils

import java.io.{ File, FileInputStream }
import java.nio.charset.StandardCharsets
import java.nio.file.{ Files, Path, Paths }
import java.security.MessageDigest
import java.util.concurrent.ThreadLocalRandom

import com.typesafe.scalalogging.LazyLogging

import scala.collection.JavaConverters._

/**
 * [[FileHelper]]
 *
 * @author <a href="mailto:nikolaos.tsapanidis@open-xchange.com">Nikolaos Tsapanidis</a>
 */
object FileHelper extends LazyLogging {
  val EMPTY_MD5 = "d41d8cd98f00b204e9800998ecf8427e"
  val pathOfResourcesDir: Path = Paths.get("src/test/resources/bodies")
  val fileTypes: Seq[String] = Seq("txt", "png", "jpg", "bmp")

  def generateTempFiles(path: Path, files: Int, fileSize: Int): Unit = {
    for (i <- 1 to files) {
      val fileType: String = fileTypes(ThreadLocalRandom.current().nextInt(fileTypes.length))

      val tmpFile: File = Files.createTempFile(path, "", s".$fileType").toFile

      val data: Array[Byte] = getByteArray(fileType, fileSize)
      write(tmpFile, data)
      tmpFile.deleteOnExit()
    }
  }

  def getByteArray(fileType: String, fileSize: Int): Array[Byte] = {
    val data: Array[Byte] = fileType match {
      case "png" | "jpg" | "jpeg" | "bmp" => RandomImgFaker(format = fileType)
      case "txt" => {
        val tmpData: Array[Byte] = new Array[Byte](ThreadLocalRandom.current().nextInt(fileSize))
        ThreadLocalRandom.current().nextBytes(tmpData)
        tmpData
      }
      case _ => Array.emptyByteArray
    }
    data
  }

  def createTempDir(prefix: String, baseDir: Path = pathOfResourcesDir): Path = {
    Files.createTempDirectory(baseDir, prefix)
  }

  /**
   * Write content to a file.
   *
   * @param file The temp file
   * @param data The content.
   * @return
   */
  def write(file: File, data: Array[Byte]): Path = {
    Files.write(Paths.get(file.getAbsolutePath), data)
  }

  /**
   * Read from a file.
   *
   * @param filePath
   * @return
   */
  def read(filePath: String): String = {
    Files.readAllLines(Paths.get(filePath), StandardCharsets.UTF_8).asScala.mkString
  }

  def calculateMD5FromFile(f: File): String = {
    val md = MessageDigest.getInstance("MD5")
    val in = new FileInputStream(f)
    val buffer = new Array[Byte](1024)

    Stream.continually(in.read(buffer))
      .takeWhile(_ != -1)
      .foreach(md.update(buffer, 0, _))

    md.digest().map(0xFF & _).map {
      "%02x".format(_)
    }.foldLeft("") {
      _ + _
    }
  }

  /**
   * Computes the MD5 digest for a byte array.
   *
   * @param bytes the data to hash
   * @return the MD5 digest, encoded as a hex string
   */
  def md5(bytes: Array[Byte]): String = {
    val digest = MessageDigest.getInstance("MD5")
    digest.reset()
    digest.update(bytes)
    digest.digest().map(0xFF & _).map {
      "%02x".format(_)
    }.foldLeft("") {
      _ + _
    }
  }

  def getSortedFiles(dir: String): Set[File] = collection.SortedSet.empty[File] ++ {
    val d = new File(dir)
    if (d.exists && d.isDirectory) {
      d.listFiles.filter(_.isFile).toList
    } else {
      List[File]()
    }
  }

  def getAllFiles(dir: File): List[File] =
    dir.listFiles.toList.flatMap { f =>
      if (f.isFile) List(f)
      else getAllFiles(f)
    }
}
