/*
 * Copyright (C) 2018 OX Software GmbH
 *
 * This file is part of OX API-Performance Tests.
 *
 * OX API-Performance Tests is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * OX API-Performance Tests is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OX API-Performance Tests.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.openexchange.testing.performance.config

import java.util.concurrent.TimeUnit

import com.openexchange.testing.performance.config.ConfigKeys._
import com.typesafe.config.ConfigFactory

import scala.collection.JavaConverters._

/**
 * [[Configuration]]
 *
 * @author <a href="mailto:nikolaos.tsapanidis@open-xchange.com">Nikolaos Tsapanidis</a>
 */
trait Configuration {
  private lazy val defaultConfig = ConfigFactory.load(getClass.getClassLoader, "ox-defaults.conf")
  private lazy val userConfig = ConfigFactory.load(getClass.getClassLoader, "ox.conf")
  private lazy val config = userConfig.withFallback(defaultConfig).resolve()

  lazy val settings = Settings(
    appsuite = AppSuite(
      uiWebPath = config.getString(appsuite.UiWebPath),
      modules = config.getObject("ox.appsuite.modules").unwrapped.asScala.map { case (k, v) => (k, v.toString.toDouble) }.toMap),
    facade = Facade(
      path = config.getString(facade.Path),
      baseUrl = s"${config.getString(facade.Protocol)}://${config.getString(facade.Host)}:${config.getInt(facade.Port)}",
      version = config.getString(facade.Version)),
    middleware = Middleware(
      path = config.getString(middleware.DispatcherPath),
      baseUrl = s"${config.getString(middleware.Protocol)}://${config.getString(middleware.Host)}:" +
        s"${config.getInt(middleware.Port)}",
      authentication = config.getString(middleware.Authentication),
      basicAuthUser = config.getOptionalString(middleware.BasicAuthUser),
      basicAuthPass = config.getOptionalString(middleware.BasicAuthPass),
      ws = WS(
        baseUrl = s"${config.getString(middleware.ws.Protocol)}://${config.getString(middleware.ws.Host)}:" +
          s"${config.getInt(middleware.ws.Port)}",
        path = config.getString(middleware.ws.Path))),
    mim = MiM(
      path = config.getString(mim.Path),
      baseUrl = s"${config.getString(mim.Protocol)}://${config.getString(mim.Host)}:${config.getInt(mim.Port)}",
      tenant = config.getString(mim.Tenant),
      ongageSecret = config.getString(mim.OngageSecret),
      campaignPlaceholderType = config.getString(mim.CampaignPlaceholderType),
      campaignPlaceholderAbsolute = config.getInt(mim.CampaignPlaceholderAbsolute),
      campaignPlaceholderPercentage = config.getDouble(mim.CampaignPlaceholderPercentage),
      campaignTimeType = config.getString(mim.CampaignTimeType),
      cava = Cava(config.getInt(mim.cava.N), config.getInt(mim.cava.Offset), config.getInt(mim.cava.Onset)),
      campaignCount = config.getInt(mim.CampaignCount),
      campaignScenarioUsers = config.getInt(mim.CampaignScenarioUsers),
      campaignUserPattern = config.getString(mim.CampaignUserPattern),
      campaignDurationTime = config.getDuration(mim.CampaignDurationTime, TimeUnit.SECONDS),
      tmpFilePath = config.getString(mim.TmpFilePath),
      aws = AWS(
        accessKey = config.getString(mim.aws.AccessKey),
        secretKey = config.getString(mim.aws.SecretKey),
        endpoint = config.getString(mim.aws.Endpoint),
        bucket = config.getString(mim.aws.Bucket),
        region = config.getString(mim.aws.Region),
        pathStyleAccess = config.getBoolean(mdb.aws.PathStyleAccess),
        signer = config.getOptionalString(mim.aws.Signer)),
      cassandraEndpoints = config.getStringList(mim.CassandraEndpoints).asScala.toList),
    mdb = MDB(
      baseUrl = s"${config.getString(mdb.Protocol)}://${config.getString(mdb.Host)}:${config.getInt(mdb.Port)}",
      tenantList = config.getStringList(mdb.TenantList).asScala.toList,
      tenantSecretKey = config.getString(mdb.TenantSecretKey),
      usersPerTenant = config.getInt(mdb.UsersPerTenant),
      espImportId = config.getString(mdb.EspImportId),
      s3Import = config.getBoolean(mdb.S3Import),
      duration = config.getDuration(mdb.DurationTime, TimeUnit.SECONDS),
      aws = AWS(
        accessKey = config.getString(mdb.aws.AccessKey),
        secretKey = config.getString(mdb.aws.SecretKey),
        endpoint = config.getString(mdb.aws.Endpoint),
        bucket = config.getString(mdb.aws.Bucket),
        region = config.getString(mdb.aws.Region),
        pathStyleAccess = config.getBoolean(mdb.aws.PathStyleAccess),
        signer = config.getOptionalString(mim.aws.Signer)),
      tmpFilePath = config.getString(mdb.tmpFilePath)),
    influx = {
      val protocol = config.getOptionalString(metrics.Protocol)
      val host = config.getOptionalString(metrics.Host)
      val port = config.getOptionalInt(metrics.Port)
      val db = config.getOptionalString(metrics.DB)

      if (protocol.isEmpty || host.isEmpty || port.isEmpty || db.isEmpty) {
        Influx("")
      } else {
        Influx(s"${protocol.get}://${host.get}:${port.get}/write?db=${db.get}")
      }
    },
    simulation = Simulation(
      performanceType = config.getOptionalString(simulation.Type),
      injection = Injection(
        duration = config.getDuration(simulation.injection.DurationTime, TimeUnit.SECONDS),
        rampup = config.getInt(simulation.injection.Rampup),
        users = config.getInt(simulation.injection.Users),
        usersPerSec = config.getInt(simulation.injection.UsersPerSec),
        loopTime = config.getDuration(simulation.injection.LoopTime, TimeUnit.SECONDS),
        repeatCounter = config.getInt(simulation.injection.RepeatCounter)),
      data = Data(
        keepData = config.getBoolean(simulation.data.KeepData),
        locale = config.getString(simulation.data.Locale),
        user = User(
          domain = config.getString(simulation.data.user.Domain)),
        drive = Drive(
          numberOfFiles = config.getInt(simulation.data.drive.NumberOfFiles),
          numberOfFolder = config.getInt(simulation.data.drive.NumberOfFolder),
          maxFileSize = config.getBytes(simulation.data.drive.MaxFileSize)),
        mail = Mail(
          numberOfMails = config.getOptionalInt(simulation.data.mail.NumberOfMails),
          conversationView = config.getOptionalBoolean(simulation.data.mail.ConversationView)),
        contacts = Contacts(
          numberOfContacts = config.getOptionalInt(simulation.data.contacts.NumberOfContacts)),
        calendar = Calendar(
          numberOfAppointments = config.getOptionalInt(simulation.data.calendar.NumberOfAppointments),
          numberOfRecurringAppointments = config.getOptionalInt(simulation.data.calendar.NumberOfRecurringAppointments)),
        tasks = Tasks(numberOfTasks = config.getOptionalInt(simulation.data.tasks.NumberOfTasks)),
        probabilities = Probabilities(
          nonIdle = config.getOptionalInt(simulation.data.probabilities.NonIdle),
          mail = config.getOptionalInt(simulation.data.probabilities.Mail),
          calendar = config.getOptionalInt(simulation.data.probabilities.Calendar),
          contacts = config.getOptionalInt(simulation.data.probabilities.Contacts),
          tasks = config.getOptionalInt(simulation.data.probabilities.Tasks),
          drive = config.getOptionalInt(simulation.data.probabilities.Drive))),
      assertions = Assertions(config.getInt(simulation.assertions.SuccessfulRequests)),
      userFeeder = config.getString(simulation.UserFeeder),
      userAgent = config.getString(simulation.UserAgent)),
    soap = Soap(
      startId = config.getInt(soap.StartId),
      startName = config.getOptionalInt(soap.StartName),
      endId = config.getInt(soap.EndId),
      maxQuota = config.getInt(soap.MaxQuota),
      adminmaster = config.getString(soap.Adminmaster),
      admin = config.getString(soap.Admin),
      adminpassword = config.getString(soap.Adminpassword),
      masterpassword = config.getString(soap.Masterpassword),
      userpassword = config.getString(soap.Userpassword),
      reseller = config.getBoolean(soap.Reseller),
      accessCombinationName = config.getString(soap.AccessCombinationName),
      imapServer = config.getString(soap.ImapServer),
      smtpServer = config.getString(soap.SmtpServer),
      filestoreId = config.getOptionalInt(soap.FilestoreId),
      writeDBPoolId = config.getOptionalInt(soap.WriteDBPoolId),
      schemaFeeder = config.getOptionalString(soap.SchemaFeeder),
      missingFeeder = config.getOptionalString(soap.MissingFeeder),
      createContext = config.getOptionalInt(soap.CreateContext),
      usersPerContext = config.getOptionalInt(soap.UsersPerContext),
      contextFeeder = config.getOptionalString(soap.ContextFeeder),
      userFeeder = config.getOptionalString(soap.UserFeeder),
      capabilitiesToAdd = config.getOptionalString(soap.CapabilitiesToAdd),
      capabilitiesToRemove = config.getOptionalString(soap.CapabilitiesToRemove),
      capabilitiesToDrop = config.getOptionalString(soap.CapabilitiesToDrop)))
}

case class AppSuite(uiWebPath: String, modules: Map[String, Double])

case class Assertions(successfulRequests: Int)

case class AWS(accessKey: String, secretKey: String, endpoint: String, bucket: String, region: String, pathStyleAccess: Boolean, signer: Option[String])

case class Cava(n: Int, offset: Int, onset: Int)

case class Data(
    keepData:      Boolean,
    locale:        String,
    user:          User,
    drive:         Drive,
    mail:          Mail,
    contacts:      Contacts,
    calendar:      Calendar,
    tasks:         Tasks,
    probabilities: Probabilities)

case class Mail(numberOfMails: Option[Int], conversationView: Option[Boolean])

case class Contacts(numberOfContacts: Option[Int])

case class Calendar(numberOfAppointments: Option[Int], numberOfRecurringAppointments: Option[Int])

case class Tasks(numberOfTasks: Option[Int])

case class Probabilities(
    nonIdle:  Option[Int],
    mail:     Option[Int],
    calendar: Option[Int],
    tasks:    Option[Int],
    contacts: Option[Int],
    drive:    Option[Int])

case class Drive(numberOfFiles: Int, numberOfFolder: Int, maxFileSize: Long)

case class Facade(baseUrl: String, path: String, version: String)

case class Injection(duration: Long, rampup: Int, users: Int, usersPerSec: Int, loopTime: Long, repeatCounter: Int)

case class Influx(endpoint: String)

case class Middleware(path: String, authentication: String, baseUrl: String, basicAuthUser: Option[String],
                      basicAuthPass: Option[String], ws: WS)

case class MiM(
    baseUrl:                       String,
    path:                          String,
    tenant:                        String,
    ongageSecret:                  String,
    campaignPlaceholderType:       String,
    campaignPlaceholderAbsolute:   Int,
    campaignPlaceholderPercentage: Double,
    campaignTimeType:              String,
    campaignDurationTime:          Long,
    cava:                          Cava,
    campaignCount:                 Int,
    campaignScenarioUsers:         Int,
    campaignUserPattern:           String,
    aws:                           AWS,
    tmpFilePath:                   String,
    cassandraEndpoints:            List[String])

case class MDB(
    baseUrl:         String,
    tenantList:      List[String],
    tenantSecretKey: String,
    usersPerTenant:  Int,
    espImportId:     String,
    s3Import:        Boolean,
    duration:        Long,
    aws:             AWS,
    tmpFilePath:     String)

case class Settings(appsuite: AppSuite, facade: Facade, middleware: Middleware, mim: MiM, mdb: MDB,
                    influx: Influx, simulation: Simulation, soap: Soap)

case class Simulation(performanceType: Option[String], injection: Injection, assertions: Assertions, data: Data,
                      userFeeder: String, userAgent: String)

case class Soap(
    startId:               Int,
    startName:             Option[Int],
    endId:                 Int,
    maxQuota:              Int,
    adminmaster:           String,
    admin:                 String,
    masterpassword:        String,
    adminpassword:         String,
    userpassword:          String,
    reseller:              Boolean,
    accessCombinationName: String,
    imapServer:            String,
    smtpServer:            String,
    filestoreId:           Option[Int],
    writeDBPoolId:         Option[Int],
    schemaFeeder:          Option[String],
    missingFeeder:         Option[String],
    createContext:         Option[Int],
    usersPerContext:       Option[Int],
    contextFeeder:         Option[String],
    userFeeder:            Option[String],
    capabilitiesToAdd:     Option[String],
    capabilitiesToRemove:  Option[String],
    capabilitiesToDrop:    Option[String])

case class User(domain: String)

case class WS(baseUrl: String, path: String)
