/*
 * Copyright (C) 2018 OX Software GmbH
 *
 * This file is part of OX API-Performance Tests.
 *
 * OX API-Performance Tests is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * OX API-Performance Tests is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OX API-Performance Tests.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.openexchange.testing.performance.groupware.folders

/**
 * [[PermissionFlags]]
 *
 * @author <a href="mailto:nikolaos.tsapanidis@open-xchange.com">Nikolaos Tsapanidis</a>
 */

object PermissionFlags {

  sealed abstract class Permissions(val value: Long, val bit: Long) {
    lazy val permissionBits: Long = value << bit

    override def toString: String = permissionBits.toString
  }

  //================================================================================
  // Folder permissions (0-6)
  //================================================================================
  case object FOLDER_NO_PERMISSIONS extends Permissions(0, 0)
  case object FOLDER_SEE_FOLDER extends Permissions(1, 0)
  case object FOLDER_CREATE_OBJECTS extends Permissions(2, 0)
  case object FOLDER_CREATE_SUBFOLDERS extends Permissions(4, 0)
  case object FOLDER_ADMIN extends Permissions(64, 0)

  //================================================================================
  // Read permissions for objects in the folder (7-13)
  //================================================================================
  case object READ_NO_PERMISSIONS extends Permissions(0, 7)
  case object READ_ONLY_OWN_OBJECTS extends Permissions(1, 7)
  case object READ_ALL_OBJECTS extends Permissions(2, 7)
  case object READ_ADMIN extends Permissions(64, 7)

  //================================================================================
  // Write permissions for objects in the folder (14-20)
  //================================================================================
  case object WRITE_NO_PERMISSIONS extends Permissions(0, 14)
  case object WRITE_ONLY_OWN_OBJECTS extends Permissions(1, 14)
  case object WRITE_ALL_OBJECTS extends Permissions(2, 14)
  case object WRITE_ADMIN extends Permissions(64, 14)

  //================================================================================
  // Delete permissions for objects in the folder (21-27)
  //================================================================================
  case object DELETE_NO_PERMISSIONS extends Permissions(0, 21)
  case object DELETE_ONLY_OWN_OBJECTS extends Permissions(1, 21)
  case object DELETE_ALL_OBJECTS extends Permissions(2, 21)
  case object DELETE_ADMIN extends Permissions(64, 21)

  //================================================================================
  // Admin flag (28)
  //================================================================================
  case object ADMIN_NO_PERMISSIONS extends Permissions(0, 28)
  case object ADMIN_ADMIN extends Permissions(1, 28)

  /** Grants full permission to the folder. */
  val OWNER = FOLDER_ADMIN.permissionBits |
    READ_ADMIN.permissionBits |
    WRITE_ADMIN.permissionBits |
    DELETE_ADMIN.permissionBits |
    ADMIN_ADMIN.permissionBits

  /** Grants full permission to the folder, but not possible to change the folder owner. */
  val PUBLISHING_EDITOR = FOLDER_CREATE_SUBFOLDERS.permissionBits |
    READ_ALL_OBJECTS.permissionBits |
    WRITE_ALL_OBJECTS.permissionBits |
    WRITE_ALL_OBJECTS.permissionBits |
    ADMIN_NO_PERMISSIONS.permissionBits

  /** Grants full permission to the folder, except the ability to create new folders. */
  val EDITOR = FOLDER_CREATE_OBJECTS.permissionBits |
    READ_ALL_OBJECTS.permissionBits |
    WRITE_ALL_OBJECTS.permissionBits |
    DELETE_ALL_OBJECTS.permissionBits |
    ADMIN_NO_PERMISSIONS.permissionBits

  /** Similar to [[PUBLISHING_EDITOR]], except the granted user can't delete files from owner. */
  val PUBLISHING_AUTHOR = FOLDER_CREATE_SUBFOLDERS.permissionBits |
    READ_ALL_OBJECTS.permissionBits |
    WRITE_ONLY_OWN_OBJECTS.permissionBits |
    DELETE_ONLY_OWN_OBJECTS.permissionBits |
    ADMIN_NO_PERMISSIONS.permissionBits

  /** Similar to [[EDITOR]], except the granted user can't delete files from owner. */
  val AUTHOR = FOLDER_CREATE_OBJECTS.permissionBits |
    READ_ALL_OBJECTS.permissionBits |
    WRITE_ONLY_OWN_OBJECTS.permissionBits |
    DELETE_ONLY_OWN_OBJECTS.permissionBits |
    ADMIN_NO_PERMISSIONS.permissionBits

  /** Use can create items, not folders, but can't editing anything. */
  val NONEDITING_AUTHOR = FOLDER_CREATE_OBJECTS.permissionBits |
    READ_ALL_OBJECTS.permissionBits |
    WRITE_NO_PERMISSIONS.permissionBits |
    DELETE_NO_PERMISSIONS.permissionBits |
    ADMIN_NO_PERMISSIONS.permissionBits

  /** User can see folders. */
  val REVIEWER = FOLDER_SEE_FOLDER.permissionBits |
    READ_ALL_OBJECTS.permissionBits |
    WRITE_NO_PERMISSIONS.permissionBits |
    DELETE_NO_PERMISSIONS.permissionBits |
    ADMIN_NO_PERMISSIONS.permissionBits

  /** User can create items. */
  val CONTRIBUTOR = FOLDER_CREATE_OBJECTS.permissionBits |
    READ_NO_PERMISSIONS.permissionBits |
    WRITE_NO_PERMISSIONS.permissionBits |
    DELETE_NO_PERMISSIONS.permissionBits |
    ADMIN_NO_PERMISSIONS.permissionBits

  /** User has no permissions. */
  val NONE = FOLDER_NO_PERMISSIONS.permissionBits |
    READ_NO_PERMISSIONS.permissionBits |
    WRITE_NO_PERMISSIONS.permissionBits |
    DELETE_ONLY_OWN_OBJECTS.permissionBits |
    ADMIN_NO_PERMISSIONS.permissionBits
}