/*
 * Copyright (C) 2018 OX Software GmbH
 *
 * This file is part of OX API-Performance Tests.
 *
 * OX API-Performance Tests is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * OX API-Performance Tests is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OX API-Performance Tests.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.openexchange.testing.performance.groupware.mail

import com.fasterxml.jackson.annotation.JsonUnwrapped
import com.openexchange.testing.performance.groupware.BaseObject

/**
 * [[Mail]]
 *
 * @author <a href="mailto:nikolaos.tsapanidis@open-xchange.com">Nikolaos Tsapanidis</a>
 */
case class Mail(@JsonUnwrapped data: MailData)

/**
 * [[MailAttachment]]
 *
 * @author <a href="mailto:nikolaos.tsapanidis@open-xchange.com">Nikolaos Tsapanidis</a>
 * @param content_type MIME type. plain text ("text/plain"), as html ("text/html") or as multipart/alternative ("ALTERNATIVE")
 * @param content Content as text. Present only if easily convertible to text.
 * @param filename Displayed filename (mutually exclusive with content).
 * @param size Size of the attachment in bytes.
 * @param disp Attachment's disposition: null, inline, attachment or alternative.
 */
case class MailAttachment(
    content_type: String,
    content:      String,
    filename:     Option[String]           = None,
    size:         Option[Int]              = None,
    disp:         Option[String]           = None,
    uploaded:     Option[Int]              = None,
    meta:         Option[Map[String, Any]] = None,
    raw:          Option[Boolean]          = None)

/**
 * [[MailSecurity]]
 *
 * @author <a href="mailto:philipp.Schumacher@open-xchange.com">Philipp Schumacher</a>
 * @param encrypt Enable encryption.
 * @param pgpInline Enable legacy coding PGP/INLINE.
 * @param sign Set to sign the mail.
 */
case class MailSecurity(
    encrypt:   Boolean,
    pgpInline: Boolean,
    sign:      Boolean)

/**
 * [[MailData]]
 *
 * @author <a href="mailto:nikolaos.tsapanidis@open-xchange.com">Nikolaos Tsapanidis</a>
 * @param from Each element is a two-element array specifying one sender. The first element of
 *             each address is the personal name, the second element is the email address.
 * @param to Each element is a two-element array (see the from field) specifying one receiver.
 * @param cc Each element is a two-element array (see the from field) specifying one carbon-copy receiver.
 * @param bcc Each element is a two-element array (see the from field) specifying one blind carbon-copy receiver.
 * @param subject Subject line.
 * @param priority Value of message's ''X-Priority'' header:
 * @param dispositionNotificationTo Content of message's header ''Disposition-Notification-To''.
 * @param security Guard security options
 * @param attachments Each element is an attachment as described in [[MailAttachment]].
 * @param infostore_ids Defines a JSON array of infostore document ID(s).
 */
case class MailData(
    from:                      Array[Array[String]],
    to:                        Array[Array[String]],
    subject:                   String,
    priority:                  Int,
    vcard:                     Option[Int]                   = None,
    cc:                        Option[Array[Array[String]]]  = None,
    bcc:                       Option[Array[Array[String]]]  = None,
    dispositionNotificationTo: Option[String]                = None,
    headers:                   Option[Map[String, Any]]      = None,
    security:                  Option[MailSecurity]          = None,
    attachments:               Option[Array[MailAttachment]] = None,
    contacts_ids:              Option[Array[String]]         = None,
    infostore_ids:             Option[Array[String]]         = None,
    dataSources:               Option[Array[String]]         = None) extends BaseObject