/*
 * Copyright (C) 2018 OX Software GmbH
 *
 * This file is part of OX API-Performance Tests.
 *
 * OX API-Performance Tests is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * OX API-Performance Tests is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OX API-Performance Tests.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.openexchange.testing.performance.simulations

import java.io.{ File, IOException }
import java.nio.file._
import java.security.MessageDigest
import java.util.concurrent.TimeUnit

import com.openexchange.testing.performance.ajax.config.ConfigAPI
import com.openexchange.testing.performance.ajax.folders.FoldersAPI
import com.openexchange.testing.performance.ajax.login.LoginAPI
import com.openexchange.testing.performance.common.SessionConstants._
import com.openexchange.testing.performance.common.utils.FileHelper._
import com.openexchange.testing.performance.config.{ Configuration, HTTPConfig }
import com.openexchange.testing.performance.drive._
import com.typesafe.scalalogging.LazyLogging
import io.gatling.core.Predef._
import org.apache.commons.io.FileUtils

import scala.concurrent.duration._

/**
 * [[QAPerfOXDrive]] - Simulates an OXDrive-Client
 *
 * @author <a href="mailto:nikolaos.tsapanidis@open-xchange.com">Nikolaos Tsapanidis</a>
 */
class QAPerfOXDrive extends Simulation with Configuration with LazyLogging {

  // 1. Create temp directory
  private val clientTempDir: Path = createTempDir(getClass.getSimpleName)

  // 2. Generate random temp files
  generateTempFiles(clientTempDir, settings.simulation.data.drive.numberOfFiles,
    settings.simulation.data.drive.maxFileSize.toInt)

  private val allFilesInFolder: Set[File] = getSortedFiles(clientTempDir.toString)

  private val directoryChecksum = {
    val digest = MessageDigest.getInstance("MD5")

    digest.reset()
    // for (i <- allFilesInFolder.toArray) yield digest.update(Files.readAllBytes(Paths.get(i.getName)))
    for (i <- allFilesInFolder.toArray) yield {
      digest.update(i.getName.getBytes("UTF-8")); digest.update(calculateMD5FromFile(i).getBytes)
    }

    digest.digest().map(0xFF & _).map {
      "%02x".format(_)
    }.foldLeft("") {
      _ + _
    }
  }

  private val clientOriginalVersion: Array[DirectoryVersion] = Array(DirectoryVersion("/", EMPTY_MD5))
  private val clientDirectoryVersion: Array[DirectoryVersion] = Array(
    DirectoryVersion("/", EMPTY_MD5), DirectoryVersion("/" + clientTempDir.getFileName.toString, directoryChecksum))

  private val clientFileVersion: Array[FileVersion] = for (file <- allFilesInFolder.toArray) yield FileVersion(Some(file.getName), Some(calculateMD5FromFile(file)))
  private val clientOriginalFileVersion: Array[FileVersion] = Array.empty[FileVersion]

  before {
    logger.debug("Created folder {}", clientTempDir)
  }

  after {
    try {
      // Wipe all data on the client (files and directory)
      logger.debug("Deleted all files in {}", clientTempDir)

      FileUtils.deleteQuietly(clientTempDir.toFile)
      logger.debug("Deleted folder {}", clientTempDir)

    } catch {
      case e: NoSuchFileException        => logger.error("File not exists.")
      case e: DirectoryNotEmptyException => logger.error("Directory {} not empty.", clientTempDir)
      case e: IOException                => e.printStackTrace()
    }

  }

  val scn = scenario(getClass.getSimpleName)
    .during(new FiniteDuration(settings.simulation.injection.duration - (settings.simulation.injection.duration /
      settings.simulation.injection.rampup), TimeUnit.SECONDS), exitASAP = false) {
      feed(csv(settings.simulation.userFeeder).circular)
        .tryMax(3) {
          exec(LoginAPI.login())
        }

        .doIf(session => session.contains(SESSION_ID)) {
          exec(ConfigAPI.getDefaultFolderInfostore)
            .exec(ConfigAPI.getDefaultFolderInfostoreTrash)
            .pause(250 milliseconds, 500 milliseconds)

            // clear the infostore folder
            .exec(FoldersAPI.clearFolders(Array($(FID_INFOSTORE))))
            .pause(50 milliseconds, 100 milliseconds)

            .exec(DriveAPI.getDriveQuota($(FID_INFOSTORE)))
            .pause(50 milliseconds, 100 milliseconds)

            .exec(DriveAPI.synchronizeFolders($(FID_INFOSTORE), Some(clientOriginalVersion), clientDirectoryVersion))

            .exec(DriveAPI.synchronizeFiles($(FID_INFOSTORE), "/" + clientTempDir.getName(4), Some(clientOriginalFileVersion), clientFileVersion))

            .foreach("${syncFilesResponse}", "action") {
              doSwitch(session => session("action").as[SynchronizeFilesResponse].action.get)(
                "acknowledge" -> exec(),
                "edit" -> exec(),
                "download" -> exec(),
                "upload" -> exec(DriveAPI.upload($(FID_INFOSTORE), "${action}")),
                "remove" -> exec(),
                "sync" -> exec(),
                "error" -> exec())
            }

            .pause(2 seconds)

            // Delete-Action
            .tryMax(10) {
              exec(DriveAPI.synchronizeFolders($(FID_INFOSTORE), Some(clientDirectoryVersion), clientOriginalVersion))
            }

            .pause(2 seconds)

            // Finally clear the infostore trash folder
            .exec(FoldersAPI.clearFolders(Array($(FID_INFOSTORE_TRASH))))

        }
        .exec(LoginAPI.logout)
    }

  setUp(scn.inject(rampUsers(settings.simulation.injection.users) over new FiniteDuration(settings.simulation.injection
    .duration / settings.simulation.injection.rampup, TimeUnit.SECONDS)))
    .protocols(HTTPConfig.appsuiteConf
      .acceptCharsetHeader("application/json, text/javascript, */*; q=0.01"))
    .assertions(global.successfulRequests.percent.gte(settings.simulation.assertions.successfulRequests))
}