/*
 * Copyright (C) 2018 OX Software GmbH
 *
 * This file is part of OX API-Performance Tests.
 *
 * OX API-Performance Tests is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * OX API-Performance Tests is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OX API-Performance Tests.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.openexchange.testing.performance.simulations

import java.io.File
import java.util.concurrent.TimeUnit

import com.openexchange.testing.performance.aws.AwsS3Client
import com.typesafe.scalalogging.LazyLogging
import com.openexchange.testing.performance.common.utils.CryptoHelper
import com.openexchange.testing.performance.config.{ Configuration, HTTPConfig }
import com.openexchange.testing.performance.mdb._
import com.openexchange.testing.performance.util.CSVGenerator

import scala.concurrent.duration.{ FiniteDuration, _ }
import scala.collection.mutable
import io.gatling.core.Predef._
import io.gatling.core.structure.ScenarioBuilder

/**
 * [[MDBSimulation]]
 *
 * @author <a href="mailto:philipp.schumacher@open-xchange.com">Philipp Schumacher</a>
 */
class MDBSimulation extends Simulation with Configuration with LazyLogging {

  private val s3Client: AwsS3Client = new AwsS3Client(
    settings.mdb.aws.accessKey,
    settings.mdb.aws.secretKey,
    settings.mdb.aws.endpoint,
    settings.mdb.aws.region,
    settings.mdb.aws.bucket,
    settings.mdb.aws.pathStyleAccess,
    settings.mim.aws.signer)

  private val listOfTenants: mutable.ArrayStack[String] = mutable.ArrayStack() ++= settings.mdb.tenantList

  before {
    if (settings.mdb.s3Import) {
      // Clean S3 bucket before uploading new CSV files
      val tenant = listOfTenants.pop()
      s3Client.deleteBucketsContent(settings.mdb.aws.bucket)
      CSVGenerator.createCSVonFilesystem(settings.mdb.usersPerTenant, tenant)
      s3Client.createObject(settings.mdb.aws.bucket, tenant + "/" + settings.mdb.espImportId + "/" + "import-" + tenant + ".csv", new File(s"${settings.mdb.tmpFilePath}/import-" + tenant + ".csv"))
      listOfTenants.push(tenant)
    }
  }

  private val mdbScenario: ScenarioBuilder = scenario("mdbScenario")

    // Create a new tenant from given list
    .exec(session => {
      session.set("tenantName", listOfTenants.pop())
    })
    .exec(new POSTTenants(true, new com.openexchange.testing.performance.util.Tenant(
      id = "${tenantName}",
      name = "${tenantName}",
      keyspaceName = "mdb_${tenantName}",
      configuration = new com.openexchange.testing.performance.util.Configuration(
        secretKey = settings.mdb.tenantSecretKey,
        shutdown = false,
        ldapBrandName = "${tenantName}"),
      dataSources = List(new com.openexchange.testing.performance.util.DataSource(
        id = "${tenantName}_dataSource_1",
        typ = "ongage",
        defaultDataSource = true,
        secretKey = settings.mdb.tenantSecretKey,
        importFilename = "import.csv",
        deleteTemporaryImportFiles = false,
        awsConnectionInformation = new com.openexchange.testing.performance.util.AwsConnectionInformation(
          authorizationType = "aws",
          accessKey = settings.mdb.aws.accessKey,
          secretKey = settings.mdb.aws.secretKey,
          signer = null,
          pathStyleAccessEnabled = false,
          endpoint = null,
          region = null,
          bucket = null,
          proxyHost = null,
          proxyPort = null,
          proxyUsername = null,
          proxyPassword = null))),
      customFieldConfigurations = null)))

    // Import CSV data either from S3 or via file upload
    .doIfOrElse(settings.mdb.s3Import) {
      exec(session => {
        session.set("hash", CryptoHelper.hmacSha1(session("tenantName").as[String] + "now", settings.mdb.tenantSecretKey))
      })
        .doIfOrElse(session => !session.contains("tenantDataSourceId")) {
          exec(new GETTenantsDataSources("${tenantName}"))
            .exec(new GETTenantsDataSources("${tenantName}", Some("${tenantDataSourcesIds.random()}")))
        } {
          exec(new GETTenantsDataSources("${tenantName}", Some("${tenantDataSourcesIds.random()}")))
        }
        .exec(new POSTUserImportsS3("${tenantName}", "${tenantDataSourcesIds.random()}", settings.mdb.espImportId, "now", "${hash}"))
    } {
      exec(session => {
        session.set("csv", CSVGenerator.createCSV(settings.mdb.usersPerTenant, session("tenantName").as[String]))
      })
        .exec(session => {
          session.set("hash", CryptoHelper.hmacSha1(session("tenantName").as[String] + "now", settings.mdb.tenantSecretKey))
        })
        .exec(new POSTUserImportsFormData("${tenantName}", "now", "${hash}", "${csv}"))
    }
    .exec(new GETUserImports("${tenantName}", Some("${importId}")))
    .asLongAs(session => session("result").as[String] == "pending") {
      exec(new GETUserImports("${tenantName}", Some("${importId}")))
        .pause(1 seconds)
    }
    .exec(new GETUserImportsErrors("${tenantName}", "${importId}"))
    .exec(session => {
      session.removeAll("csv", "hash")
    })

    // Execute random actions
    .during(new FiniteDuration(settings.simulation.injection.loopTime, TimeUnit.SECONDS)) {
      randomSwitch(
        10d -> exec(new GETTenants()),
        10d -> exec(new GETTenants(Some("${tenantName}"))),
        10d -> exec(new GETTenantsConfig("${tenantName}")),
        10d -> exec(new GETTenantsDataSources("${tenantName}")),
        10d -> doIfOrElse(session => !session.contains("tenantDataSourceId")) {
          exec(new GETTenantsDataSources("${tenantName}"))
            .exec(new GETTenantsDataSources("${tenantName}", Some("${tenantDataSourcesIds.random()}")))
        } {
          exec(new GETTenantsDataSources("${tenantName}", Some("${tenantDataSourcesIds.random()}")))
        },
        //TODO /shutdown/all - Enable tenant again after shutdown
        10d -> exec(new PUTTenantsAllShutdown),
        10d -> doIfOrElse(session => !session.contains("tenantConfig")) {
          exec(new GETTenantsConfig("${tenantName}"))
            .exec(new PUTTenantsConfig("${tenantName}", "${tenantConfig}"))
        } {
          exec(new PUTTenantsConfig("${tenantName}", "${tenantConfig}"))
        },
        10d -> doIfOrElse(session => !session.contains("tenantDataSources")) {
          exec(new GETTenantsDataSources("${tenantName}"))
            .exec(new PUTTenantsDataSources("${tenantName}", "${tenantDataSources}"))
        } {
          exec(new PUTTenantsDataSources("${tenantName}", "${tenantDataSources}"))
        },
        10d -> exec(new GETUserImports("${tenantName}")),
        //TODO /tenants/:tenantId/:userId - How provide userId from LDAP? i.e. '3@1'
        /*0d -> exec(session => {
          session.set("userId", "3@1")
        })
          .exec(new GETUsers("${tenantName}", "${userId}")),*/
        10d -> exec(new GETDebug("${tenantName}"))
      //TODO /debug/:tenantId/:userId - How provide userId from LDAP? i.e. '3@1'
      /*0d -> doIfOrElse(session => !session.contains("userId")) {
            exec(session => {
              session.set("userId", "3@1")
            })
              .exec(new GETDebug(tenantId = "${tenantName}", Some("${userId}")))
          } {
            exec(new GETDebug(tenantId = "${tenantName}", Some("${userId}")))
          } */
      )
    }
    .exec(new DELETETenants("${tenantName}"))
    .exec(session => {
      listOfTenants.push(session("tenantName").as[String])
      session
    })

  setUp(mdbScenario.inject(atOnceUsers(settings.mdb.tenantList.size))).protocols(HTTPConfig.mdbConf)

}
