/*
 * Copyright (C) 2018 OX Software GmbH
 *
 * This file is part of OX API-Performance Tests.
 *
 * OX API-Performance Tests is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * OX API-Performance Tests is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OX API-Performance Tests.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.openexchange.testing.performance.simulations

import java.io.File
import java.nio.file.{ Files, Paths }
import java.text.SimpleDateFormat
import java.util.{ Calendar, Locale }

import com.openexchange.testing.performance.aws.AwsS3Client
import com.openexchange.testing.performance.common.utils.CryptoHelper
import com.openexchange.testing.performance.config.{ Configuration, HTTPConfig }
import com.openexchange.testing.performance.mim._
import com.openexchange.testing.performance.util.CampaignGenerator
import com.typesafe.scalalogging.LazyLogging
import io.gatling.core.Predef._
import io.gatling.core.structure.ScenarioBuilder

import scala.collection.mutable
import scala.collection.mutable.HashMap
import scala.concurrent.duration._

/**
 * [[ImportAsyncSimulation]]
 *
 * @author <a href="mailto:nikolaos.tsapanidis@open-xchange.com">Nikolaos Tsapanidis</a>
 * @author <a href="mailto:philipp.schumacher@open-xchange.com">Philipp Schumacher</a>
 * @author <a href="mailto:alexander.schulze-ardey@open-xchange.com">Alexander Schulze-Ardey</a>
 */
class ImportAsyncSimulation extends Simulation with Configuration with LazyLogging {

  private val s3Client: AwsS3Client = new AwsS3Client(
    settings.mim.aws.accessKey,
    settings.mim.aws.secretKey,
    settings.mim.aws.endpoint,
    settings.mim.aws.region,
    settings.mim.aws.bucket,
    settings.mim.aws.pathStyleAccess,
    settings.mim.aws.signer)
  private val dateFormat = new SimpleDateFormat("EEE, dd MMM YYYY HH:mm:ss XXX", Locale.ENGLISH)

  before {
    val startCleaningS3Bucket = System.currentTimeMillis()
    if (s3Client.listBucketsContent().nonEmpty) {
      logger.info("start cleaning bucket \"{}\"", settings.mim.aws.bucket)
      s3Client.deleteBucketsContent()
    }

    logger.info("cleaning s3 bucket took: {} ms", System.currentTimeMillis() - startCleaningS3Bucket)

    logger.info("start generating campaign data")
    val startGeneratingCampaignData = System.currentTimeMillis()

    val generator = new CampaignGenerator
    val guids = generator.generateCampaigns(settings.mim.campaignCount)

    logger.info("campaign generation took: {} ms", System.currentTimeMillis() - startGeneratingCampaignData)
    val startUploadingCampaignData = System.currentTimeMillis()
    for (i <- guids) {
      s3Client.createObject(settings.mim.aws.bucket, s"${settings.mim.tenant}/$i/template.json", new File(s"${settings.mim.tmpFilePath}/template-$i.json"))
      s3Client.createObject(settings.mim.aws.bucket, s"${settings.mim.tenant}/$i/placeholders.json", new File(s"${settings.mim.tmpFilePath}/placeholders-$i.json"))
      s3Client.createObject(settings.mim.aws.bucket, s"${settings.mim.tenant}/$i/metadata.json", new File(s"${settings.mim.tmpFilePath}/metadata-$i.json"))
      val placeholdersJSONPath = Paths.get(s"${settings.mim.tmpFilePath}/placeholders-$i.json")
      val templateJSONPath = Paths.get(s"${settings.mim.tmpFilePath}/template-$i.json")
      val metadataJSONPath = Paths.get(s"${settings.mim.tmpFilePath}/metadata-$i.json")
      Files.delete(placeholdersJSONPath)
      Files.delete(templateJSONPath)
      Files.delete(metadataJSONPath)
    }
    logger.info("uploaded new campaign(s) to s3 storage: {} ms", System.currentTimeMillis() - startUploadingCampaignData)
  }

  private val importScenario: ScenarioBuilder = scenario("generateCampaigns")
    .asLongAs(session => !session.contains("matadataFileCount")) {
      exec(session => {
        val files = s3Client.listBucketsContent(settings.mim.aws.bucket).filter(_.getKey.endsWith("metadata.json"))
        val pattern = "[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}".r
        if (files.length == settings.mim.campaignCount) {
          session.set("matadataFileCount", files.length).set("metadataJSONGuids", files.map(x => pattern findFirstIn x.getKey).toList.flatten)
        } else {
          session
        }
      })
    }
    .exec(session => {
      session.set("importIdMap", mutable.HashMap[String, String]())
    })
    .foreach(session => session("metadataJSONGuids").as[List[String]], "campaignGuid") {
      exec(session => {
        session.set("date", dateFormat.format(Calendar.getInstance.getTime))
      })
        .exec(session => {
          val hash = CryptoHelper.hmacSha1(session("campaignGuid").as[String] + session("date").as[String], settings.mim.ongageSecret)
          session.set("hash", hash)
        })
        .exec(session => {
          logger.info("start importing campaign {}", session("campaignGuid").as[String])
          session.set("startImportCampaign", System.currentTimeMillis())
        })
        .exec(new ImportAsyncCampaign(
          tenantId = settings.mim.tenant,
          campaignId = "${campaignGuid}",
          date = "${date}",
          hash = "${hash}"))
        .exec(session => {
          val importId = session("importId").as[String]
          val campaignGuid = session("campaignGuid").as[String]
          val importIdMap = session("importIdMap").as[HashMap[String, String]]
          session.set("importIdMap", importIdMap += campaignGuid -> importId)
        })
    }
    .asLongAs(session => !session("importIdMap").as[mutable.HashMap[String, String]].isEmpty) {
      foreach(session => session("importIdMap").as[mutable.HashMap[String, String]].keySet.toSeq, "campaignGuid") {
        exec(session => {
          val importIdMap = session("importIdMap").as[mutable.HashMap[String, String]]
          val importId = importIdMap.get(session("campaignGuid").as[String]).get
          session.set("currentImportId", importId)
        })
          .exec(new ImportCampaignStatus(
            settings.mim.tenant,
            "${campaignGuid}",
            "${currentImportId}"))
          .doIf(session => session("result").as[String] == "successful") {
            exec(session => {
              val importIdMap = session("importIdMap").as[mutable.HashMap[String, String]]
              val newImportIdMap = importIdMap.remove(session("currentImportId").as[String])
              session.set("importIdMap", newImportIdMap)
            })
          }
      }
        .pause(5 seconds)
    }

  setUp(importScenario.inject(atOnceUsers(1))).protocols(HTTPConfig.mimConf)
}

