/*
 * Copyright (C) 2018 OX Software GmbH
 *
 * This file is part of OX API-Performance Tests.
 *
 * OX API-Performance Tests is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * OX API-Performance Tests is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OX API-Performance Tests.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.openexchange.testing.performance.simulations

import java.io.File
import java.nio.file.{ Files, Paths }
import java.text.SimpleDateFormat
import java.util.concurrent.ThreadLocalRandom
import java.util.{ Calendar, Locale }

import com.amazonaws.util.IOUtils
import com.fasterxml.jackson.databind.ObjectMapper
import com.openexchange.testing.performance.aws.AwsS3Client
import com.openexchange.testing.performance.common.utils.CryptoHelper
import com.openexchange.testing.performance.config.{ Configuration, HTTPConfig }
import com.openexchange.testing.performance.mim._
import com.openexchange.testing.performance.util.CampaignGenerator
import com.typesafe.scalalogging.LazyLogging
import io.gatling.core.Predef._
import io.gatling.core.structure.ScenarioBuilder
import io.gatling.jsonpath.JsonPath

import scala.collection.mutable.ArrayBuffer
import scala.concurrent.duration._

/**
 * [[MiMSimulation]]
 *
 * @author <a href="mailto:nikolaos.tsapanidis@open-xchange.com">Nikolaos Tsapanidis</a>
 * @author <a href="mailto:philipp.schumacher@open-xchange.com">Philipp Schumacher</a>
 */
class MiMSimulation extends Simulation with Configuration with LazyLogging {

  private val s3Client: AwsS3Client = new AwsS3Client(
    settings.mim.aws.accessKey,
    settings.mim.aws.secretKey,
    settings.mim.aws.endpoint,
    settings.mim.aws.region,
    settings.mim.aws.bucket,
    settings.mim.aws.pathStyleAccess,
    settings.mim.aws.signer)
  private val dateFormat = new SimpleDateFormat("EEE, dd MMM YYYY HH:mm:ss XXX", Locale.ENGLISH)
  private val arrayOfCampaignsGuids: ArrayBuffer[String] = ArrayBuffer[String]()

  before {
    val startCleaningS3Bucket = System.currentTimeMillis()
    if (s3Client.listBucketsContent().nonEmpty) {
      logger.info("start cleaning bucket \"{}\"", settings.mim.aws.bucket)
      s3Client.deleteBucketsContent()
    }
    logger.info("cleaning s3 bucket took: {} ms", System.currentTimeMillis() - startCleaningS3Bucket)

    logger.info("start generating campaign data")
    val startGeneratingCampaignData = System.currentTimeMillis()

    val generator = new CampaignGenerator
    val arrayOfCampaignsGuids = generator.generateCampaigns(settings.mim.campaignCount)
    logger.info("campaign generation took: {} ms", System.currentTimeMillis() - startGeneratingCampaignData)

    val startUploadingCampaignData = System.currentTimeMillis()
    for (i <- arrayOfCampaignsGuids) {
      s3Client.createObject(settings.mim.aws.bucket, s"$i/template.json", new File(s"${settings.mim.tmpFilePath}/template-$i.json"))
      s3Client.createObject(settings.mim.aws.bucket, s"$i/placeholders.json", new File(s"${settings.mim.tmpFilePath}/placeholders-$i.json"))
      val placeholdersJSONPath = Paths.get(s"${settings.mim.tmpFilePath}/placeholders-$i.json")
      val templateJSONPath = Paths.get(s"${settings.mim.tmpFilePath}/template-$i.json")
      Files.delete(placeholdersJSONPath)
      Files.delete(templateJSONPath)
    }
    s3Client.createObject(settings.mim.aws.bucket, "metadata.json", new File(s"${settings.mim.tmpFilePath}/metadata.json"))
    val metadataJSON = Paths.get(s"${settings.mim.tmpFilePath}/metadata.json")
    Files.delete(metadataJSON)
    logger.info("uploaded new campaign(s) to s3 storage: {} ms", System.currentTimeMillis() - startUploadingCampaignData)
  }

  private val importScenario: ScenarioBuilder = scenario("generateCampaigns")
    .asLongAs(session => !session.contains("metadataJSONGuids")) {
      exec(session => {
        val metadataJSONStream = s3Client.getObject(settings.mim.aws.bucket, "metadata.json")
        if (metadataJSONStream != null) {
          val metadataJSON = IOUtils.toString(metadataJSONStream)
          val json = (new ObjectMapper).readValue(metadataJSON, classOf[Object])
          val metadataJSONGuids = JsonPath.query("$..guid", json).map(_.toList)
          metadataJSONGuids match {
            case Right(it) => session.set("metadataJSONGuids", it)
            case Left(_)   => session
          }
        } else {
          session
        }
      })
    }
    .foreach(session => session("metadataJSONGuids").as[List[String]], "campaignGuid") {
      exec(session => {
        session.set("date", dateFormat.format(Calendar.getInstance.getTime))
      })
        .exec(session => {
          val hash = CryptoHelper.hmacSha1(session("campaignGuid").as[String] + session("date").as[String], settings.mim.ongageSecret)
          session.set("hash", hash)
        })
        .exec(session => {
          logger.info("start importing campaign {}", session("campaignGuid").as[String])
          session.set("startImportCampaign", System.currentTimeMillis())
        })
        .exec(new ImportCampaign(
          date = "${date}",
          guid = "${campaignGuid}",
          hash = "${hash}"))
        .exec(session => {
          logger.info("campaign import of {} took: {} ms", session("campaignGuid").as[String], System.currentTimeMillis() - session("startImportCampaign").as[Long])
          session
        })
    }

  private val campaignScenario: ScenarioBuilder = scenario("campaignScenario")
    .during(settings.simulation.injection.duration seconds) {
      exec(session => {
        session.set("startChooseRandomUser", System.currentTimeMillis())
      })
        .exec(session => {
          session.set("randomUser", s"user-${ThreadLocalRandom.current().nextInt(settings.mim.campaignPlaceholderAbsolute)}")
        })
        .exec(session => {
          logger.debug("choosing random user took: {} ms", System.currentTimeMillis() - session("startChooseRandomUser").as[Long])
          session
        })
        .exec(session => {
          session.set("createAndSetUserHash", System.currentTimeMillis())
        })
        .exec(session => {
          session.set("randomUserHash", CryptoHelper.sha256(session("randomUser").as[String]))
        })
        .exec(session => {
          logger.debug("creating user hash took: {} ms", System.currentTimeMillis() - session("createAndSetUserHash").as[Long])
          session
        })
        .exec(new GetUserCampaigns("${randomUserHash}").buildHttpRequest())
        .doIf(session => session.contains("allUserCampaignGuids")) {
          exec(session => {
            session.set("createRandomGuid", System.currentTimeMillis())
          })
            .exec(session => {
              val campaignVector = session("allUserCampaignGuids").as[Vector[Any]]
              session.set("randomCampaignGuid", campaignVector(ThreadLocalRandom.current().nextInt(campaignVector.length)))
            })
            .exec(session => {
              logger.debug("creating random guid took: {} ms", System.currentTimeMillis() - session("createRandomGuid").as[Long])
              session
            })
            .exec(new GetCampaign("${randomCampaignGuid}"))

        }
        .randomSwitch(20d -> exec(new GetAllCampaigns()))
    }

  setUp(
    importScenario.inject(atOnceUsers(1)),
    campaignScenario.inject(atOnceUsers(settings.mim.campaignScenarioUsers))).protocols(HTTPConfig.mimConf)
}

