/*
 * Copyright (C) 2018 OX Software GmbH
 *
 * This file is part of OX API-Performance Tests.
 *
 * OX API-Performance Tests is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * OX API-Performance Tests is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OX API-Performance Tests.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.openexchange.testing.performance.simulations

import java.io.File
import java.nio.file.{ Files, Paths }
import java.text.SimpleDateFormat
import java.util.concurrent.ThreadLocalRandom
import java.util.{ Calendar, Locale }

import com.openexchange.testing.performance.aws.AwsS3Client
import com.openexchange.testing.performance.common.utils.CryptoHelper
import com.openexchange.testing.performance.config.{ Configuration, HTTPConfig }
import com.openexchange.testing.performance.mim._
import com.openexchange.testing.performance.util.CampaignGenerator
import com.typesafe.scalalogging.LazyLogging
import io.gatling.core.Predef._
import io.gatling.core.structure.ScenarioBuilder

import scala.collection.mutable.ArrayBuffer
import scala.concurrent.duration._

/**
 * [[ReferenceAsyncMiMSimulation]]
 *
 * @author <a href="mailto:nikolaos.tsapanidis@open-xchange.com">Nikolaos Tsapanidis</a>
 * @author <a href="mailto:philipp.schumacher@open-xchange.com">Philipp Schumacher</a>
 */
class ReferenceAsyncMiMSimulation extends Simulation with Configuration with LazyLogging {

  private val s3Client: AwsS3Client = new AwsS3Client(
    settings.mim.aws.accessKey,
    settings.mim.aws.secretKey,
    settings.mim.aws.endpoint,
    settings.mim.aws.region,
    settings.mim.aws.bucket,
    settings.mim.aws.pathStyleAccess,
    settings.mim.aws.signer)
  private val dateFormat = new SimpleDateFormat("EEE, dd MMM YYYY HH:mm:ss XXX", Locale.ENGLISH)
  private var guids: ArrayBuffer[String] = _

  before {
    val startCleaningS3Bucket = System.currentTimeMillis()
    if (s3Client.listBucketsContent().nonEmpty) {
      logger.info("start cleaning bucket \"{}\"", settings.mim.aws.bucket)
      s3Client.deleteBucketsContent()
    }

    logger.info("cleaning s3 bucket took: {} ms", System.currentTimeMillis() - startCleaningS3Bucket)

    logger.info("start generating campaign data")
    val startGeneratingCampaignData = System.currentTimeMillis()

    val generator = new CampaignGenerator
    guids = generator.generateCampaigns(settings.mim.campaignCount)

    logger.info("campaign generation took: {} ms", System.currentTimeMillis() - startGeneratingCampaignData)
    logger.info("before: guids content: {}", guids)
  }

  private val importScenario: ScenarioBuilder = scenario("importCampaigns")
    .exec(session => {
      logger.info("importScenario: guids content: {}", guids)
      session.set("campaignGuids", guids)
    })
    .foreach(session => session("campaignGuids").as[Seq[String]], "campaignGuid") {
      exec(session => {
        // Upload campaign to S3
        val startS3Upload = System.currentTimeMillis()
        val i = session("campaignGuid").as[String]
        logger.info("importScenario.foreach: current guid: {}", i)
        s3Client.createObject(settings.mim.aws.bucket, s"${settings.mim.tenant}/$i/template.json", new File(s"${settings.mim.tmpFilePath}/template-$i.json"))
        s3Client.createObject(settings.mim.aws.bucket, s"${settings.mim.tenant}/$i/placeholders.json", new File(s"${settings.mim.tmpFilePath}/placeholders-$i.json"))
        s3Client.createObject(settings.mim.aws.bucket, s"${settings.mim.tenant}/$i/metadata.json", new File(s"${settings.mim.tmpFilePath}/metadata-$i.json"))
        val placeholdersJSONPath = Paths.get(s"${settings.mim.tmpFilePath}/placeholders-$i.json")
        val templateJSONPath = Paths.get(s"${settings.mim.tmpFilePath}/template-$i.json")
        val metadataJSONPath = Paths.get(s"${settings.mim.tmpFilePath}/metadata-$i.json")
        Files.delete(placeholdersJSONPath)
        Files.delete(templateJSONPath)
        Files.delete(metadataJSONPath)
        logger.info("S3 upload of campaign {} took: {} ms", i, System.currentTimeMillis() - startS3Upload)
        session
      })
        .exec(session => {
          session.set("date", dateFormat.format(Calendar.getInstance.getTime))
        })
        .exec(session => {
          val hash = CryptoHelper.hmacSha1(session("campaignGuid").as[String] + session("date").as[String], settings.mim.ongageSecret)
          session.set("hash", hash)
        })
        .exec(session => {
          logger.info("start importing campaign {}", session("campaignGuid").as[String])
          session.set("startImportCampaign", System.currentTimeMillis())
        })
        .exec(new ImportAsyncCampaign(
          tenantId = settings.mim.tenant,
          campaignId = "${campaignGuid}",
          date = "${date}",
          hash = "${hash}"))
        .asLongAs(session => session.contains("importId")) {
          exec(new ImportCampaignStatus(
            settings.mim.tenant,
            "${campaignGuid}",
            "${importId}"))
            .doIf(session => session("result").as[String] == "successful") {
              exec(session => {
                // TODO: cope with error status
                logger.info("done with import of campaign {}", session("campaignGuid").as[String])
                session.remove("importId")
              })
            }
            .pause(5 seconds)
        }
    }

  private val campaignScenario: ScenarioBuilder = scenario("campaignScenario")
    .exec(session => {
      session.set("startChooseRandomUser", System.currentTimeMillis())
    })
    .exec(session => {
      session.set("randomUser", s"${settings.mim.campaignUserPattern}${ThreadLocalRandom.current().nextInt(1, settings.mim.campaignPlaceholderAbsolute)}")
    })
    .exec(session => {
      logger.debug("choosing random '{}' user took: {} ms", session("randomUser").as[String], System.currentTimeMillis() - session("startChooseRandomUser").as[Long])
      session
    })
    .exec(session => {
      session.set("createAndSetUserHash", System.currentTimeMillis())
    })
    .exec(session => {
      session.set("randomUserHash", CryptoHelper.sha256(session("randomUser").as[String]))
    })
    .exec(session => {
      logger.debug("creating user hash {} took: {} ms", session("randomUserHash").as[String], System.currentTimeMillis() - session("createAndSetUserHash").as[Long])
      session
    })
    .exec(new GetUserCampaigns("${randomUserHash}").buildHttpRequest())
    .randomSwitch(0.02d -> doIf(session => session.contains("allUserCampaignGuids")) {
      exec(session => {
        session.set("createRandomGuid", System.currentTimeMillis())
      })
        .exec(session => {
          val campaignVector = session("allUserCampaignGuids").as[Vector[Any]]
          session.set("randomCampaignGuid", campaignVector(ThreadLocalRandom.current().nextInt(campaignVector.length)))
        })
        .exec(session => {
          logger.debug("creating random guid took: {} ms", System.currentTimeMillis() - session("createRandomGuid").as[Long])
          session
        })
        .exec(new GetCampaign("${randomCampaignGuid}").buildHttpRequest())
    })

  setUp(
    importScenario.inject(atOnceUsers(1)),
    campaignScenario.inject(
      rampUsersPerSec(1) to settings.simulation.injection.usersPerSec during (60 seconds),
      constantUsersPerSec(settings.simulation.injection.usersPerSec) during (settings.simulation.injection.duration seconds)))
    .protocols(HTTPConfig.mimConf)

}

