/*
 * Copyright (C) 2018 OX Software GmbH
 *
 * This file is part of OX API-Performance Tests.
 *
 * OX API-Performance Tests is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * OX API-Performance Tests is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OX API-Performance Tests.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.openexchange.testing.performance.util

import java.io.{ BufferedWriter, FileWriter }
import java.nio.charset.CodingErrorAction
import java.time.Instant
import java.util.concurrent.ThreadLocalRandom

import com.openexchange.testing.performance.common.utils.CryptoHelper
import com.openexchange.testing.performance.common.utils.FakerHelper.faker
import com.openexchange.testing.performance.config.Configuration
import com.typesafe.scalalogging.LazyLogging
import org.apache.commons.lang3.{ RandomStringUtils, StringEscapeUtils }

import scala.collection.mutable.ArrayBuffer
import scala.io.Codec
import scala.util.Random

class CampaignGenerator extends Configuration with LazyLogging {

  implicit val codec = Codec("UTF-8")
  codec.onMalformedInput(CodingErrorAction.REPLACE)
  codec.onUnmappableCharacter(CodingErrorAction.REPLACE)

  private val bodyTemplate = "test" //Source.fromResource("bodies/promo_mail_template.eml").getLines.mkString

  private val campaignPlaceholderType = settings.mim.campaignPlaceholderType
  private val campaignPlaceholderAbsolute = settings.mim.campaignPlaceholderAbsolute
  private val campaignPlaceholderPercentage = settings.mim.campaignPlaceholderPercentage
  private val campaignTimeType = settings.mim.campaignTimeType
  private val cavaN = settings.mim.cava.n
  private val cavaOffset = settings.mim.cava.offset
  private val cavaOnset = settings.mim.cava.onset
  private val campaignSecondsValid = settings.mim.campaignDurationTime
  private val campaignUserPattern = settings.mim.campaignUserPattern

  private val random: ThreadLocalRandom = ThreadLocalRandom.current
  private val guids: ArrayBuffer[String] = ArrayBuffer[String]()
  private val startDates = new ArrayBuffer[Instant]
  private val endDates = new ArrayBuffer[Instant]

  private var cavaCounter = 0

  def generateCampaigns(campaignCount: Int): ArrayBuffer[String] = {

    if (campaignPlaceholderPercentage < 0d || campaignPlaceholderPercentage > 100d && campaignPlaceholderType == "Percentage") {
      throw new NumberFormatException("userPercentage needs to be between 0 and 100")
    }

    var placeholdersCount: Long = 0L

    if (campaignPlaceholderType == "Percentage") {
      placeholdersCount = campaignPlaceholderAbsolute * campaignPlaceholderPercentage.toLong / 100L
    } else if (campaignPlaceholderType == "Absolute") {
      placeholdersCount = campaignPlaceholderAbsolute
    } else if (campaignPlaceholderType == "Random") {
      placeholdersCount = random.nextLong(1, campaignPlaceholderAbsolute)
    } else {
      throw new Exception("Unknown campaignPlaceholdersType")
    }

    val startGeneratingCampaignData = System.currentTimeMillis()

    for (i <- 0 to campaignCount - 1) {
      guids += java.util.UUID.randomUUID().toString.toLowerCase
      if (campaignTimeType == "Random") {
        startDates += Instant.now
        endDates += Instant.now.plusSeconds(random.nextLong(1, campaignSecondsValid))
      } else if (campaignTimeType == "CAVA") {
        startDates += Instant.now.plusSeconds((campaignSecondsValid / cavaN) * (cavaCounter) + cavaOnset)
        endDates += Instant.now.plusSeconds((campaignSecondsValid / cavaN) * (cavaCounter + 1) + cavaOffset)
        if (cavaCounter == cavaN - 1) cavaCounter = 0 else cavaCounter += 1
      } else if (campaignTimeType == "Fixed") {
        startDates += Instant.now
        endDates += Instant.now.plusSeconds(campaignSecondsValid)
      } else {
        throw new Exception("Unknown campaignTimeType")
      }
      val replacedBody = bodyTemplate
        .replaceAll("\\{RND_SENTENCE\\}", faker.lorem().sentence(random.nextInt(100)))
        .replaceAll("\\{RND_LINK\\}", faker.internet().image())
        .replaceAll("\\{RND_EMAIL\\}", faker.internet().emailAddress())
        .replaceAll("\\{RND_STRING\\}", faker.lorem().word)
        .replaceAll("\\{RND_GOT_HOUSE\\}", faker.gameOfThrones().house())
      val templatesFW = new FileWriter(s"${settings.mim.tmpFilePath}/template-${guids(i)}.json")
      val placeholdersFW = new FileWriter(s"${settings.mim.tmpFilePath}/placeholders-${guids(i)}.json")
      val templatesBW = new BufferedWriter(templatesFW)
      val placeholdersBW = new BufferedWriter(placeholdersFW)
      //metadata
      val metadataFW = new FileWriter(s"${settings.mim.tmpFilePath}/metadata-${guids(i)}.json")
      val metadataBW = new BufferedWriter(metadataFW)
      metadataBW.write("{")
      metadataBW.write("\"tenantId\":\"" + settings.mim.tenant + "\",")
      metadataBW.write("\"campaignId\":\"" + guids(i) + "\",")
      metadataBW.write("\"timestamp\":" + Instant.now.toEpochMilli + ",")
      metadataBW.write("\"subject\":\"" + faker.lorem().sentence(random.nextInt(8)) + "\",")
      metadataBW.write("\"sender\":\"" + s"${faker.name().fullName()} <${faker.internet().emailAddress()}>" + "\",")
      metadataBW.write("\"startDate\":" + startDates(i).toEpochMilli + ",")
      metadataBW.write("\"endDate\":" + endDates(i).toEpochMilli + ",")
      metadataBW.write("\"positionFixed\":" + Random.nextBoolean() + "")
      metadataBW.write("}")
      metadataBW.close()
      //template
      templatesBW.write("{")
      templatesBW.write("\"campaignId\":\"" + guids(i) + "\",")
      templatesBW.write("\"body\":\"" + StringEscapeUtils.escapeJava(replacedBody) + "\"")
      templatesBW.write("}")
      templatesBW.close()

      //placeholders
      placeholdersBW.write("[")
      for (j <- 1L to placeholdersCount) {
        placeholdersBW.write("{")
        placeholdersBW.write("\"userId\":\"" + CryptoHelper.sha256(campaignUserPattern.format(j)) + "\",")
        placeholdersBW.write("\"ongage_tracking_identifier\":\"" + RandomStringUtils.randomAlphanumeric(64) + "\"")
        placeholdersBW.write("}")
        if (j != placeholdersCount) placeholdersBW.write(",")
      }
      placeholdersBW.write("]")
      placeholdersBW.close()

    }
    logger.info("campaign generation took: {} ms", System.currentTimeMillis() - startGeneratingCampaignData)
    guids
  }

}
