/*
 * Copyright (C) 2017 OX Software GmbH
 *
 * This file is part of OX API-Performance Tests.
 *
 * OX API-Performance Tests is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * OX API-Performance Tests is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OX API-Performance Tests.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.openexchange.testing.performance.simulations

import java.util
import java.util.Locale
import java.util.concurrent.TimeUnit

import com.openexchange.testing.performance.config.{ Configuration, HTTPConfig }
import com.openexchange.testing.performance.soap.webservices.SoapAPI
import com.openexchange.testing.performance.soap.{ SoapFeeder, WSDLServiceName }
import io.gatling.core.Predef._
import io.gatling.core.structure.ScenarioBuilder

import scala.concurrent.duration.FiniteDuration

/* Parameters:
 *
 * simulation.injection.users : concurrency
 * simulation.injection.repeatCounter : iterations   -> total number of contexts provisioned: users*repeatCounter
 * simulation.soap.startId : start context id
 * simulation.soap.startName : integer for generating start context name
 * simulation.soap.usersPerContext : users per context. total number of users provisioned: users*repeatCounter*usersPerContext
 * simulation.soap.createContext : 0|1 : whether to create a context (normal usecase) or only create users (if you have done a createcontext-only situation before)
 */

/**
 * [[CreateContextAndUserSimulation]]
 *
 * @author <a href="mailto:nikolaos.tsapanidis@open-xchange.com">Nikolaos Tsapanidis</a>
 */
class CreateContextAndUserSimulation extends Simulation with Configuration {

  private val usersPerContext = settings.soap.usersPerContext.getOrElse(10)

  val schemaFeeder = csv(settings.soap.schemaFeeder.getOrElse("data/schemas-dummy.csv")).circular.convert {
    case ("writeDBPoolId", string) => string.toInt
    case ("filestoreId", string)   => string.toInt
  }

  val missingFeeder = csv(settings.soap.missingFeeder.getOrElse("data/missing-dummy.csv")).convert {
    case ("ctx_id", string) => string.toInt
  }

  val scn: ScenarioBuilder = scenario(getClass.getSimpleName)
    .exec(session => {
      session.setAll(
        "oxadminmaster" -> settings.soap.adminmaster,
        "masterpassword" -> settings.soap.masterpassword,
        "adminpassword" -> settings.soap.adminpassword,
        "oxadmin" -> settings.soap.admin,
        "userpassword" -> settings.soap.userpassword,
        "ctx_maxQuota" -> settings.soap.maxQuota,
        "language" -> Locale.US.toString)
    })
    .doIf(settings.soap.filestoreId.isDefined) {
      exec(session => session.set("filestoreId", settings.soap.filestoreId.get))
    }
    .doIf(settings.soap.writeDBPoolId.isDefined) {
      exec(session => session.set("writeDBPoolId", settings.soap.writeDBPoolId.get))
    }

    .repeat(settings.simulation.injection.repeatCounter) {
      pace(new FiniteDuration((usersPerContext + (if (settings.soap.createContext.getOrElse(1) > 0) 1 else 0)) * 1000, TimeUnit.MILLISECONDS))
        .feed(SoapFeeder.contextFeeder)
        .doIf(settings.soap.schemaFeeder.isDefined) {
          feed(schemaFeeder)
        }
        .doIf(settings.soap.missingFeeder.isDefined) {
          feed(missingFeeder)
        }
        .doIf(settings.soap.createContext.getOrElse(1) > 0) {
          exec(SoapAPI.createModuleAccessByName(
            WSDLServiceName.OX_CONTEXT_SERVICE,
            Some(settings.soap.accessCombinationName)))
        }
        .exec(session => {
          val displayNames: Set[String] = Set()
          session.set("displayNames", displayNames)
        })
        .repeat(usersPerContext, "userCounter") {
          pace(new FiniteDuration(1000, TimeUnit.MILLISECONDS))
            .exec(session => {
              var doIt: Boolean = true
              // gatling starts counting by 0, I want to have it starting with 1
              val userCounter = session("userCounter").as[Int] + 1
              var userData = SoapFeeder.userMap(session("ctx_name").as[String], userCounter)
              var displayNames = session("displayNames").as[Set[String]]
              while (displayNames(userData("display_name"))) {
                printf("DisplayName %s not unique, re-rolling.\n", userData("display_name"))
                userData = SoapFeeder.userMap(session("ctx_name").as[String], userCounter)
              }
              displayNames += userData("display_name")
              session.setAll(userData).set("displayNames", displayNames)
            })
            .exec(SoapAPI.createUser(WSDLServiceName.OX_USER_SERVICE))
        }
        .exec(session => {
          session.remove("displayNames")
        })
    }

  setUp(scn.inject(atOnceUsers(settings.simulation.injection.users))).protocols(HTTPConfig.appsuiteConf)
}
