/*
 * Copyright (C) 2016 OX Software GmbH
 *
 * This file is part of OX Gatling Performance Tests.
 *
 * OX Gatling Performance Tests is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * OX Gatling Performance Tests is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OX Gatling Performance Tests.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.openexchange.testing.performance.soap

import io.gatling.commons.validation.{ Failure, Success, Validation }
import java.io.{ ByteArrayInputStream, ByteArrayOutputStream }

import javax.xml.bind.{ JAXBContext, Marshaller, Unmarshaller }
import javax.xml.parsers.DocumentBuilderFactory
import javax.xml.soap.{ MessageFactory, SOAPBody, SOAPMessage }
import org.w3c.dom.Document

import scala.reflect._

/**
 * [[JaxbHelper]]
 *
 * @author <a href="mailto:dominik.epple@open-xchange.com">Dominik Epple</a>
 */
object JaxbHelper {

  val jaxbContextContext = JAXBContext.newInstance("com.openexchange.wsdl.context.soap")
  val jaxbUserContext = JAXBContext.newInstance("com.openexchange.wsdl.user.soap")

  def serializeCtx[T: ClassTag](create: T): Validation[String] = {

    val jaxbMarshaller: Marshaller = jaxbContextContext.createMarshaller

    // output pretty printed
    //jaxbMarshaller.setProperty(Marshaller.JAXB_FORMATTED_OUTPUT, true)
    jaxbMarshaller.setProperty(Marshaller.JAXB_FRAGMENT, true)

    val document: Document = DocumentBuilderFactory.newInstance.newDocumentBuilder.newDocument

    jaxbMarshaller.marshal(create, document)
    val soapMessage: SOAPMessage = MessageFactory.newInstance.createMessage

    // cosmetic
    // change namespace from SOAP-ENV to soap
    /*
    soapMessage.getSOAPPart().getEnvelope().removeNamespaceDeclaration("SOAP-ENV")
    soapMessage.getSOAPPart().getEnvelope().addNamespaceDeclaration("soap", "http://www.w3.org/2001/12/soap-envelope")
    soapMessage.getSOAPPart().getEnvelope().setPrefix("soap")
    soapMessage.getSOAPHeader().setPrefix("soap")
    soapMessage.getSOAPBody().setPrefix("soap")
    */
    // cosmetic end

    soapMessage.getSOAPBody.addDocument(document)

    val outputStream: ByteArrayOutputStream = new ByteArrayOutputStream();
    soapMessage.writeTo(outputStream);

    Success(outputStream.toString())

  }

  def getSoapBody(soapResponse: String): SOAPBody = {
    val soapMessage: SOAPMessage = MessageFactory.newInstance().createMessage(null, new ByteArrayInputStream(soapResponse.getBytes()))
    soapMessage.getSOAPBody
  }

  def unserializeCtx[T: ClassTag](soapResponse: String): Validation[T] = {
    val soapBody = getSoapBody(soapResponse)
    if (!soapBody.hasFault) {
      val unmarshaller: Unmarshaller = jaxbContextContext.createUnmarshaller
      val response = unmarshaller.unmarshal(soapBody.extractContentAsDocument)
      response match {
        case r: T => Success(r)
        case _    => Failure("unserializeCtx: SOAP hasFault flag was not set, but unmarshalling failed.")
      }
    } else Failure("unserializeCtx: SOAP hasFault flag was set: Code=" + soapBody.getFault.getFaultCode + ", String=" + soapBody.getFault.getFaultString)
  }

  def serializeUser[T: ClassTag](create: T): Validation[String] = {

    val jaxbMarshaller: Marshaller = jaxbUserContext.createMarshaller

    // output pretty printed
    //jaxbMarshaller.setProperty(Marshaller.JAXB_FORMATTED_OUTPUT, true)
    jaxbMarshaller.setProperty(Marshaller.JAXB_FRAGMENT, true)

    val document: Document = DocumentBuilderFactory.newInstance.newDocumentBuilder.newDocument

    jaxbMarshaller.marshal(create, document)
    val soapMessage: SOAPMessage = MessageFactory.newInstance.createMessage

    // cosmetic
    // change namespace from SOAP-ENV to soap
    /*
    soapMessage.getSOAPPart().getEnvelope().removeNamespaceDeclaration("SOAP-ENV")
    soapMessage.getSOAPPart().getEnvelope().addNamespaceDeclaration("soap", "http://www.w3.org/2001/12/soap-envelope")
    soapMessage.getSOAPPart().getEnvelope().setPrefix("soap")
    soapMessage.getSOAPHeader().setPrefix("soap")
    soapMessage.getSOAPBody().setPrefix("soap")
    */
    // cosmetic end

    soapMessage.getSOAPBody.addDocument(document)

    val outputStream: ByteArrayOutputStream = new ByteArrayOutputStream();
    soapMessage.writeTo(outputStream);

    Success(outputStream.toString())

  }

  def unserializeUser[T: ClassTag](soapResponse: String): Validation[T] = {
    val soapBody = getSoapBody(soapResponse)
    if (!soapBody.hasFault) {
      val unmarshaller: Unmarshaller = jaxbUserContext.createUnmarshaller
      val response = unmarshaller.unmarshal(soapBody.extractContentAsDocument)
      response match {
        case r: T => Success(r)
        case _    => Failure("unserializeUser: SOAP hasFault flag was not set, but unmarshalling failed.")
      }
    } else Failure("unserializeUser: SOAP hasFault flag was set: Code=" + soapBody.getFault.getFaultCode + ", String=" + soapBody.getFault.getFaultString)
  }
}
