/*
 * Copyright (C) 2017 OX Software GmbH
 *
 * This file is part of OX API-Performance Tests.
 *
 * OX API-Performance Tests is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * OX API-Performance Tests is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OX API-Performance Tests.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.openexchange.testing.performance.soap.webservices

import com.openexchange.testing.performance.soap.JaxbHelper
import com.openexchange.testing.performance.soap.webservices.admin.context._
import com.openexchange.testing.performance.soap.webservices.admin.user.{ ChangeCapabilities, CreateUser }
import com.typesafe.scalalogging.LazyLogging
import io.gatling.core.Predef._
import io.gatling.commons.validation.{ Failure, Success }
import io.gatling.http.Predef._

import com.openexchange.wsdl.context.soap.ListResponse

import scala.concurrent.duration._
import scala.language.reflectiveCalls

import scala.collection.JavaConverters._

/**
 * [[SoapAPI]]
 *
 * @author <a href="mailto:nikolaos.tsapanidis@open-xchange.com">Nikolaos Tsapanidis</a>
 */
object SoapAPI extends LazyLogging {

  def listAllContext(service: String) = group(service) {
    exec(new ListAllContext(service))
  }

  def listContext(service: String, searchPattern: String, startCtxId: Int, keepResponse: Boolean = false) = group(service) {
    // set searchpattern
    exec(session => {
      session.set("search_pattern", searchPattern)
    })

      .exec(new ListContext(service).build()
        .check(bodyString.saveAs("soapResponse")))

      .exec(session => {
        var s = session
        for {
          soapResponse <- session("soapResponse").validate[String]
        } yield {
          JaxbHelper.unserializeCtx[ListResponse](soapResponse) match {
            case Success(lr) => {
              val contextIdsList = for (v <- lr.getReturn.asScala.map(a => a.getId) if v >= startCtxId) yield (v)
              s = s.set("ctx_ids", contextIdsList)

              /*
              println("size of returned list: " + contextIdsList.size)
              for (contextid <- contextIdsList) {
                println("context id: " + contextid)
              }
*/
            }
            case Failure(error) => println("Failure: " + error)
          }
        }
        if (!keepResponse) s = s.remove("soapResponse")
        s.remove("search_pattern")
      })

  }

  def createContext(service: String) = group(service) {
    exec(new CreateContext(service).build()
      .check(bodyString.saveAs("soapResponse")))
      .exec(session => {
        for {
          soapResponse <- session("soapResponse").validate[String]
        } yield {
          JaxbHelper.unserializeCtx[com.openexchange.wsdl.context.soap.CreateResponse](soapResponse) match {
            case Success(cr) => {
              val ctx = cr.getReturn
              // was logger.debug, but I cannot see that anywhere, even with trace logging
              // println("Successfully created context with id " + ctx.getId + " using wsdl-based framework")
            }
            case Failure(error) => println("Failure: " + error)
          }
        }
        session.remove("soapResponse")
      })
  }

  def deleteContext(service: String) = group(service) {
    exec(new DeleteContext(service).build()
      .check(bodyString.saveAs("soapResponse")))

      .exec(session => {
        for {
          soapResponse <- session("soapResponse").validate[String]
        } yield {
          val soapBody = JaxbHelper.getSoapBody(soapResponse)
          if (soapBody.hasFault) {
            println("deleteContext: SOAP hasFault flag was set: Code=" + soapBody.getFault.getFaultCode + ", String=" + soapBody.getFault.getFaultString)
            println("Additional info: contextid=" + session("ctx_id").as[Int].toString)
          }
        }
        session.remove("soapResponse")
      })
  }

  def deleteAllContexts(service: String) = group(service) {
    doIf(session => session.contains("ctx_ids")) {
      foreach("${ctx_ids}", "ctx_id") {
        exec(new DeleteContext(service))
          .pause(100 milliseconds)
          .exec(session => for {
            cid <- session("ctx_id").validate[String]
          } yield {
            // TODO error handling ... see createModuleAccessByName for example
            logger.debug("Successfully deleted context with id {}", cid)
            session
          })
      }
    }
  }

  def createModuleAccessByName(service: String, moduleAccessByName: Option[String]) = group(service) {
    exec(session => {
      session.setAll("access_combination_name" -> moduleAccessByName.getOrElse("webmail_plus"))
    })

      .exec(new CreateModuleAccessByName(service).build()
        .check(bodyString.saveAs("soapResponse")))

      .exec(session => {
        // TODO
        // well, we assume we always get a soapBody, even in an error case. For more robust error handling, we
        // should also check that we have a soapBody in the following, otherwise we'll fail silently.
        for {
          soapResponse <- session("soapResponse").validate[String]
          acn <- session("access_combination_name").validate[String]
        } yield {
          JaxbHelper.unserializeCtx[com.openexchange.wsdl.context.soap.CreateModuleAccessByNameResponse](soapResponse) match {
            case Success(cr) => {
              val ctx = cr.getReturn
              // was logger.debug, but I cannot see that anywhere, even with trace logging
              // println("Successfully created context with id " + ctx.getId + " and access-combination-name=" + acn + " using wsdl-based framework")
            }
            case Failure(error) => println("Failure: " + error)
          }
        }
        session.removeAll("soapResponse", "access_combination_name")
      })

  }

  def createUser(service: String) = group(service) {
    exec(new CreateUser().build()
      .check(bodyString.saveAs("soapResponse")))
      .exec(session => {
        for {
          soapResponse <- session("soapResponse").validate[String]
        } yield {
          JaxbHelper.unserializeUser[com.openexchange.wsdl.user.soap.CreateResponse](soapResponse) match {
            case Success(cr) => {
              val user = cr.getReturn
              // was logger.debug, but I cannot see that anywhere, even with trace logging
              // println("Successfully created user with id " + user.getId + " in context " + session("ctx_id").as[Int] + " using wsdl-based framework")
            }
            case Failure(error) => {
              println("Failure: " + error)
              println("Additional info: contextid=" + session("ctx_id").as[Int].toString + ", username=" + session("name").as[String] + ", email1=" + session("email1").as[String])
            }
          }
        }
        session.removeAll("soapResponse")
      })
  }

  def changeCapabilities(service: String) = group(service) {
    exec(new ChangeCapabilities().build()
      .check(bodyString.saveAs("soapResponse")))
      .exec(session => {
        for {
          soapResponse <- session("soapResponse").validate[String]
        } yield {
          val soapBody = JaxbHelper.getSoapBody(soapResponse)
          if (soapBody.hasFault) {
            println("changeCapabilities: SOAP hasFault flag was set: Code=" + soapBody.getFault.getFaultCode + ", String=" + soapBody.getFault.getFaultString)
            println("Additional info: contextid=" + session("ctx_id").as[Int].toString)
          }
        }
        session.remove("soapResponse")
      })
  }

}
