/*
 *
 *    OPEN-XCHANGE legal information
 *
 *    All intellectual property rights in the Software are protected by
 *    international copyright laws.
 *
 *
 *    In some countries OX, OX Open-Xchange, open xchange and OXtender
 *    as well as the corresponding Logos OX Open-Xchange and OX are registered
 *    trademarks of the OX Software GmbH group of companies.
 *    The use of the Logos is not covered by the GNU General Public License.
 *    Instead, you are allowed to use these Logos according to the terms and
 *    conditions of the Creative Commons License, Version 2.5, Attribution,
 *    Non-commercial, ShareAlike, and the interpretation of the term
 *    Non-commercial applicable to the aforementioned license is published
 *    on the web site http://www.open-xchange.com/EN/legal/index.html.
 *
 *    Please make sure that third-party modules and libraries are used
 *    according to their respective licenses.
 *
 *    Any modifications to this package must retain all copyright notices
 *    of the original copyright holder(s) for the original code used.
 *
 *    After any such modifications, the original and derivative code shall remain
 *    under the copyright of the copyright holder(s) and/or original author(s)per
 *    the Attribution and Assignment Agreement that can be located at
 *    http://www.open-xchange.com/EN/developer/. The contributing author shall be
 *    given Attribution for the derivative code and a license granting use.
 *
 *     Copyright (C) 2017-2020 OX Software GmbH
 *     Mail: info@open-xchange.com
 *
 *
 *     This program is free software; you can redistribute it and/or modify it
 *     under the terms of the GNU General Public License, Version 2 as published
 *     by the Free Software Foundation.
 *
 *     This program is distributed in the hope that it will be useful, but
 *     WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *     or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 *     for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc., 59
 *     Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

package com.openexchange.test.performance.mysql;

import java.net.ConnectException;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;
import java.util.UUID;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;
import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.OptionBuilder;
import org.apache.commons.cli.OptionGroup;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.PosixParser;

/**
 * {@link ConfigDBFiller}
 *
 * @author <a href="mailto:ioannis.chouklis@open-xchange.com">Ioannis Chouklis</a>
 */
public class ConfigDBFiller {

    private static final String GET_FILESTORE_MAX_ID = "SELECT MAX(id) AS id FROM filestore";
    private static final String FILL_FILESTORE = "INSERT INTO filestore (id, uri, size, max_context) VALUES (?,?,?,?)";

    private static final String FILL_DB_POOL = "INSERT INTO db_pool (db_pool_id, url, driver, login, password, hardlimit, max, initial, name) VALUES (?,?,?,?,?,?,?,?,?)";
    private static final String GET_DB_POOL_MAX_ID = "SELECT MAX(db_pool_id) AS db_pool_id FROM db_pool";

    private static final String FILL_DB_CLUSTER = "INSERT INTO db_cluster (cluster_id, read_db_pool_id, write_db_pool_id, weight, max_units) VALUES (?,?,?,?,?)";

    private static final String FILL_CONTEXT = "INSERT INTO context (cid, name, enabled, filestore_id, filestore_name, quota_max) VALUES (?,?,?,?,?,?)";
    private static final String FILL_LOGIN_2_CONTEXT = "INSERT INTO login2context (cid, login_info) VALUES (?,?)";

    private static final String FILL_CONTEXT_SERVER2DB_POOL = "INSERT INTO context_server2db_pool (server_id, cid, read_db_pool_id, write_db_pool_id, db_schema) VALUES (?,?,?,?,?)";

    private static final String[] TABLES = { "filestore", "db_pool", "db_cluster", "context", "login2context", "context_server2db_pool" };
    private static final String TRUNCATE = "TRUNCATE TABLE ";

    private static final Options options;

    private static CountDownLatch done;
    static {
        OptionGroup og = new OptionGroup();

        Option truncate = createOption("t", "truncate", false, "Truncates the configdb", false);
        og.addOption(truncate);

        options = new Options();
        options.addOption(createOption("f", "filestores", true, "The amount of filestores to create (Defaults to 10)", false));
        options.addOption(createOption("p", "db-pools", true, "The amount of database pools to create (Defaults to 20)", false));
        options.addOption(createOption("r", "db-clusters", true, "The amount of database clusters to create (Defaults to 10)", false));
        options.addOption(createOption("c", "contexts", true, "The amount of contexts to create (Defaults to 100)", false));
        options.addOption(createOption("s", "threads", true, "The amount of threads (Defaults to 10)", false));
        options.addOption(createOption("o", "host", true, "The database host (Defaults to localhost)", false));
        options.addOption(createOption("t", "port", true, "The database port (Defaults to 3306)", false));
        options.addOption(createOption("u", "user", true, "The user (Defaults to root)", false));
        options.addOption(createOption("w", "password", true, "The password (Defaults to empty password)", false));
        options.addOption(createOption("h", "help", false, "The help screen", false));

        options.addOptionGroup(og);
    }

    /**
     * Create an {@link Option} with the {@link OptionBuilder}
     *
     * @param shortName short name of the option
     * @param longName long name of the option
     * @param hasArgs whether it has arguments
     * @param hasOptArgs whether it has optional arguments
     * @param description short description
     * @param mandatory whether it is mandatory
     * @return
     */
    @SuppressWarnings("static-access")
    private static final Option createOption(String shortName, String longName, boolean hasArgs, String description, boolean mandatory) {
        return OptionBuilder.withLongOpt(longName).hasArg(hasArgs).withDescription(description).isRequired(mandatory).create(shortName);
    }

    /**
     * Print usage
     *
     * @param exitCode
     */
    private static final void printUsage(int exitCode) {
        HelpFormatter hf = new HelpFormatter();
        hf.setWidth(120);
        hf.printHelp("configdbfiller -flag <argument>", options);
        System.exit(exitCode);
    }

    /**
     * Entry point
     * 
     * @param args
     * @throws ClassNotFoundException
     * @throws SQLException
     */
    public static void main(String[] args) throws ClassNotFoundException, SQLException {
        CommandLineParser parser = new PosixParser();
        try {
            CommandLine cl = parser.parse(options, args);

            if (cl.hasOption('h')) {
                printUsage(0);
            }

            // Fetch the database options 
            String host = cl.getOptionValue('o', "localhost");
            int port = Integer.parseInt(cl.getOptionValue('t', "3306"));
            String user = cl.getOptionValue('u', "root");
            String password = cl.getOptionValue('w', "");

            // Initialise the database and the filler
            Database database = new Database(host, user, password, port, "configdb");
            ConfigDBFiller configDBFiller = new ConfigDBFiller(database);

            // Check for truncation option
            if (cl.hasOption('t')) {
                configDBFiller.truncate();
                System.exit(1);
            }

            // Fetch the amounts
            int filestoreAmount = Integer.parseInt(cl.getOptionValue('f', "10"));
            int dbPoolAmount = Integer.parseInt(cl.getOptionValue('p', "20"));
            int dbClusterAmount = Integer.parseInt(cl.getOptionValue('r', "10"));
            int contextsAmount = Integer.parseInt(cl.getOptionValue('c', "100"));
            int threadCount = Integer.parseInt(cl.getOptionValue('s', "10"));

            // Fill configDB
            configDBFiller.fillFilestore(filestoreAmount);
            configDBFiller.fillDBPool(dbPoolAmount);
            configDBFiller.fillDBCluster(dbClusterAmount);

            // Fill contexts
            done = new CountDownLatch(threadCount);
            long start = System.currentTimeMillis();
            ExecutorService executorService = Executors.newFixedThreadPool(threadCount);
            int contextsChunk = contextsAmount / threadCount;
            for (int index = 0; index < contextsAmount; index += contextsChunk) {
                executorService.execute(new ContextFiller(configDBFiller, contextsChunk, index + 1));
            }
            executorService.shutdown();
            done.await();
            System.out.println("Finished execution of all context inserts in " + TimeUnit.MILLISECONDS.toSeconds(System.currentTimeMillis() - start) + " seconds");
        } catch (Throwable e) {
            e.printStackTrace();
            printUsage(-1);
        }
    }

    /**
     * {@link ContextFiller}
     */
    private static final class ContextFiller implements Runnable {

        private int amount;
        private int startId;
        private ConfigDBFiller filler;

        /**
         * Initialises a new {@link ConfigDBFiller.ContextFiller}.
         */
        public ContextFiller(ConfigDBFiller filler, int amount, int startId) {
            super();
            this.filler = filler;
            this.amount = amount;
            this.startId = startId;
        }

        /*
         * (non-Javadoc)
         * 
         * @see java.lang.Runnable#run()
         */
        @Override
        public void run() {
            long start = System.currentTimeMillis();
            System.out.println("Begining execution of batch " + startId);
            filler.fillContext(amount, startId);
            System.out.println("Finished execution of batch " + startId + " in " + TimeUnit.MILLISECONDS.toSeconds(System.currentTimeMillis() - start) + " seconds.");
            done.countDown();
        }
    }

    private final List<Integer> dbPools;
    private final List<Integer> filestores;

    private final Database database;

    /**
     * Initialises a new {@link ConfigDBFiller}.
     * 
     * @throws SQLException
     * @throws ClassNotFoundException
     */
    public ConfigDBFiller(Database database) throws ClassNotFoundException, SQLException {
        super();
        this.database = database;
        dbPools = new ArrayList<>();
        filestores = new ArrayList<>();
    }

    /**
     * 
     * @param amount
     */
    private void fillFilestore(int amount) {
        PreparedStatement statement = null;
        Connection connection = null;
        try {
            connection = getDatabaseConnection(database);
            connection.setAutoCommit(false);

            statement = connection.prepareStatement(GET_FILESTORE_MAX_ID);
            ResultSet resultSet = statement.executeQuery();
            int startId = 1;
            if (resultSet.next()) {
                startId = resultSet.getInt("id") + 1;
            }

            ProgressBar bar = new ProgressBar();
            bar.update(0, amount);

            for (int c = startId; c < amount + startId; c++) {
                int parameterIndex = 1;
                statement = connection.prepareStatement(FILL_FILESTORE);
                statement.setInt(parameterIndex++, c);
                statement.setString(parameterIndex++, "file:/tmp/filestore_" + c);
                statement.setLong(parameterIndex++, 1048576000000L);
                statement.setInt(parameterIndex, 2000000);

                statement.executeUpdate();
                statement.close();
                filestores.add(c);
                bar.update(c, amount);
            }
            connection.setAutoCommit(true);
        } catch (SQLException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } finally {
            DBUtil.closeStatement(statement);
            DBUtil.closeConnection(connection);
        }
    }

    /**
     * 
     * @param amount
     */
    private void fillDBPool(int amount) {
        PreparedStatement statement = null;
        Connection connection = null;
        try {
            connection = getDatabaseConnection(database);
            connection.setAutoCommit(false);
            statement = connection.prepareStatement(GET_DB_POOL_MAX_ID);
            ResultSet resultSet = statement.executeQuery();
            int startId = 1;
            if (resultSet.next()) {
                startId = resultSet.getInt("db_pool_id") + 1;
            }

            ProgressBar bar = new ProgressBar();
            bar.update(0, amount);
            for (int c = startId; c < amount + startId; c++) {
                int parameterIndex = 1;
                statement = connection.prepareStatement(FILL_DB_POOL);
                statement.setInt(parameterIndex++, c);
                statement.setString(parameterIndex++, "jdbc:mysql://localhost");
                statement.setString(parameterIndex++, "com.mysql.jdbc.Driver");
                statement.setString(parameterIndex++, "somelogin");
                statement.setString(parameterIndex++, "somepassword");
                statement.setInt(parameterIndex++, 1);
                statement.setInt(parameterIndex++, 1);
                statement.setInt(parameterIndex++, 1);
                statement.setString(parameterIndex++, UUID.randomUUID().toString());

                statement.executeUpdate();
                statement.close();
                dbPools.add(c);
                bar.update(c, amount);
            }
            connection.setAutoCommit(true);
        } catch (SQLException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } finally {
            DBUtil.closeStatement(statement);
            DBUtil.closeConnection(connection);
        }
    }

    /**
     * 
     * @param amount
     */
    private void fillDBCluster(int amount) {
        PreparedStatement statement = null;
        Connection connection = null;
        try {
            connection = getDatabaseConnection(database);
            connection.setAutoCommit(false);

            int parameterIndex = 1;
            int poolSize = dbPools.size();
            ProgressBar bar = new ProgressBar();
            bar.update(0, amount);
            for (int c = 1; c <= amount; c++) {
                statement = connection.prepareStatement(FILL_DB_CLUSTER);
                statement.setInt(parameterIndex++, c);
                statement.setInt(parameterIndex++, dbPools.get(c % poolSize));
                statement.setInt(parameterIndex++, dbPools.get((c + 1) % poolSize));
                statement.setInt(parameterIndex++, 100);
                statement.setInt(parameterIndex++, 100);

                statement.executeUpdate();
                statement.close();
                parameterIndex = 1;
                bar.update(c, amount);
            }
            connection.setAutoCommit(true);
        } catch (SQLException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } finally {
            DBUtil.closeStatement(statement);
            DBUtil.closeConnection(connection);
        }
    }

    /**
     * 
     * @param amount,
     */
    private void fillContext(int amount, int startId) {
        PreparedStatement statement = null;
        Connection connection = null;
        try {
            connection = getDatabaseConnection(database);
            int filestoreSize = filestores.size();
            int poolSize = dbPools.size();
            int parameterIndex = 1;

            ProgressBar bar = new ProgressBar();
            bar.update(0, amount + startId);

            for (int c = startId; c < amount + startId; c++) {
                connection.setAutoCommit(false);
                statement = connection.prepareStatement(FILL_CONTEXT);
                statement.setInt(parameterIndex++, c);
                statement.setString(parameterIndex++, "Name-" + c);
                statement.setInt(parameterIndex++, 1);
                statement.setInt(parameterIndex++, filestores.get(c % filestoreSize));
                statement.setString(parameterIndex++, c + "_ctx_store");
                statement.setLong(parameterIndex++, 5242880000L);

                statement.executeUpdate();
                statement.close();
                parameterIndex = 1;

                statement = connection.prepareStatement(FILL_LOGIN_2_CONTEXT);
                statement.setInt(parameterIndex++, c);
                statement.setString(parameterIndex++, Integer.toString(c));
                statement.executeUpdate();
                statement.close();
                parameterIndex = 1;

                statement = connection.prepareStatement(FILL_LOGIN_2_CONTEXT);
                statement.setInt(parameterIndex++, c);
                statement.setString(parameterIndex++, "Name-" + c);
                statement.executeUpdate();
                statement.close();
                parameterIndex = 1;

                statement = connection.prepareStatement(FILL_CONTEXT_SERVER2DB_POOL);
                statement.setInt(parameterIndex++, 1);
                statement.setInt(parameterIndex++, c);
                statement.setInt(parameterIndex++, dbPools.get(c % poolSize));
                statement.setInt(parameterIndex++, dbPools.get((c + 1) % poolSize));
                statement.setString(parameterIndex++, "db_schema_" + c);
                statement.executeUpdate();
                statement.close();
                parameterIndex = 1;
                connection.setAutoCommit(true);
                bar.update(c, amount + startId);
            }
        } catch (SQLException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } finally {
            DBUtil.closeStatement(statement);
            DBUtil.closeConnection(connection);
        }
    }

    /**
     * Truncates all tables
     */
    private void truncate() {
        PreparedStatement statement = null;
        Connection connection = null;
        try {
            connection = getDatabaseConnection(database);
            connection.setAutoCommit(false);
            statement = connection.prepareStatement("SET FOREIGN_KEY_CHECKS = 0");
            statement.executeQuery();

            for (String table : TABLES) {
                statement = connection.prepareStatement(TRUNCATE + table);
                statement.executeUpdate();
                statement.close();
            }

            statement = connection.prepareStatement("SET FOREIGN_KEY_CHECKS = 1");
            statement.executeQuery();
            connection.setAutoCommit(true);
        } catch (SQLException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } finally {
            DBUtil.closeStatement(statement);
            DBUtil.closeConnection(connection);
        }
    }

    /**
     * Establishes a connection to the specified {@link Database}
     * 
     * @param database The {@link Database} to establish connection to
     * @return The {@link ConnectException}
     * @throws ClassNotFoundException
     * @throws SQLException
     */
    private Connection getDatabaseConnection(Database database) throws SQLException {
        String driverName = "com.mysql.jdbc.Driver";
        StringBuilder urlBuilder = new StringBuilder();
        urlBuilder.append("jdbc:mysql://").append(database.getHost()).append("/").append(database.getDatabaseName());
        try {
            Class.forName(driverName);
        } catch (ClassNotFoundException e) {
            e.printStackTrace();
            System.exit(-1);
        }
        Connection connection = DriverManager.getConnection(urlBuilder.toString(), database.getUser(), database.getPasswd());
        return connection;
    }
}
