/*
 * Copyright (C) 2016 OX Software GmbH
 *
 * This file is part of uimetrics.
 *
 * uimetrics is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * uimetrics is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with uimetrics.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.openexchange.test.performance.uimetrics.cli;

import com.openexchange.test.performance.uimetrics.config.Configuration;
import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.OptionGroup;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;
import org.apache.commons.cli.PosixParser;

/**
 * The {@link CLIParser} is responsible for the user input
 *
 * @author <a href="mailto:markus.wagner@open-xchange.com">Markus Wagner</a>
 */
public class CLIParser {

    private static String[] ARGUMENTS;

    private static Options OPTIONS;

    private static CommandLineParser PARSER;

    private static Configuration CONFIGURATION;

    private static boolean INITIALIZED = false;

    final static String HELP = "h";

    final static String URL = "U";

    final static String USER = "u";

    final static String CONTEXT = "c";

    final static String PASSWORD = "p";

    final static String IE = "I";

    final static String FIREFOX = "F";

    final static String CHROME = "C";

    final static String SAFARI = "S";

    final static String BROWSER_BINARY = "B";

    public CLIParser(String[] arguments) {
        ARGUMENTS = arguments;
        PARSER = new PosixParser();
        CONFIGURATION = Configuration.getInstance();

        defineOptions();
        checkForHelp();
        parseOptions();
    }

    /**
     * Methode to check if the cmd parser is ready
     *
     * @return true, if the command line was succesfully parsed
     */
    public boolean isInitialized() {
        return INITIALIZED;
    }

    /**
     * check if the user has given the -h or --help option
     */
    private void checkForHelp() {
        try {
            if (PARSER.parse(new Options().addOption("h", "help", false, "Show this help"), ARGUMENTS, true).hasOption(HELP) || ARGUMENTS.length == 0) {
                HelpFormatter helpFormatter = new HelpFormatter();
                helpFormatter.setWidth(300);
                helpFormatter.printHelp("java -jar " + new java.io.File(CLIParser.class.getProtectionDomain().getCodeSource().getLocation().getPath()).getName(), OPTIONS, true);
                System.exit(0);
            }
        } catch (ParseException e) {
            System.err.println(e.getLocalizedMessage());
            System.exit(0);
        }
    }

    /**
     * initialize the command line options
     */
    private void defineOptions() {
        OPTIONS = new Options();

        OPTIONS.addOption(URL, "url", true, "URL to the OX-Server, e.g. https;//example.org");

        OptionGroup browser = new OptionGroup();
        browser.addOption(new Option(IE, "iexplorer", false, "Use the Internet Explorer (only on Windows)"));
        browser.addOption(new Option(FIREFOX, "firefox", false, "Use Firefox"));
        browser.addOption(new Option(CHROME, "chrome", false, "Use Chrome"));
        browser.addOption(new Option(SAFARI, "safari", false, "Use Safari (only on Mac OSX)"));
        browser.setRequired(true);

        OPTIONS.addOptionGroup(browser);

        OPTIONS.addOption(BROWSER_BINARY, "binary", true, "path to the browser binary (e.g. \"C:\\Programme\\Mozilla Firefox\\firefox.exe\")");
        OPTIONS.addOption(USER, "user", true, "OX App Suite username");
        OPTIONS.addOption(CONTEXT, "context", true, "OX App Suite user context");
        OPTIONS.addOption(PASSWORD, "password", true, "OX App Suite user password");

        OPTIONS.getOption(URL).setRequired(true);
        OPTIONS.getOption(USER).setRequired(true);
        OPTIONS.getOption(PASSWORD).setRequired(true);
    }

    private void parseOptions() {
        try {
            CommandLine cmd = PARSER.parse(OPTIONS, ARGUMENTS);

            CONFIGURATION.set(Configuration.Property.OX_HOST, cmd.getOptionValue(URL));
            CONFIGURATION.set(Configuration.Property.USER, cmd.getOptionValue(USER));
            if (cmd.hasOption(CONTEXT)) {
                CONFIGURATION.set(Configuration.Property.CONTEXT, cmd.getOptionValue(CONTEXT));
            }
            CONFIGURATION.set(Configuration.Property.PASSWORD, cmd.getOptionValue(PASSWORD));

            if (cmd.hasOption(IE)) {
                CONFIGURATION.set(Configuration.Property.USE_IE, true);
                if (cmd.hasOption(BROWSER_BINARY)) {
                    CONFIGURATION.set(Configuration.Property.IE_BINARY, cmd.getOptionValue(BROWSER_BINARY));
                }
            } else if (cmd.hasOption(FIREFOX)) {
                CONFIGURATION.set(Configuration.Property.USE_FIREFOX, true);
                if (cmd.hasOption(BROWSER_BINARY)) {
                    CONFIGURATION.set(Configuration.Property.FIREFOX_BINARY, cmd.getOptionValue(BROWSER_BINARY));
                }
            } else if (cmd.hasOption(CHROME)) {
                CONFIGURATION.set(Configuration.Property.USE_CHROME, true);
                if (cmd.hasOption(BROWSER_BINARY)) {
                    CONFIGURATION.set(Configuration.Property.CHROME_BINARY, cmd.getOptionValue(BROWSER_BINARY));
                }
            } else if (cmd.hasOption(SAFARI)) {
                CONFIGURATION.set(Configuration.Property.USE_SAFARAI, true);
            }
            INITIALIZED = true;
        } catch (ParseException e) {
            System.err.println(e.getLocalizedMessage());
            System.exit(0);
        }
    }
}
