/*
 * Copyright (C) 2016 OX Software GmbH
 *
 * This file is part of uimetrics.
 *
 * uimetrics is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * uimetrics is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with uimetrics.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.openexchange.test.performance.uimetrics.config;

import java.net.MalformedURLException;
import java.net.URL;
import java.util.Properties;
import org.apache.commons.lang3.SystemUtils;

/**
 * Basic configuration, more or less initialized and set by the {@link CLIParser}
 *
 * @author Markus Wagner <markus.wagner@open-xchange.com>
 */
public class Configuration {

    private static Configuration config = null;

    private final Properties p = new Properties();

    public enum Property {
        OX_HOST("ox.host"),
        USER("ox.user"),
        CONTEXT("ox.context"),
        PASSWORD("ox.password"),
        USE_IE("ox.ie"),
        IE_BINARY("ox.ie.binary"),
        USE_FIREFOX("ox.firefox"),
        FIREFOX_BINARY("ox.firefox.binary"),
        USE_CHROME("ox.chrome"),
        CHROME_BINARY("ox.chrome.binary"),
        USE_SAFARAI("ox.safari");

        private String name;

        private Property(final String name) {
            this.name = name;
        }

        public String getName() {
            return name;
        }
    }

    private Configuration() {
        // set some defaults
        p.setProperty(Property.IE_BINARY.name, "iexplorer.exe");
        p.setProperty(Property.FIREFOX_BINARY.name, SystemUtils.IS_OS_WINDOWS ? "firefox.exe" : "firefox");
        p.setProperty(Property.CHROME_BINARY.name, SystemUtils.IS_OS_WINDOWS ? "chrome.exe" : "google-chrome");
        p.setProperty(Property.IE_BINARY.name, "iexplorer.exe");
        p.setProperty(Property.IE_BINARY.name, "iexplorer.exe");
    }

    public static Configuration getInstance() {
        if (null == config) {
            config = new Configuration();
        }
        return config;
    }

    public String get(final Property property) {
        return get(property.getName());
    }

    public boolean getAsBoolean(final Property property) {
        return getAsBoolean(property.getName());
    }

    /**
     * Get the OX host URL
     *
     * @return the hosts URL as {@link String} (e.g. http://example.org/)
     */
    public String getOXHost() {
        String host = null;
        try {
            host = new URL(this.get(Property.OX_HOST)).toString();
            if (!host.endsWith("/")) {
                host += "/";
            }
        } catch (final MalformedURLException e) {
            e.printStackTrace();
        }
        return host;
    }

    public void set(final Property property, final String value) {
        p.setProperty(property.name, value);
    }

    public void set(final Property property, final Boolean value) {
        p.setProperty(property.name, Boolean.toString(value));
    }

    // below are some old c&p methods, leave them here, perhaps we can use it

    public String get(final String key, final String def) {
        final String value = p.getProperty(key);
        return (value == null || value.equals("")) ? def : value;
    }

    public String get(final String key) {
        return get(key, null);
    }

    public String get(final Property property, final String def) {
        return get(property.getName(), def);
    }

    public boolean getAsBoolean(final String name, final String def) {
        final String value = get(name, def);
        if ("true".equals((def).toLowerCase()) || "true".equals((value).toLowerCase())) {
            return true;
        } else {
            return false;
        }
    }

    public boolean getAsBoolean(final String name) {
        return getAsBoolean(name, "false");
    }

    public boolean getAsBoolean(final Property property, final String def) {
        return getAsBoolean(property.getName(), def);
    }

    public int getAsInt(final String name, final int def) {
        final String value = get(name);
        if (value == null) {
            return def;
        }
        return Integer.parseInt(value);
    }

    public int getAsInt(final String name) {
        return getAsInt(name, -1);
    }

    public int getAsInt(final Property property, final int def) {
        return getAsInt(property.getName(), def);
    }

    public int getAsInt(final Property property) {
        return getAsInt(property.getName());
    }

}
